import es

from . import console
from . import srcds_types

is_loaded = False

info = es.AddonInfo()
info.name = "SRCDS tools"
info.basename = "srcds"
info.description = "Python package for advanced SRCDS stuff in EventScripts 2.1"
info.author = "GODJonez"
info.version = "dev20110416"


class SRCDSError(RuntimeError):
    '''Error caused by limitations in the Source engine.'''
    pass


class ImplementationWarning(Warning):
    '''Warning of a feature that has not been properly implemented yet.'''
    pass


def load():
    global is_loaded
    is_loaded = True
    if not es.exists('command', '_start_console_capture'):
        es.regcmd('_start_console_capture', 'srcds/ccmd_start_console_capture', 'Start capturing console output')
    if not es.exists('command', '_end_console_capture'):
        es.regcmd('_end_console_capture', 'srcds/ccmd_end_console_capture', 'Stop capturing console output')
    if not es.exists('command', 'srcds_ent'):
        es.regcmd('srcds_ent', 'srcds/ccmd_srcds_ent', 'Info on entities implemented in srcds.entities')


def unload():
    global is_loaded
    is_loaded = False


def ccmd_start_console_capture():
    if es.getargc() != 2:
        es.dbgmsg(0, 'Error: This command must not be used manually')
        return False
    try:
        thisid = int(es.getargv(1))
    except ValueError:
        es.dbgmsg(0, 'Error: This command must not be used manually')
        return False
    if thisid not in console._concmd_handlers:
        es.dbgmsg(0, 'Error: invalid handler id')
        return False
    console._concmd_handlers.start(thisid)
    return True


def ccmd_end_console_capture():
    if es.getargc() != 2:
        es.dbgmsg(0, 'Error: This command must not be used manually')
        return False
    try:
        thisid = int(es.getargv(1))
    except ValueError:
        es.dbgmsg(0, 'Error: This command must not be used manually')
        return False
    if thisid not in console._concmd_handlers:
        es.dbgmsg(0, 'Error: invalid handler id')
        return False
    console._concmd_handlers.end(thisid)
    return True


def ccmd_srcds_ent():
    from . import entities
    if es.getargc() == 1:
        es.dbgmsg(0, 'Usage: srcds_ent {classname|entindex}')
        es.dbgmsg(0, ' - show information about specific entity class')
        es.dbgmsg(0, ' - or show information about specific entity on game')
        return False
    arg = es.getargv(1)
    if arg.isdigit():
        entindex = int(arg)
        try:
            entity = entities.get_entity(entindex)
        except SRCDSError, error:
            es.dbgmsg(0, str(error))
            return False
        entity_data = entity.get_entity_data()
        classname = entity_data.pop('classname')
        print_keyvalues('classname', classname, None)
        for key, value in sorted(entity_data.iteritems()):
            print_keyvalues(key, value, None)
    else:
        try:
            entclass = entities.entity_class_lookup[arg]
        except KeyError, error:
            es.dbgmsg(0, 'Unknown entity class')
            return False
        keys, methods = entclass.get_entity_classdata()
        es.dbgmsg(0, '')
        for line in pretty_doc(entclass.__doc__):
            es.dbgmsg(0, line)
        es.dbgmsg(0, '')
        es.dbgmsg(0, 'Keyvalues for "%s":' % (arg,))
        for keyvalue_pack in keys:
            es.dbgmsg(0, ' ')
            if not keyvalue_pack[1]:
                es.dbgmsg(0, '  %s (nothing)' % (keyvalue_pack[0],))
            else:
                es.dbgmsg(0, '  %s:' % (keyvalue_pack[0],))
                for key, value in sorted(keyvalue_pack[1].iteritems()):
                    print_keyvalues(key, value[0], value[1])
        es.dbgmsg(0, ' ')
        es.dbgmsg(0, 'Methods for "%s":' % (arg,))
        es.dbgmsg(0, ' ')
        for method_pack in methods:
            if not method_pack[1]:
                es.dbgmsg(0, '  %s (nothing)' % (method_pack[0],))
            else:
                es.dbgmsg(0, '  %s:' % (method_pack[0],))
                for name, value in sorted(method_pack[1].iteritems()):
                    doc, additional = value
                    if additional is not None:
                        additional = ' (overridden by '+additional+')'
                    else:
                        additional = ''
                    es.dbgmsg(0, '      %s%s' % (name, additional))
                    for line in pretty_doc(doc, 10):
                        es.dbgmsg(0, line)
                    es.dbgmsg(0, '      ')


def print_keyvalues(key, value, overridden):
    '''Print out info of a keyvalues object.'''
    if overridden is None:
        additional = ''
    else:
        additional = ' (overridden by '+overridden+')'
    if isinstance(value, srcds_types.SymbolicValue):
        es.dbgmsg(0, '      %-25s  %s%s' % (key, value.getvalue(), additional))
        for symname, symvalue in sorted(
            value._keywords.iteritems(),
            key=lambda item: item[1]
        ):
            es.dbgmsg(0, '    %8d  %s' % (symvalue, symname))
    elif isinstance(value, srcds_types.Flags):
        es.dbgmsg(0, '      %-25s  %s' % (key, additional))
        for symname, symvalue in sorted(
            value._flags.iteritems(),
            key=lambda item: item[1]
        ):
            if value._value & symvalue:
                es.dbgmsg(0, '          [X]  %s' % (symname,))
            else:
                es.dbgmsg(0, '          [ ]  %s' % (symname,))  
    elif isinstance(value, srcds_types.RangedValue):
        es.dbgmsg(0, '      %-25s  %-8s [%s - %s]' % (
            key, value.getvalue(), value._min, value._max))
    else:
        es.dbgmsg(0, '      %-25s  %s%s' % (key, value, additional))


def pretty_doc(text, indent=4, width=80):
    '''
    Generate pretty format for documentation with specified indent and max
    line width. Yields one line at a time.
    '''
    lines = text.splitlines()
    # discard empty lines at start
    try:
        line = lines[0]
        while not line.strip():
            lines.pop(0)
            line = lines[0]
    except IndexError:
        # list became empty, no documentation
        return
    original_indent = len(line) - len(line.lstrip())
    # discard empty line from end
    try:
        line = lines[-1]
        while not line.strip():
            lines.pop()
            line = lines[-1]
    except IndexError:
        # list became empty, no documentation
        return
    section = []
    sections = [section]
    for line in lines:
        if not line.strip():
            section = []
            sections.append(['']) # empty line
            sections.append(section)
            continue
        preindent = line[:original_indent]
        if preindent.strip():
            # unexpected unindent
            section = []
            sections.append(section)
            newline = line.lstrip()
        else:
            newline = line[original_indent:]
            if len(newline.lstrip()) != len(newline):
                # indent, force new section
                section = []
                sections.append(section)
        section.extend(newline.split(' '))
    for section in sections:
        line = [' '*(indent-1)]
        length = indent
        for word in section:
            addlen = len(word)+1
            if length+addlen <= width:
                line.append(word)
                length += addlen
            else:
                yield ' '.join(line)
                line = [' '*(indent-1), word]
                length = indent+addlen
        if length > indent:
            yield ' '.join(line)
