'''entity classes'''

import random

import es
dbgmsg = es.dbgmsg

import gamethread

from .. import srcds_types
from .. import srcds

# DEBUG DEBUG
from ..debug import DebugProxy
# uncomment the next line to dump on console all the es-module interactions from here
# es = DebugProxy(es)
# DEBUG DEBUG

Vector = srcds_types.Vector
Color = srcds_types.Color
RangedValue = srcds_types.RangedValue
SymbolicValue = srcds_types.SymbolicValue
SymbolicBoolean = srcds_types.SymbolicBoolean
Flags = srcds_types.Flags

SRCDSError = srcds.SRCDSError

INVALID_INDEX = -1
NO_CREATE = -2


def get_bases():
    '''
    Return a dictionary holding all the global names that could be needed
    by other modules defining more entities.
    '''
    bases = {
        'Vector': Vector,
        'Color': Color,
        'RangedValue': RangedValue,
        'SymbolicValue': SymbolicValue,
        'SymbolicBoolean': SymbolicBoolean,
        'Flags': Flags,
        '_KeyValue': _KeyValue,
        'KeyvalueInput': KeyvalueInput,
    }
    baseclasses = dict(
        filter(
            lambda item: (
                item[0][0] == '_' and isinstance(item[1], type) and
                issubclass(
                    item[1],
                    _BaseEntity
                )
            ),
            globals().iteritems()
        )
    )
    bases.update(baseclasses)
    return bases


class _KeyValue(object):
    '''Internal class to handle entity keyvalues.'''
    def __init__(self, value, name=None, readonly=False):
        self.name = name
        self.default_value = value
        self._type = type(value)
        self._readonly = readonly
    
    def __get__(self, entity, entclass):
        '''
        Get this keyvalue.
        '''
        if entity is None:
            # EntityClass.keyvalue
            return self
        raw_value = es.entitygetvalue(entity._index, self.name)
        try:
            value = self.coerce(raw_value)
        except (ValueError, TypeError), error:
            es.dbgmsg(0, 'Entity #%d (%s) has incompatible value %r for %s (%s)' % (
                entity._index, entclass.__name__, raw_value, self.name, error))
            return raw_value
        else:
            if hasattr(value, 'set_callback'):
                value.set_callback(lambda new_value: self.__set__(entity, new_value))
            return value
    
    def __set__(self, entity, value):
        '''
        Set new value.
        '''
        if self._readonly:
            raise TypeError('Cannot set value for read-only keyvalues.')
        raw_value = self.engine_value(self.coerce(value))
        es.entitysetvalue(entity._index, self.name, raw_value)

    def coerce(self, value):
        '''Convert value into same type as the default value.'''
        if type(value) != self._type:
            # incompatible type, try creating a new instance of the original type
            if hasattr(self.default_value, 'coerce'):
                value = self.default_value.coerce(value)
            else:
                if not value:
                    value = self._type()
                value = self._type(value)
        
        return value

    @staticmethod
    def engine_value(value):
        '''Get the key value as it should be given to the engine.'''
        if hasattr(value, 'getvalue_engine'):
            return value.getvalue_engine()
        else:
            return str(value)


class KeyvalueInput(_KeyValue):
    '''
    A type that has an Input of the exact same name to change value.
    '''
    class Proxy(object):
        def __init__(self, value):
            self.__value = value
        
        def __getattr__(self, attribute):
            if attribute == '_Proxy__value':
                return super(Proxy, self).__getattr__(attribute)
            return getattr(self.__value, attribute)
        
        def __setattr__(self, attribute, value):
            if attribute == '_Proxy__value':
                return super(Proxy, self).__setattr__(attribute, value)
            return setattr(self.__value, attribute)
    
    
    def __get__(self, entity, entclass):
        value = Proxy(_KeyValue.__get__(self, entity, entclass))
        def callfunc(proxy, param, userid=0):
            return entity.fire(self.name, param, userid=userid)
        value.__call__ = callfunc


class _MetaEntity(type):
    '''
    Meta class for entities that does magic for _KeyValue objects
    and adds classname keyvalue.
    '''
    def __init__(self, name, bases, contents):
        '''
        Creating a class with this metaclass as base.
        
        Add classname attribute to the class.
        '''
        self._keyvalues = {}
        for key, value in contents.iteritems():
            if isinstance(value, _KeyValue):
                value.name = key
                self._keyvalues[key] = value
    
    def __setattr__(self, attribute, value):
        '''
        Class attribute definition, check if value is _KeyValue object.
        '''
        if isinstance(value, _KeyValue):
            value.name = attribute
            self._keyvalues[attribute] = value
        type.__setattr__(self, attribute, value)


class _BaseEntity(object):
    '''
    Base entity class for all entities.

    Adds the following keyvalues to entities:
        classname
            The entity's classname (e.g. info_target)

        angles <pitch yaw roll>
            An angular world orientation in degrees.

        origin <x y z>
            The location in the world.
    '''
    __metaclass__ = _MetaEntity
    angles = _KeyValue(Vector((0,0,0)))
    origin = _KeyValue(Vector((0,0,0)))
    
    def __init__(self, index=INVALID_INDEX):
        cls = self.__class__
        cls.classname = _KeyValue(cls.__name__)
        
        if index == INVALID_INDEX:
            self._create()
        elif index == NO_CREATE:
            self._index = INVALID_INDEX
        elif index >= 0:
            self._index = index
        
        self._tempname = None # used for self.fire
    
    def __repr__(self):
        return '%s(%d)' % (self.__class__.__name__, self._index)
    
    def __int__(self):
        '''int(x), returns the entity index of the entity.'''
        return int(self._index)
    
    def _create(self):
        '''Create the entity and assign it entity index.'''
        self._index = es.createentity(self.__class__.__name__)
        return self._index
    
    def spawn(self):
        '''Spawn this entity to the world.'''
        es.spawnentity(self._index)

    def _random_name(self):
        '''Generate a random targetname.'''
        return "%s_%010.8f"%(self.__class__.__name__, random.random())
    
    def _reset_name(self):
        '''Reset name after firing inputs with unique name.'''
        if self._index == INVALID_INDEX:
            return
        self._tempname[0] -= 1
        if self._tempname[0] > 0:
            gamethread.delayed(0, self._reset_name)
        else:
            self._tempname = None
            es.setentityname(self._index, self.targetname)
    
    def fire(self, fire_input, input_params=None, userid=0, fatal=False):
        '''Fire an input on the entity.'''
        if self._index == INVALID_INDEX:
            raise SRCDSError('Entity does not exist.')
        userid = userid or es.getuserid()
        if userid:
            if self._tempname is None:
                self._tempname = [2, self._random_name()]
                gamethread.delayed(0, self._reset_name)
                es.setentityname(self._index, self._tempname[1])
            elif self._tempname[0] == 1:
                self._tempname[0] = 2
            if input_params is None:
                es.fire(userid, self._tempname[1], fire_input)
            else:
                es.fire(userid, self._tempname[1], fire_input, str(input_params))
            if fatal:
                self._index = INVALID_INDEX
        else:
            raise SRCDSError('''Cannot fire inputs when no users are active.
        Your input was <%s>'''%fire_input)

    def fire_group(self, fire_input, input_params=None, userid=0):
        '''
        Fire an input on all the entities that share the targetname/classname.
        '''
        userid = userid or es.getuserid()
        if userid:
            if input_params is None:
                es.fire(userid, self.targetname or self.classname, fire_input)
            else:
                es.fire(userid, self.targetname or self.classname, fire_input, str(input_params))
        else:
            raise SRCDSError('''Cannot fire inputs when no users are active.
        Your input was <%s>'''%fire_input)

    def copy(self):
        '''Create a new entity using this entity as a template.'''
        ent_class = type(self)
        if ent_class is _UnknownEntity:
            raise SRCDSError('Cannot copy unknown entities')
        new_entity = ent_class()
        # copy keyvalues using the _KeyValue magic inside
        for attribute in type(self)._inherited_keyvalues():
            setattr(new_entity, attribute, getattr(self, attribute))
        return new_entity

    def remove(self):
        '''Remove entity from world.'''
        # to prevent crashes, check if the entity really exists
        if self._index in es.getEntityIndexes():
            es.remove(self._index)
        self._index = INVALID_INDEX
    
    def get_entity_data(self):
        '''Return all keyvalues and their raw data in a dictionary.'''
        return dict(
            ( # generator syntax
                (key, getattr(self, key)) for key in type(self)._inherited_keyvalues()
            )
        )
    
    @classmethod
    def _inherited_keyvalues(cls):
        '''
        Return a set of keyvalue names from this and all baseclasses.
        '''
        values = set()
        for baseclass in cls.__mro__:
            if hasattr(baseclass, '_keyvalues'):
                values.update(baseclass._keyvalues)
        return values
    
    @classmethod
    def get_entity_classdata(cls):
        '''
        Return information of all keyvalues and methods supported by the class.
        
        Return value is a tuple (keyvalues, methods), where each member is a
        list of tuples (classname, contents), ordered by the method resolution
        order (entity class itself is the first, _BaseEntity is the last).
        
        classname is the name of the class where the contents are defined in,
        contents is a dictionary with keys being names (ketvalues or methods)
        and values being tuples (value, overridden).
        
        For keyvalues the value is the raw value instance with default value,
        for methods it is the method docstring. overridden specifies the
        classname which ultimately has overridden the current item, or None.
        '''
        keys = []
        methods = []
        seen_keys = dict()
        seen_methods = dict()
        for baseclass in cls.__mro__:
            clsname = baseclass.__name__
            if hasattr(baseclass, '_keyvalues'):
                ckeys = {}
                keys.append((clsname, ckeys))
                for keyname, keyvalue in baseclass._keyvalues.iteritems():
                    if keyname not in seen_keys:
                        overridden = None
                        seen_keys[keyname] = clsname
                    else:
                        overridden = seen_keys[keyname]
                    ckeys[keyname] = keyvalue.default_value, overridden
            cmethods = {}
            methods.append((clsname, cmethods))
            for keyname, keyvalue in baseclass.__dict__.iteritems():
                if keyname[0] != '_' and callable(keyvalue):
                    if keyname not in seen_keys:
                        overridden = None
                        seen_methods[keyname] = clsname
                    else:
                        overridden = seen_methods[keyname]
                    cmethods[keyname] = keyvalue.__doc__, overridden
        return keys, methods
                    

class _UnknownKeys(_BaseEntity):
    '''
    Baseclass allowing getting and setting of unknown keys.
    '''
    def __setattr__(self, attribute, value):
        if attribute in self._keyvalues:
            super(_UnknownKeys, self).__setattr__(attribute, value)
        elif attribute[0] != '_':
            es.entitysetvalue(self._index, attribute, value)
        else:
            super(_UnknownKeys, self).__setattr__(attribute, value)
    
    def __getattr__(self, attribute):
        return es.entitygetvalue(self._index, attribute)


class _UnknownEntity(_UnknownKeys):
    '''
    Entity class used for entities not explicitly implemented.
    '''
    pass


class _Targetname(_BaseEntity):
    '''
    Implement targetname keyvalue and related inputs.
    '''
    targetname = _KeyValue('')
    
    def __str__(self):
        '''str(x), returns the targetname of the entity.'''
        return str(self.targetname)

    def Kill(self, userid=0):
        '''Remove this entity from the world.'''
        self.fire('Kill', userid=userid, fatal=True)

    def KillHierarchy(self, userid=0):
        '''Remove this entity and all its children from the world.'''
        self.fire('KillHierarchy', userid=userid, fatal=True)

    def AddOutput(self, text, userid=0):
        '''
        Evaluate a keyvalue/output on this entity.
        Format: <key> <value>
        Format: <output name> <targetname>,<inputname>,<parameter>,<delay>,<max times to fire (-1 == infinite)>
        '''
        self.fire('AddOutput', text, userid=userid)

    def FireUser1(self, userid=0):
        '''Cause this entity's OnUser1 output to be fired.'''
        self.FireUser(self, 1, userid)

    def FireUser2(self, userid=0):
        '''Cause this entity's OnUser2 output to be fired.'''
        self.FireUser(self, 2, userid)

    def FireUser3(self, userid=0):
        '''Cause this entity's OnUser3 output to be fired.'''
        self.FireUser(self, 3, userid)

    def FireUser4(self, userid=0):
        '''Cause this entity's OnUser4 output to be fired.'''
        self.FireUser(self, 4, userid)

    def FireUser(self, num, userid=0):
        '''Cause this entity's OnUser<num> output to be fired.'''
        self.fire('FireUser%d'%num, userid=userid)


class _Shadow(_BaseEntity):
    '''
    Base class for shadow casting entities.
    '''
    disableshadows = _KeyValue(SymbolicBoolean(0))

    def DisableShadow(self, userid=0):
        '''Turn shadow off.'''
        self.fire('DisableShadow', userid=userid)

    def EnableShadow(self, userid=0):
        '''Turn shadow on.'''
        self.fire('EnableShadow', userid=userid)


class _Studiomodel(_Targetname, _Shadow):
    '''
    Base class for entities having specific model.
    '''
    model = _KeyValue('')
    skin = _KeyValue(0)

    def Skin(self, skin, userid=0):
        '''Change the model skin to the specified number.'''
        self.fire('Skin', skin, userid=userid)

    def AlternativeSorting(self, sorting, userid=0):
        '''Attempt to fix sorting problems when rendering.'''
        self.fire('AlternativeSorting', sorting, userid=userid)


class _BasePlat(_BaseEntity):
    '''
    Base entity class for platforms.
    '''
    def Toggle(self, userid=0):
        '''Toggle the platform state.'''
        self.fire('Toggle', userid=userid)
    
    def GoUp(self, userid=0):
        '''Tell the platform to go up.'''
        self.fire('GoUp', userid=userid)
    
    def GoDown(self, userid=0):
        '''Tell the platform to go down.'''
        self.fire('GoDown', userid=userid)


class _Parentname(_BaseEntity):
    '''
    Base entity class for parentable entities.
    
    Adds the following keyvalues to entities:
         parentname <target_destination>
             Specifies the targetname of this entity's Parent in a movement
            hierarchy. Child entities mimic the movements of their Parent.
    '''
    parentname = _KeyValue('')
    
    def SetParent(self, target, userid=0):
        '''Change the entity's parent in the movement hierarchy.'''
        if target[0] != '!':
            self.parentname = target
        self.fire('SetParent', target, userid=userid)
    
    def SetParentAttachment(self, attachment, userid=0):
        '''
        Change this entity to attach to a specific attachment point on its
        parent. Entities must be parented before being sent this input.
        The parameter passed in should be the name of the attachment.
        '''
        self.fire('SetParentAttachment', attachment, userid=userid)
    
    def SetParentAttachmentMaintainOffset(self, attachment, userid=0):
        '''
        Change this entity to attach to a specific attachment point on its
        parent. Entities must be parented before being sent this input.
        The parameter passed in should be the name of the attachment.
        The entity will maintain its position relative to the Parent's
        Origin and Angles at the time it is attached.
        '''
        self.fire('SetParentAttachmentMaintainOffset', attachment, userid=userid)
    
    def ClearParent(self, userid=0):
        '''Remove this entity from its current movement hierarchy.'''
        self.fire('ClearParent', userid=userid)


class _BaseBrush(_BaseEntity):
    '''
    Base entity used for brushes in HL1 ports.
    '''
    def SetTextureIndex(self, index, userid=0):
        '''Set the brush texture index.'''
        self.fire('SetTextureIndex', index, userid=userid)
    
    def IncrementTextureIndex(self, userid=0):
        '''Increment the brush texture index.'''
        self.fire('IncrementTextureIndex', userid=userid)


class _EnableDisable(_BaseEntity):
    '''
    Base class for entities that can be enabled/disabled.
    '''
    StartDisabled = _KeyValue(SymbolicBoolean(0))
    
    def Enable(self, userid=0):
        '''Enable this entity.'''
        self.fire('Enable', userid=userid)
    
    def Disable(self, userid=0):
        '''Disable this entity.'''
        self.fire('Disable', userid=userid)


class _PlayerClass(_BaseEntity):
    '''A base class used in Hammer to display a player model in place of box.'''
    pass


class _RenderFxChoices(_BaseEntity):
    '''
    Rendering effects.
    '''

    renderfx_values = {
        'Normal': 0,
        'Slow Pulse': 1,
        'Fast Pulse': 2,
        'Slow Wide Pulse': 3,
        'Fast Wide Pulse': 4,
        'Slow Fade Away': 5,
        'Fast Fade Away': 6,
        'Slow Become Solid': 7,
        'Fast Become Solid': 8,
        'Slow Strobe': 9,
        'Fast Strobe': 10,
        'Faster Strobe': 11,
        'Slow Flicker': 12,
        'Fast Flicker': 13,
        'Constant Glow': 14,
        'Distort': 15,
        'Hologram': 16, # Distort + Fade
        #'Scale up': 17,
        #'Spotlight FX': 22,
        'Cull By Distance': 23, # TEST
        #'Fade Wider Pulse': 24,
        'Spotlight FX': 24, # no more 22?
        'Fade Near': 26,
    }
    
    renderfx = _KeyValue(SymbolicValue('Normal', renderfx_values))


class _RenderFields(_RenderFxChoices):
    '''
    Rendered entities.
    '''
    rendermode_values = {
        'Normal': 0,
        'Color': 1,
        'Texture': 2,
        'Glow': 3,
        'Solid': 4,
        'Additive': 5,
        #
        'Additive Fractional Frame': 7,
        #
        'World Space Glow': 9,
        'Dont Render': 10,
    }
    
    HDRColorScale = _KeyValue(1.0)
    rendermode = _KeyValue(SymbolicValue('Normal', rendermode_values))
    renderamt = _KeyValue(RangedValue(0,0,255))
    rendercolor = _KeyValue(Color(0,0,0))
    disablereceiveshadows = _KeyValue(SymbolicBoolean(0))
    
    def Color(self, color, userid=0):
        '''Set the rendercolor.'''
        self.fire('Color', Color(color).for_input(), userid=userid)
    
    def ColorRedValue(self, red, userid=0):
        '''
        Set the red color channel's value (0 - 255).
        '''
        self.fire('ColorRedValue', red, userid=userid)
    
    def ColorGreenValue(self, green, userid=0):
        '''
        Set the green color channel's value (0 - 255).
        '''
        self.fire('ColorGreenValue', green, userid=userid)
    
    def ColorBlueValue(self, blue, userid=0):
        '''
        Set the blue color channel's value (0 - 255).
        '''
        self.fire('ColorBlueValue', blue, userid=userid)
 
    def Alpha(self, alpha):
        '''Set the renderamt.'''
        self.fire('Alpha', alpha)


class _DXLevelChoice(_BaseEntity):
    '''
    Only include an object in a range of dx levels.
    '''
    dxlevel_values = {
        'default': 0,
        'lowest': 0,
        'default (lowest)': 0,
        'dx7': 70,
        'dx8.0': 80,
        'dx8.0 (4600Ti)': 80,
        'dx8.1': 81,
        'dx8.1 (FX5200)': 81,
        'dx9 SM2.0': 90,
        'dx9 SM3.0': 95,
    }
    
    mindxlevel = _KeyValue(SymbolicValue('default', dxlevel_values))
    maxdxlevel = _KeyValue(SymbolicValue('default', dxlevel_values))


class _Inputfilter(_BaseEntity):
    '''
    Base entity class for input filtering.
    '''
    InputFilter = _KeyValue(SymbolicValue(0, {
        'Allow all inputs': 0,
        'Ignore Touch/Untouch': 8,
        'Ignore Use': 16,
        'Ignore All': 32,
    }))


class _Global(_BaseEntity):
    '''
    Global entity name for level transitions.
    '''
    globalname = _KeyValue('')


class _EnvGlobal(_Targetname):
    '''
    Base class for env_global.
    '''
    initialstate = _KeyValue(SymbolicValue('Off', {
        'Off': 0,
        'On': 1,
        'Dead': 2,
    }))
    counter = _KeyValue(0)
    spawnflags = _KeyValue(Flags({
        'Set Initial State': 1,
    }))
    
    def TurnOn(self, userid=0):
        '''Set state of global to ON.'''
        self.fire('TurnOn', userid=userid)
    
    def TurnOff(self, userid=0):
        '''Set state of global to OFF.'''
        self.fire('TurnOff', userid=userid)
    
    def Toggle(self, userid=0):
        '''Toggle state of global between ON and OFF.'''
        self.fire('Toggle', userid=userid)
    
    def Remove(self, userid=0):
        '''Set state of global to DEAD.'''
        self.fire('Remove', userid=userid)
    
    def SetCounter(self, value, userid=0):
        '''Set the counter value of this global.'''
        self.counter=int(counter)
        self.fire('SetCounter', self.counter, userid=userid)
    
    def AddToCounter(self, value, userid=0):
        '''Add to the counter value. Negative value subtracts.'''
        self.counter+=int(value)
        self.fire('AddToCounter', value, userid=userid)
    
    def GetCounter(Self, userid=0):
        '''Fire Counter output.'''
        self.fire('GetCounter', userid=userid)


class _DamageFilter(_BaseEntity):
    '''
    Base class for entities supporting damage filtering.
    '''
    damagefilter = _KeyValue('')
    
    def SetDamageFilter(self, target, userid=0):
        '''Set the entity to use as damage filter.'''
        self.fire('SetDamageFilter', target, userid=userid)


class _ResponseContext(_BaseEntity):
    '''
    Base class for entities supporting response contexts.
    '''
    ResponseContext = _KeyValue('')
    
    def AddContext(self, context, userid=0):
        '''
        Add a context to this entity's list of response contexts.
        The format should be 'key:value'.
        '''
        self.fire('AddContext', context, userid=userid)
    
    def RemoveContext(self, key, userid=0):
        '''
        Remove a context from this entity's list of response contexts.
        The key should match an existing context.
        '''
        self.fire('RemoveContext', key, userid=userid)
    
    def ClearContext(self, userid=0):
        '''
        Remove all contexts in this entity's list of response contexts.
        '''
        self.fire('ClearContext', userid=userid)


class _Breakable(_Targetname, _DamageFilter, _Shadow):
    '''
    Base class for breakable entities.
    '''
    ExplodeDamage = _KeyValue(0.0)
    ExplodeRadius = _KeyValue(0.0)
    PerformanceMode = _KeyValue(SymbolicValue('Normal', {
        'Normal': 0,
        'No Gibs': 1,
        'Full Gibs on All Platforms': 2,
        'Reduced gibs': 3,
    }))
    
    def Break(self, userid=0):
        '''Break the breakable.'''
        self.fire('Break', userid=userid)
    
    def SetHealth(self, health, userid=0):
        '''Set a new value for the breakable's health.'''
        self.fire('SetHealth', health, userid=userid)
    
    def AddHealth(self, health, userid=0):
        '''Add health to the breakable.'''
        self.fire('AddHealth', health, userid=userid)

    def RemoveHealth(self, health, userid=0):
        '''Remove health from the breakable.'''
        self.fire('RemoveHealth', health, userid=userid)

    def EnablePhyscannonPickup(self, userid=0):
        '''Make the breakable able to picked up by the physcannon.'''
        self.fire('EnablePhyscannonPickup', userid=userid)

    def DisablePhyscannonPickup(self, userid=0):
        '''Make the breakable not able to picked up by the physcannon.'''
        self.fire('DisablePhyscannonPickup', userid=userid)
    
    def SetMass(self, mass, userid=0):
        '''Set mass of this object.'''
        self.fire('SetMass', mass, userid=userid)


class _BreakableBrush(_Breakable, _Parentname, _Global):
    '''
    Breakable brush entity.
    '''
    spawnflags_values = {
        'Only Break on Trigger': 1,
        'Break on Touch': 2,
        'Break on Pressure': 4,
        'Break immediately on Physics': 512,
        "Don't take physics damage": 1024,
        "Don't allow bullet penetration": 2048,
    }

    propdata_values = {
        'None': 0,
        'Wooden.Tiny': 1,
        'Wooden.Small': 2,
        'Wooden.Medium': 3,
        'Wooden.Large': 4,
        'Wooden.Huge': 5,
        'Metal.Small': 6,
        'Metal.Medium': 7,
        'Metal.Large': 8,
        'Cardboard.Small': 9,
        'Cardboard.Medium': 10,
        'Cardboard.Large': 11,
        'Stone.Small': 12,
        'Stone.Medium': 13,
        'Stone.Large': 14,
        'Stone.Huge': 15,
        'Glass.Small': 16,
        'Plastic.Small': 17,
        'Plastic.Medium': 18,
        'Plastic.Large': 19,
        'Pottery.Small': 20,
        'Pottery.Medium': 21,
        'Pottery.Large': 22,
        'Pottery.Huge': 23,
        'Glass.Window': 24,
    }
    
    material_values = {
        'Glass': 0,
        'Wood': 1,
        'Metal': 2,
        'Flesh': 3,
        'CinderBlock': 4,
        'CeilingTile': 5,
        'Computer': 6,
        'Unbreakable Glass': 7,
        'Rocks': 8,
        'Web': 9, # defunct
        'None': 10,
    }
    
    explosion_values = {
        'Random': 0,
        'Relative to Attack': 1,
        'Use Precise Gib Dir': 2,
    }
    
    spawnobject_values = {
        '': 0, # for default value, not listed in fgd
        'item_battery': 1,
        'item_healthkit': 2,
        'item_ammo_pistol': 3,
        'item_ammo_pistol_large': 4,
        'item_ammo_smg1': 5,
        'item_ammo_smg1_large': 6,
        'item_ammo_ar2': 7,
        'item_ammo_ar2_large': 8,
        'item_box_buckshot': 9,
        'item_flare_round': 10,
        'item_box_flare_rounds': 11,
        'item_ml_grenade': 12,
        'item_smg1_grenade': 13,
        'item_box_sniper_rounds': 14,
        'unused1': 15,
        'weapon_stunstick': 16,
        'weapon_ar1': 17,
        'weapon_ar2': 18,
        'unused2': 19,
        'weapon_ml': 20,
        'weapon_smg1': 21,
        'weapon_smg2': 22,
        'weapon_slam': 23,
        'weapon_shotgun': 24,
        'weapon_molotov': 25,
        'item_dynamic_resupply': 26,
    }
    
    spawnflags = _KeyValue(Flags(spawnflags_values, 0))
    propdata = _KeyValue(SymbolicValue('None', propdata_values))
    health = _KeyValue(1)
    material = _KeyValue(SymbolicValue(0, material_values))
    explosion = _KeyValue(SymbolicValue(0, explosion_values))
    gibdir = _KeyValue(Vector(0,0,0))
    nodamageforces = _KeyValue(SymbolicBoolean(0))
    gibmodel = _KeyValue('')
    spawnobject = _KeyValue(SymbolicValue(0, spawnobject_values))
    explodemagnitude = _KeyValue(0)
    pressuredelay = _KeyValue(0.0)
    
    def EnableDamageForces(self, userid=0):
        '''Damaging the entity applies physics forces to it.'''
        self.fire('EnableDamageForces', userid=userid)
    
    def DisableDamageForces(self, userid=0):
        '''Damaging the entity does not apply forces to it.'''
        self.fire('DisableDamageForces', userid=userid)


class _BreakableProp(_Breakable):
    '''
    Base entity class for breakable props.
    '''
    spawnflags = _KeyValue(Flags({
        'Break on Touch': 16,
        'Break on Pressure': 32,
    }))
    pressuredelay = _KeyValue(0.0)


class _BaseNPC(_Targetname, _RenderFields, _DamageFilter, _ResponseContext, _Shadow):
    '''
    Base class for non-playable characters.
    '''
    target = _KeyValue('')
    squadname = _KeyValue('')
    hintgroup = _KeyValue('')
    hintlimiting = _KeyValue(SymbolicBoolean(0))
    spawnflags = _KeyValue(Flags({
        'Wait Till Seen': 1,
        'Gag': 2,
        'Fall to ground': 4,
        'Drop Healthkit': 8,
        'Efficient': 16,
        'Wait For Script': 128,
        'Long Visibility/Shoot': 256,
        'Fade Corpse': 512,
        'Think outside PVS': 1024,
        'Template NPC': 2048,
        'Do Alternate collision for this NPC': 4096,
        "Don't drop weapons": 8192,
        'Ignore player push': 16384,
    }, 'Fall to ground', 'Fade Corpse'))
    #initialidle = _KeyValue('')
    sleepstate = _KeyValue(SymbolicValue(0, {
        'None': 0,
        'Waiting for threat': 1,
        'Waiting for PVS': 2,
        'Waiting for input, ignore PVS': 3,
        'Auto PVS': 4,
        'Auto PVS after PVS': 5,
    }))
    wakeradius = _KeyValue(0.0)
    wakesquad = _KeyValue(SymbolicBoolean(0))
    enemyfilter = _KeyValue('')
    ignoreunseenenemies = _KeyValue(SymbolicBoolean(0))
    physdamagescale = KeyvalueInput(1.0)
    
    def SetRelationship(self, *relationship, **kw):
        '''
        Change this entity's relationship with another entity or class.
        Format: <entityname/classname> <D_HT/D_FR/D_LI/D_NU> <priority>
        '''
        userid = kw.get('userid', 0)
        if len(relationship) == 1:
            self.fire('SetRelationship', relationship, userid=userid)
        elif len(relationship) == 3:
            self.fire('SetRelationship', ' '.join(relationship), userid=userid)
        else:
            raise TypeError('Needs correct relationship parameter')
    
    def SetEnemyFilter(self, target, userid=0):
        '''Change this NPC's enemy filter to the named filter.'''
        self.fire('SetEnemyFilter', target, userid=userid)
    
    def SetHealth(self, health, userid=0):
        '''Set this NPC's health.'''
        self.fire('SetHealth', health, userid=userid)
    
    def SetBodyGroup(self, group, userid=0):
        '''HACK: Set this NPC's body group (0-n).'''
        self.fire('SetBodyGroup', group, userid=userid)
    
    def Ignite(self, userid=0):
        '''Ignite, burst into flames.'''
        self.fire('Ignite', userid=userid)
    
    def IgniteLifetime(self, lifetime, userid=0):
        '''Ignite for specific time.'''
        self.fire('IgniteLifetime', lifetime, userid=userid)
    
    def IgniteNumHitboxFires(self, fires, userid=0):
        '''Ignite with specific number of hitbox fires.'''
        self.fire('IgniteNumHitboxFires', fires, userid=userid)
    
    def IgniteHitboxFireScale(self, scale, userid=0):
        '''Ignite with hitbox fire scale.'''
        self.fire('IgniteHitboxFireScale', scale, userid=userid)
    
    def Break(self, userid=0):
        '''Break, smash into pieces.'''
        self.fire('Break', userid=userid)
    
    def StartScripting(self, userid=0):
        '''Enter scripting state.'''
        self.fire('StartScripting', userid=userid)
    
    def StopScripting(self, userid=0):
        '''Exit scripting state.'''
        self.fire('StopScripting', userid=userid)
        
    def SetSquad(self, squad, userid=0):
        '''Set the name of this NPC's squad.'''
        self.fire('SetSquad', squad, userid=userid)
    
    def Wake(self, userid=0):
        '''Wake up the NPC if it is sleeping.'''
        self.fire('Wake', userid=userid)
    
    def ForgetEntity(self, entity, userid=0):
        '''Clear out the NPC's knowledge of a named entity.'''
        self.fire('ForgetEntity', entity, userid=userid)
    
    def GagEnable(self, userid=0):
        '''Turn on the NPC Gag flag.'''
        self.fire('GagEnable', userid=userid)
    
    def GagDisable(self, userid=0):
        '''Turn off the NPC Gag flag.'''
        self.fire('GagDisable', userid=userid)
    
    def IgnoreDangerSounds(self, seconds, userid=0):
        '''Ignire danger sounds for the specified time.'''
        self.fire('IgnoreDangerSounds', seconds, userid=userid)
    
    def HolsterWeapon(self, userid=0):
        '''Force the NPC to holster their weapon.'''
        self.fire('HolsterWeapon', userid=userid)
    
    def HolsterAndDestroyWeapon(self, userid=0):
        '''As HolsterWeapone except that the weapon is destroyed.'''
        self.fire('HolsterAndDestroyWeapon', userid=userid)
    
    def UnholsterWeapon(self, userid=0):
        '''Force the NPC to draw their weapon.'''
        self.fire('UnholsterWeapon', userid=userid)
    
    def ForceInteractionWithNPC(self, npc, interaction, userid=0):
        '''Force the NPC to use dynamic interaction with another NPC.'''
        self.fire('ForceInteractionWithNPC', "%s %s"%(npc,interaction), userid=userid)
    
    def UpdateEnemyMemory(self, target, userid=0):
        '''Update or create this NPC's memory of an enemy.'''
        self.fire('UpdateEnemyMemory', target, userid=userid)
    
    def BecomeRagdoll(self, userid=0):
        '''This NPC will instantly become a ragdoll with ZERO force.'''
        self.fire('BecomeRagdoll', userid=userid)


class _BaseNPCMaker(_Targetname, _EnableDisable):
    '''
    Base class for npc_maker and npc_template_maker.
    '''
    StartDisabled = _KeyValue(SymbolicBoolean(1)) #override default value
    spawnflags = _KeyValue(Flags({
        'Fade Corpse': 16,
        'Infinite Children': 32,
        'Do Not Drop': 64,
        "Don't Spawn While Visible": 128,
    }))
    MaxNPCCount = _KeyValue(1)
    SpawnFrequency = _KeyValue(5.0) # should be float, why is it string, Valve?!?
    MaxLiveChildren = _KeyValue(5)
    
    def Spawn(self, userid=0):
        '''Spawn an NPC.'''
        self.fire('Spawn', userid=userid)
    
    def Toggle(self, userid=0):
        '''Toggle the spawner enabled/disabled state.'''
        self.fire('ToggleSpawn', userid=userid)

    def AddMaxChildren(self, value, userid=0):
        '''Add to the number of NPCs that can be spawned.'''
        self.fire('AddMaxChildren', value, userid=userid)

    def SetMaxChildren(self, value, userid=0):
        '''Set the number of NPCs that can be spawned.'''
        self.fire('SetMaxChildren', value, userid=userid)

    def SetMaxLiveChildren(self, value, userid=0):
        '''Set the number of NPCs that can be alive at the same time.'''
        self.fire('SetMaxLiveChildren', value, userid=userid)
    
    def SetSpawnFrequency(self, freq, userid=0):
        '''Set how often a new NPC will be spawned.'''
        self.fire('SetSpawnFrequency', freq, userid=userid)


class _BaseHelicopter(_BaseNPC):
    '''
    Base class for helicopters.
    '''
    InitialSpeed = _KeyValue(0.0) # why string?
    target = _KeyValue('')
    spawnflags = _KeyValue(Flags({
        'No Rotorwash': 32,
        'Await Input': 64,
    }))
    
    def MoveTopSpeed(self, userid=0):
        '''
        The helicopter will immediately move at top speed toward its current
        goal, or in its current orientation if it's on top of its goal.
        '''
        self.fire('MoveTopSpeed', userid=userid)
    
    def MoveSpecifiedSpeed(self, speed, userid=0):
        '''
        The helicopter will immediately move at specified speed (units/sec).
        '''
        self.fire('MoveSpecifiedSpeed', speed, userid=userid)
    
    def ChangePathCorner(self, target, userid=0):
        '''Tell the helicopter to move to a path corner on a new path.'''
        self.fire('ChangePathCorner', target, userid=userid)
    
    def SelfDestruct(self, userid=0):
        '''Self Destruct.'''
        self.fire('SelfDestruct', userid=userid)
    
    def Activate(self, userid=0):
        '''
        Activate. Use to wake up a helicopter that spawned with the
        'Await Input' spawnflag on.
        '''
        self.fire('Activate', userid=userid)
    
    def SetTrack(self, target, userid=0):
        '''
        Set a track for the helicopter to adhere to. The helicopter will do
        nothing if he's on the same path, and will move to the closest point
        on the specified track if he's on a different path.
        '''
        self.fire('SetTrack', self.target, userid=userid)
    
    def FlyToSpecificTrackViaPath(self, target, userid=0):
        '''
        The helicopter will first fly to the closest point on the path if
        he's on a different path. Then he'll fly along the path to the
        specified track point.
        '''
        self.fire('FlyToSpecificTrackViaPath', target, userid=userid)
    
    def StartPatrol(self, userid=0):
        '''Start patrolling back and forth along the current track.'''
        self.fire('StartPatrol', userid=userid)
    
    def StopPatrol(self, userid=0):
        '''
        Stop patrolling back and forth along the track.
        This will cause the helicopter to come to rest at the track which
        he's currently flying toward.
        '''
        self.fire('StopPatrol', userid=userid)
    
    def ChooseFarthestPathPoint(self, userid=0):
        '''
        When tracking an enemy, choose the point on the path furthest from the
        enemy, but still in firing range.
        '''
        self.fire('ChooseFarthestPathPoint', userid=userid)
    
    def ChooseNearestPathPoint(self, userid=0):
        '''
        When tracking an enemy, choose the point on the path nearest from the
        enemy.
        '''
        self.fire('ChooseNearestPathPoint', userid=userid)
    
    def StartBreakableMovement(self, userid=0):
        '''
        The helicopter is now allowed to disobey direct commands to go to
        particular points if he senses an enemy. He will move to the closest
        point (or farthest point, if ChooseFarthestPathPoint is used), on the
        path if he senses an enemy.
        '''
        self.fire('StartBreakableMovement', userid=userid)
    
    def StopBreakableMovement(self, userid=0):
        '''
        The helicopter can not disobey direct commands.
        He will continue to fly along his patrol path or to his specified
        target even if he senses an enemy.
        '''
        self.fire('StopBreakableMovement', userid=userid)


class _BaseSpeaker(_Targetname, _ResponseContext):
    '''
    Base class for speaker entities.
    '''
    delaymin = _KeyValue('15') # Min Delay Between Announcements
    delaymax = _KeyValue('135') # Max Delay Between Announcements
    spawnflags = _KeyValue(Flags({
        'Start Silent': 1,
        'Play Everywhere': 2,
    }))
    rulescript = _KeyValue('') # Script file containing rules for playing appropriate sounds.
    concept = _KeyValue('') # High level concept name used as primary search key.
    
    def TurnOn(self, userid=0):
        '''
        Turn on the random announcements.
        '''
        self.fire('TurnOn', userid=userid)
    
    def TurnOff(self, userid=0):
        '''
        Turn off the random announcements.
        '''
        self.fire('TurnOff', userid=userid)
    
    def Toggle(self, userid=0):
        '''
        Toggle the random announcements off and on.
        '''
        self.fire('Toggle', userid=userid)


class _Light(_BaseEntity):
    '''
    Base class for light entities.
    '''
    style_values = {
        'Normal': 0,
        'Flicker A': 1,
        'Slow, strong pulse': 2,
        'Candle A': 3,
        'Fast strobe': 4,
        'Gentle pulse': 5,
        'Flicker B': 6,
        'Candle B': 7,
        'Candle C': 8,
        'Slow Strobe': 9,
        'Fluorescent flicker': 10,
        'Slow pulse, noblack': 11,
    }

    _light = _KeyValue(Color(255,255,255,200))
    # _lightHDR = _KeyValue(Color(-1,-1,-1,1)) # will turn to 255,255,255,1
    # removed, engine seems to return empty string hinting that this does
    # not exist actually
    _lightscaleHDR = _KeyValue(1.0)
    style = _KeyValue(SymbolicValue('Normal', style_values))
    pattern = _KeyValue('')
    _constant_attn = _KeyValue('0')
    _linear_attn = _KeyValue('0')
    _quadratic_attn = _KeyValue('1')
    _fifty_percent_distance = _KeyValue('0')
    _zero_percent_distance = _KeyValue('0')
    _hardfalloff = _KeyValue(0)
    
    def TurnOn(self, userid=0):
        '''Turn the light on.'''
        self.fire('TurnOn', userid=userid)
    
    def TurnOff(self, userid=0):
        '''Turn the light off.'''
        self.fire('TurnOff', userid=userid)
    
    def Toggle(self, userid=0):
        '''Toggle the ligiht's current state.'''
        self.fire('Toggle', userid=userid)

    def SetPattern(self, pattern, userid=0):
        '''
        Set a custom pattern of light brightness for this light.
        Pattern format is a string of characters, where
        'a' is total darkness, 'z' fully bright. i.e. 'aaggnnttzz'
        would be a steppy fade in from dark to light.
        '''
        self.fire('SetPattern', pattern, userid=userid)

    def FadeToPattern(self, pattern, userid=0):
        '''
        Fade from first value in old pattern, to first value in the new
        given pattern. Pattern format is a string of characters, where
        'a' is total darkness, 'z' fully bright. i.e. 'aaggnnttzz'
        would be a steppy fade in from dark to light.
        '''
        self.fire('FadeToPattern', pattern, userid=userid)


class _Node(_BaseEntity):
    '''
    Base class for node entities.

    Adds read only keyvalue nodeid.
    '''
    nodeid = _KeyValue(-1, readonly=True)


class _HintNode(_Node):
    '''
    Base class for hint nodes.
    '''
    hinttype_values = {
        'None': 0,
        'World: Window': 2,
        'World: Act Busy Hint': 12,
        'World: Visually Interesting': 13,
        "World: Visually Interesting (Don't aim at)": 14,
        'World: Inhibit Combine Mines within 15 feet': 15,
        'World: Visually Interesting (Stealth mode)': 16,
        'Crouch Cover Medium': 100,
        'Crouch Cover Low': 101,
        'Waste Scanner Spawn': 102,
        'Entrance / Exit Pinch': 103,
        'Guard Point': 104, # not part of base hint node?
        'Enemy Disadvantage Point': 105,
        'Health Kit': 106,
        'Antlion: Burrow Point': 400,
        'Antlion; Thumper Flee Point': 401,
        'Headcrab: Burrow Point': 450,
        'Headcrab: Exit Pod Point': 451,
        'Roller: Patrol Point': 500,
        'Roller: Cleanup Spot': 501,
        'Crow: Fly to point': 700,
        'Crow: Perch point': 701,
        'Follower: Wait point': 900,
        'Override jump permission': 901,
        'Player squad transition point': 902,
        'NPC exit point': 903,
        'Strider node': 904,
        'Player Ally: Push away destination': 950,
        'Player Ally: Fear withdrawal destination': 951,
        'HL1 World: Machinery': 1000,
        'HL1 World: Blinking Light': 1001,
        'HL1 World: Human Blood': 1002,
        'HL1 World: Alien Blood': 1003,
        'CS Hostage: Escape Point': 1100, # not part of base hint node?
    }
    
    state_values = {
        'Idle': 1,
        'Alert': 2,
        'Combat': 3,
    }
    
    spawnflags = _KeyValue(Flags({
        'Allow jump up': 65536,
    }))
    hinttype = _KeyValue(SymbolicValue('None', hinttype_values))
    hintactivity = _KeyValue('')
    nodeFOV = _KeyValue(SymbolicValue(180, {
        '45 Degrees': 45,
        '90 Degrees': 90,
        '180 Degrees': 180,
        '260 Degrees': 360,
    }))
    StartHintDisabled = _KeyValue(SymbolicBoolean(0))
    Group = _KeyValue('')
    TargetNode = _KeyValue(-1)
    IgnoreFacing = _KeyValue(SymbolicValue('Default', {
        'No': 0,
        'Yes': 1,
        'Default': 2,
    }))
    MinimumState = _KeyValue(SymbolicValue('Idle', state_values))
    MaximumState = _KeyValue(SymbolicValue('Combat', state_values))
    
    def EnableHint(self, userid=0):
        '''Enable hint.'''
        self.fire('EnableHint', userid=userid)
    
    def DisableHint(self, userid=0):
        '''Disable hint.'''
        self.fire('DisableHint', userid=userid)


class _TriggerOnce(_Targetname, _Parentname, _EnableDisable, _Global):
    '''
    Base class for trigger_once entity and other triggers.
    '''
    spawnflags = _KeyValue(Flags({
        'Clients': 1,
        'NPCs': 2,
        'Pushables': 4,
        'Physics Objects': 8,
        'Only player ally NPCs': 16,
        'Only clients in vehicles': 32,
        'Everything': 64,
        'Only clients not in vehicles': 512,
        'Physics debris': 1024,
        'Only NPCs in vehicles': 2048,
    }, 'Clients'))
    filtername = _KeyValue('')
    
    def Toggle(self, userid=0):
        '''Toggle this trigger between enabled and disabled states.'''
        self.fire('Toggle', userid=userid)


class _Trigger(_TriggerOnce):
    '''
    Base class for triggers.
    '''
    pass


class _KeyFrame(_BaseEntity):
    '''
    KeyFrame base class.
    '''
    NextKey = _KeyValue('') # Name of the next keyframe along this keyframe path.
    MoveSpeed = _KeyValue(64) # Speed (units per second)
    #!FGD NextTime = _KeyValue(0), # Time to get to next keyframe
    #!FGD TimeModifier = _KeyValue(SymbolicValue(0, {
    #    'Linear': 0,
    #    'Accel': 1,
    #    'Deaccel': 2,
    #    'Accel/Deaccel (sine)': 2,
    #}))


class _Mover(_BaseEntity):
    '''
    Mover entity base class.
    '''
    PositionInterpolator = _KeyValue(SymbolicValue(0, {
        'Linear': 0,
        'Catmull-Rom Spline': 1,
    })) # Position Interpolator


class _RopeKeyFrame(_DXLevelChoice):
    '''
    Base class for keyframe ropes
    '''
    spawnflags = _KeyValue(Flags({
        'Auto Resize': 1,
    }))
    Slack = _KeyValue(25) # How much extra length the rope has
        # (by default it has the length between its two endpoints in the editor).
    Type = _KeyValue(SymbolicValue(0, {
        'Rope': 0,
        'Semi-rigid': 1,
        'Rigid': 2,
    })) # Type
    Subdiv = _KeyValue(2) # Number of subdivisions between each rope segment.
        # Maximum value is 8. Higher values make smoother ropes, but are slower to render.
    Barbed = _KeyValue(SymbolicValue(0, {
        'No': 0,
        'Yes': 1,
    })) # Test effect that makes the rope look sharper and more barbed.
    Width = _KeyValue('2') # Width of the rope.
    TextureScale = _KeyValue('1') # This changes the texture resolution.
        # The default resolution is 4 pixels per inch.
        # Larger values stretch the texture and smaller values scrunch it up.
    Collide = _KeyValue(SymbolicValue(0, {
        'No': 0,
        'Yes': 1,
    })) # Collide with world
    Dangling = _KeyValue(SymbolicValue(0, {
        'No': 0,
        'Yes': 1,
    })) # When set to Yes, the rope starts out detached from its target endpoint.
    Breakable = _KeyValue(SymbolicValue(0, {
        'No': 0,
        'Yes': 1,
    })) # When set to yes, the rope can be detached from either endpoint when shot.
    RopeMaterial = _KeyValue('cable/cable.vmt') # The material to use when rendering the rope.
    NoWind = _KeyValue(SymbolicValue(0, {
        'No': 0,
        'Yes': 1,
    })) # When set to Yes, the rope will no longer act as though it's being affected by wind.
    
    def SetScrollSpeed(self, rate, userid=0):
        '''
        Set the speed at which the texture scrolls.
        '''
        self.fire('SetScrollSpeed', rate, userid=userid)
    
    def SetForce(self, vector, userid=0):
        '''
        Apply a force instantaneously to the rope.
        The parameter should be a vector containing the force to be applied (X Y Z).
        '''
        v = Vector(vector)
        self.fire('SetForce', v.getstr(' '), userid=userid)
    
    def Break(self, userid=0):
        '''
        Break the rope, if it's marked to do so.
        '''
        self.fire('Break', userid=userid)


class _Button(_BaseEntity):
    '''
    Base class for button entities.
    '''
    
    def Lock(self, userid=0):
        '''
        Lock the button, preventing it from functioning.
        '''
        self.fire('Lock', userid=userid)
    
    def Unlock(self, userid=0):
        '''
        Unlock the button, allowing it to function.
        '''
        self.fire('Unlock', userid=userid)
    
    def Press(self, userid=0):
        '''
        Activate the button as if it was pressed.
        '''
        self.fire('Press', userid=userid)
    
    def PressIn(self, userid=0):
        '''
        Activate the button as if it was pressed, sending it to the bottom position.
        '''
        self.fire('PressIn', userid=userid)
    
    def PressOut(self, userid=0):
        '''
        Unpress the button, sending it to the top position.
        '''
        self.fire('PressOut', userid=userid)


class _Door(_Targetname, _Parentname, _RenderFields, _Global, _Shadow):
    '''
    Base class for doors.
    '''
    speed = _KeyValue(100) # The speed at which the door moves.
    master = _KeyValue('') # Legacy support
    noise1 = _KeyValue('') # Sound to play when the door starts moving.
    noise2 = _KeyValue('') # Sound to play when the door stops moving.
    startclosesound = _KeyValue('') # (Optional) Sound to play when the door starts closing.
    closesound = _KeyValue('') # (Optional) Sound to play when the door stops closing.
    wait = _KeyValue(4) # Amount of time, in seconds, after the door has opened before it closes.
        # Once it has closed, it can be used again.
        # If the value is set to -1, the door never closes itself.
    lip = _KeyValue(0) # The amount, in inches, of the button to leave sticking out of the
    # wall it recedes into when pressed.
    # Negative values make the button recede even further into the wall.
    dmg = _KeyValue(0) # Amount of damage done to entities that block
        # the movement of this door, per frame.
    forceclosed = _KeyValue(SymbolicValue(0, {
        'No': 0,
        'Yes': 1,
    })) # If set, this door will close no matter what.
        # Useful for doors that have to close even if the player tries
        # to block them with objects.
    ignoredebris = _KeyValue(SymbolicValue(0, {
        'No': 0,
        'Yes': 1,
    })) # If set this will change the door's collision group to one that ignore
        # collisions with debris objects (note that this is not compatible with
        # the non-solid-to-player spawnflag).
    message = _KeyValue('') # Message If Triggered
    health = _KeyValue(0) # Health (shoot open)
    locked_sound = _KeyValue('') # Sound played when the player tries to use the door)
        # and fails because it's locked.
    unlocked_sound = _KeyValue('') # Sound played when the button is door. (LOLWUT?)
    spawnpos = _KeyValue(SymbolicValue(0, {
        'Closed': 0,
        'Open': 1,
    })) # Spawn Position
    spawnflags = _KeyValue(Flags({
        "Starts Open - OBSOLETE, use 'Spawn Position' key instead": 1,
        'Non-solid to Player': 4,
        'Passable': 8,
        'Toggle': 32,
        'Use Opens': 256,
        "NPCs Can't": 512,
        'Touch Opens': 1024,
        'Starts locked': 2048,
        'Door Silent': 4096,
    }, 'Touch Opens'))
    locked_sentence = _KeyValue(SymbolicValue(0, {
        'None': 0,
        'Gen. Access Denied': 1,
        'Security Lockout': 2,
        'Blast Door': 3,
        'Fire Door': 4,
        'Chemical Door': 5,
        'Radiation Door': 6,
        'Gen. Containment': 7,
        'Maintenance Door': 8,
        'Broken Shut Door': 9,
    })) # A sentence played when the player tries to use the door,
        # and fails because it's locked.
    unlocked_sentence = _KeyValue(SymbolicValue(0, {
        'None': 0,
        'Gen. Access Granted': 1,
        'Security Disengaged': 2,
        'Blast Door': 3,
        'Fire Door': 4,
        'Chemical Door': 5,
        'Radiation Door': 6,
        'Gen. Containment': 7,
        'Maintenance area': 8,
    })) # A sentence played when the door is unlocked.
    _minlight = _KeyValue('') # The minimum level of ambient light that hits this brush.
    loopmovesound = _KeyValue(SymbolicValue(0, {
        'No': 0,
        'Yes': 1,
    })) # If set to true, the door's 'Start Sound' will be continually
        # looped until the door finishes moving.
    
    def Open(self, userid=0):
        '''
        Open the door, if it is not fully open.
        '''
        self.fire('Open', userid=userid)
    
    def Close(self, userid=0):
        '''
        Close the door, if it is not fully closed.
        '''
        self.fire('Close', userid=userid)
    
    def Toggle(self, userid=0):
        '''
        Toggle the door between open and closed.
        '''
        self.fire('Toggle', userid=userid)
    
    def Lock(self, userid=0):
        '''
        Lock the door.
        '''
        self.fire('Lock', userid=userid)
    
    def Unlock(self, userid=0):
        '''
        Unlock the door.
        '''
        self.fire('Unlock', userid=userid)
    
    def SetSpeed(self, param, userid=0):
        '''
        Set the door speed.
        '''
        self.fire('SetSpeed', param, userid=userid)


class _BModelParticleSpawner(_BaseEntity):
    '''
    Base class for volumetric particle spawning entities.
    '''
    StartDisabled = _KeyValue(SymbolicValue(0, {
        'No': 0,
        'Yes': 1,
    })) # Start Disabled
    Color = _KeyValue(Color(255,255,255)) # Particle Color (R G B)
    SpawnRate = _KeyValue(40) # Number of particles to spawn, per second.
    SpeedMax = _KeyValue(13) # Maximum speed that the particles can move after spawning.
    LifetimeMin = _KeyValue(3) # Minimum number of seconds until each particle dies.
        # Particles live for a random duration between this and LifetimeMax.
    LifetimeMax = _KeyValue(5) # Maximum number of seconds until each particle dies.
        # Particles live for a random duration between LifetimeMin and this.
    DistMax = _KeyValue(1024) # Maximum distance at which particles are visible.
        # They fade to translucent at this distance.
    Frozen = _KeyValue(SymbolicValue(0, {
        'No': 0,
        'Yes': 1,
    })) # When set, this entity spawns the number of particles in SpawnRate immediately,
        # and then goes inactive.
    
    def TurnOn(self, userid=0):
        '''
        Turn on.
        '''
        self.fire('TurnOn', userid=userid)
    
    def TurnOff(self, userid=0):
        '''
        Turn off.
        '''
        self.fire('TurnOff', userid=userid)


class _gibshooterbase(_Targetname, _Parentname):
    '''
    Base class for gibshooter.
    '''
    m_iGibs = _KeyValue(3)
    delay = _KeyValue(0.0) # String? No way. Delay between shots. 0 = all at once
    gibangles = _KeyValue(Vector((0,0,0)))
    gibanglevelocity = _KeyValue(0.0) # not a string >.<
    m_flVelocity = _KeyValue(200.0) # not int
    m_flVariance = _KeyValue(0.15) # not string
    m_flGibLife = _KeyValue(4.0) # not string
    lightingorigin = _KeyValue('') # targetname
    spawnflags = _KeyValue(Flags({
        'Repeatable': 0,
    }))
    
    def Shoot(self, userid=0):
        '''Force the gibshooter to create and shoot a gib.'''
        self.fire('Shoot', userid=userid)
    

class _vgui_screen_base(_Targetname, _Parentname):
    '''
    Vgui screen in 3D.
    '''
    panelname = _KeyValue('')
    overlaymaterial = _KeyValue('')
    width = _KeyValue(32)
    height = _KeyValue(32)
    
    def SetActive(self, userid=0):
        '''Make the vgui screen visible.'''
        self.fire('SetActive', userid=userid)
    
    def SetInactive(self, userid=0):
        '''Make the vgui screen invisible.'''
        self.fire('SetInactive', userid=userid)


class _worldbase(_BaseEntity):
    '''
    Base class for the world.
    '''
    message = _KeyValue('')
    skyname = _KeyValue('sky_day01_01')
    chaptertitle = _KeyValue('')
    startdark = _KeyValue(SymbolicBoolean(0))
    gametitle = _KeyValue(SymbolicBoolean(0))
    newunit = _KeyValue(SymbolicBoolean(0))
    maxoccludeearea = _KeyValue(0.0)
    minoccluderarea = _KeyValue(0.0)
    maxoccludeearea_x360 = _KeyValue(0.0)
    minoccluderarea_x360 = _KeyValue(0.0)
    maxpropscreenwidth = _KeyValue(-1.0)
    minpropscreenwidth = _KeyValue(0.0)
    detailvbsp = _KeyValue('detail.vbsp')
    detailmaterial = _KeyValue('detail/detailsprites')
    coldworld = _KeyValue(SymbolicBoolean(0))


class _BaseFilter(_Targetname):
    '''
    Base class for Activator Filters.
    '''
    Negated = _KeyValue(SymbolicValue('Allow entities that match criteria', {
        'Allow entities that match criteria': 0,
        'Disallow entities that match criteria': 1,
    })) # If set to Allow, only entities who match the criteria will pass the filter.
        # If set to Disallow, only entities who do NOT match the criteria will
        # pass the filter.
    
    def TestActivator(self, userid=0):
        '''
        Test the activator against the filter and fires OnPass or OnFail output.
        '''
        self.fire('TestActivator', userid=userid)


class _TwoObjectPhysics(_Targetname):
    '''
    Base class for two object physics constraint entities.
    '''
    spawnflags = _KeyValue(Flags({
        'No Collision until break': 1,
        'Start inactive': 4,
        'Change mass to keep stable attachment to world': 8,
        'Do not connect entities until turned on': 16,
    }))
    attach1 = _KeyValue('') # Entity 1
    attach2 = _KeyValue('') # Entity 2
    constraintsystem = _KeyValue('') # The name of a phys_constraintsystem
        # that this constraint should be a part of.
        # All constraints on a set of entities should be placed in
        # the same system, or they will fight each other during simulation.
    forcelimit = _KeyValue(0.0) # The amount of force an impact must apply to the
        # constraint to break it. A way of calculating this is to set it
        # to the mass of an object that would break this constraint if
        # it were resting on the constrainted objects.
    torquelimit = _KeyValue(0.0) # The amount of torque required to break the constraint.
        # A way of calculating this is to multiply any reference mass by the
        # resting distance (from the center of mass of the object) needed
        # to break the constraint.
    breaksound = _KeyValue('') # A sound played when the constraint is broken.
    teleportfollowdistance = _KeyValue(0.0) # If one object teleports more than this
        # many units away it will cause the other constrained object to
        # teleport to an appropriate relative position.
    
    def Break(self, userid=0):
        '''
        Force the constraint to break.
        '''
        self.fire('Break', userid=userid)
    
    def TurnOn(self, userid=0):
        '''
        Enable the constraint.
        
        Do this when the objects don't exist when
        the constraint spawns - or when you have deactivated the constraint.
        Broken constraints can NOT be turned on. They have been deleted.
        '''
        self.fire('TurnOn', userid=userid)
    
    def TurnOff(self, userid=0):
        '''
        Disable this constraint.
        '''
        self.fire('TurnOff', userid=userid)


class _ForceController(_Targetname):
    '''
    Base class for force controller entities
    '''
    spawnflags = _KeyValue(Flags({
        'Start On': 1,
        'Apply Force': 2,
        'Apply Torque': 4,
        'Orient Locally': 8,
        'Ignore Mass': 16,
    }, 'Apply Force', 'Apply Torque', 'Orient Locally'))
    attach1 = _KeyValue('') # Object to apply the force to.
    forcetime = _KeyValue('0') # Automatic shut-off after this time has passed
        # (0 = stay on forever or until deactivated)
    
    def Activate(self, userid=0):
        '''
        Turn the force on
        '''
        self.fire('Activate', userid=userid)
    
    def Deactivate(self, userid=0):
        '''
        Turn the force off
        '''
        self.fire('Deactivate', userid=userid)
    
    def Scale(self, scale, userid=0):
        '''
        Set Force Scale
        '''
        self.fire('Scale', scale, userid=userid)


class _prop_detail_base(_BaseEntity):
    '''
    Base class for prop_detail.
    '''
    model = _KeyValue('') # World model


class _BaseFadeProp(_BaseEntity):
    '''
    Base class for prop entites fade values.
    '''
    fademindist = _KeyValue(-1.0) # Distance at which the prop starts to fade
        # (<0 = use fademaxdist). If 'Screen Space Fade' is selected,
        # this represents the number of pixels wide covered by the prop
        # when it starts to fade.
    fademaxdist = _KeyValue(0.0) # Maximum distance at which the prop is visible
        # (0 = don't fade out). If 'Screen Space Fade' is selected, this
        # represents the *minimum* number of pixels wide covered by the
        # prop when it fades.
    fadescale = _KeyValue(1.0) # If you specify a fade in the worldspawn, or if
        # the engine is running under dx7 [hl2/ep1/portal] or dx8 [ep2/tf],
        # then the engine will forcibly fade out props even if
        # fademindist/fademaxdist isn't specified.


class _prop_static_base(_BaseFadeProp, _DXLevelChoice):
    '''
    Base class for prop_static.
    '''
    model = _KeyValue('') # World Model
    skin = _KeyValue(0) # Some models have multiple versions of their textures
        # called skins. Set this to a number other than 0 to use that
        # skin instead of the default.
    solid = _KeyValue(SymbolicValue(6, {
        'Not Solid': 0,
        'Use Bounding Box': 2,
        'Use VPhysics': 6,
    })) # Collisions
    disableshadows = _KeyValue(SymbolicValue(0, {
        'No': 0,
        'Yes': 1,
    })) # Disable Shadows
    screenspacefade = _KeyValue(SymbolicValue(0, {
        'No': 0,
        'Yes': 1,
    })) # The method by which the fading distance should be determined.
        # If 'No', the fade distances is the distance from the player's
        # view to the object, in inches. If 'Yes', the fade distance is
        # the size of the object onscreen, in pixels.
    lightingorigin = _KeyValue('') # Select an info_lighting to specify a location
        # to sample lighting from, instead of using this entity's origin.
    disablevertexlighting = _KeyValue(SymbolicValue(0, {
        'No': 0,
        'Yes': 1,
    })) # Disable Vertex lighting
    disableselfshadowing = _KeyValue(SymbolicValue(0, {
        'No': 0,
        'Yes': 1,
    })) # Disable Self-Shadowing with vertex lighting
    ignorenormals = _KeyValue(SymbolicValue(0, {
        'No': 0,
        'Yes': 1,
    })) # Ignore surface normal for computing vertex lighting


class _prop_dynamic_base(
    _Parentname, _Global, _Studiomodel, _BreakableProp, _DXLevelChoice,
    _BaseFadeProp, _RenderFields):
    '''
    Base class for prop_dynamic.
    '''
    solid = _KeyValue(SymbolicValue(6, {
        'Not Solid': 0,
        'Use Bounding Box': 2,
        'Use VPhysics': 6,
    })) # Collisions
    spawnflags = _KeyValue(Flags({
        'Use Hitboxes for Renderbox': 64,
        'Start with collision disabled': 256,
    }))
    DefaultAnim = _KeyValue('') # The name of the idle animation that this prop
        # will revert to whenever it finishes a random or forced animation.
    RandomAnimation = _KeyValue(SymbolicValue(0, {
        'No': 0,
        'Yes': 1,
    })) # If set, this prop will randomly choose and play animations,
        # based upon the times specified in 'Min/Max Random Anim Time'.
        # Inbetween the random animations, it will revert to playing
        # the 'Default Animation'.
    MinAnimTime = _KeyValue(5.0) # Minimum time between random animations.
    MaxAnimTime = _KeyValue(10.0) # Maximum time between random animations.
    SetBodyGroup = KeyvalueInput(0) # Body Group
    lightingorigin = _KeyValue('') # Select a path_corner entity to specify a
        # location to sample lighting from, instead of using this entity's origin.
    
    def SetAnimation(self, name_of_animation, userid=0):
        '''
        Force the prop to play an animation.
        The parameter should be the name of the animation.
        '''
        self.fire('SetAnimation', name_of_animation, userid=userid)
    
    def SetDefaultAnimation(self, name_of_animation, userid=0):
        '''
        Set the Default Animation to the one specified in the parameter.
        '''
        self.fire('SetDefaultAnimation', name_of_animation, userid=userid)
    
    def SetPlaybackRate(self, rate, userid=0):
        '''
        Set the playback rate for the animation.
        '''
        self.fire('SetPlaybackRate', rate, userid=userid)
    
    def TurnOn(self, userid=0):
        '''
        Make the prop visible.
        '''
        self.fire('TurnOn', userid=userid)
    
    def TurnOff(self, userid=0):
        '''
        Make the prop invisible.
        '''
        self.fire('TurnOff', userid=userid)
    
    def EnableCollision(self, userid=0):
        '''
        Enable collision on the prop.
        '''
        self.fire('EnableCollision', userid=userid)
    
    def DisableCollision(self, userid=0):
        '''
        Disable collision on the prop.
        '''
        self.fire('DisableCollision', userid=userid)


class _BasePropPhysics(
    _Global, _Studiomodel, _BreakableProp, _DXLevelChoice,
    _BaseFadeProp):
    '''
    Base class for prop_physics entities.
    '''
    spawnflags = _KeyValue(Flags({
        'Start Asleep': 1,
        "Don't take physics damage": 2,
        "Debris - Don't collide with the player or other debris": 4,
        'Motion Disabled': 8,
        'Enable motion on Physcannon grab': 64,
        'Not affected by rotor wash': 128,
        'Generate output on +USE ': 256,
        'Prevent pickup': 512,
        'Prevent motion enable on player bump': 1024,
        'Debris with trigger interaction': 4096,
        'Force server-side (Multiplayer only)': 8192,
        'Physgun can ALWAYS pick up. No matter what.': 1048576,
    }, 'Generate output on +USE '))
    minhealthdmg = _KeyValue(0) # The prop will ignore any damage events if
        # the damage is less than this amount.
    shadowcastdist = _KeyValue(0.0) # Use this to override how far this object
        # casts shadows. 0 = default distance.
    physdamagescale = _KeyValue(0.1) # Scales damage energy when this object is
        # hit by a physics object. Note: 0 means this feature is disabled
        # for backwards compatibility.
        # Set to 1.0 for materials as strong as flesh,
        # smaller numbers indicate stronger materials.
    Damagetype = _KeyValue(SymbolicValue(0, {
        'Blunt': 0,
        'Sharp': 1,
    })) # Impact damage type
    nodamageforces = _KeyValue(SymbolicValue(0, {
        'No': 0,
        'Yes': 1,
    })) # Used to determine whether or not damage should cause the brush to move.
    inertiaScale = _KeyValue(1.0) # Scales the angular mass of an object.
        # Used to hack angular damage and collision response.
    massScale = _KeyValue(0.0) # A scale multiplier for the object's mass.
    overridescript = _KeyValue('') # A list of physics key/value pairs that are
        # usually in a physics prop .qc file.
        # Format is 'key,value,key,value,etc'.
    damagetoenablemotion = _KeyValue(0) # If specified, this object will start
        # motion disabled. Once its health has dropped below this
        # specified amount, it will enable motion.
    forcetoenablemotion = _KeyValue(0.0) # If specified, this object will start
        # motion disabled. Any impact that imparts a force greater than
        # this value on the physbox will enable motion.
    puntsound = _KeyValue('') # Sound to make when punted
    
    def Wake(self, userid=0):
        '''
        Wake up this physics object, if it is sleeping.
        '''
        self.fire('Wake', userid=userid)
    
    def Sleep(self, userid=0):
        '''
        Put this physics object to sleep.
        It will wake if given the Wake input, or if force is applied to it.
        Note that physics objects go to sleep automatically after coming to
        rest for a while, so you don't really need to use this.
        '''
        self.fire('Sleep', userid=userid)
    
    def EnableMotion(self, userid=0):
        '''
        Enable physics motion/collision response.
        '''
        self.fire('EnableMotion', userid=userid)
    
    def DisableMotion(self, userid=0):
        '''
        Disable physics motion/collision response.
        '''
        self.fire('DisableMotion', userid=userid)
    
    def DisableFloating(self, userid=0):
        '''
        Disable fluid/floating simulation to reduce cost.
        '''
        self.fire('DisableFloating', userid=userid)
    
    def SetBodyGroup(self, param, userid=0):
        '''
        Set this prop's body group (from 0 - n).
        '''
        self.fire('SetBodyGroup', param, userid=userid)
    
    def physdamagescale(self, param, userid=0):
        '''
        Set the Physics Impact Damage Scale for this character.
        Note: 0 means this feature is disabled for backwards compatibility.
        '''
        self.fire('physdamagescale', param, userid=userid)
    
    def EnableDamageForces(self, userid=0):
        '''
        Damaging the entity applies physics forces to it.
        '''
        self.fire('EnableDamageForces', userid=userid)
    
    def DisableDamageForces(self, userid=0):
        '''
        Damaging the entity does *not* apply physics forces to it.
        '''
        self.fire('DisableDamageForces', userid=userid)
    
    def EnablePuntSound(self, userid=0):
        '''
        Allow this prop to play its own sound when punted
        '''
        self.fire('EnablePuntSound', userid=userid)
    
    def DisablePuntSound(self, userid=0):
        '''
        Prevent this prop from playing its own sound when punted
        '''
        self.fire('DisablePuntSound', userid=userid)


class _PlatSounds(_BaseEntity):
    '''
    Platform sounds.
    '''
    movesnd = _KeyValue(SymbolicValue(0, {
        'No Sound': 0,
        'big elev 1': 1,
        'big elev 2': 2,
        'tech elev 1': 3,
        'tech elev 2': 4,
        'tech elev 3': 5,
        'freight elev 1': 6,
        'freight elev 2': 7,
        'heavy elev': 8,
        'rack elev': 9,
        'rail elev': 10,
        'squeek elev': 11,
        'odd elev 1': 12,
        'odd elev 2': 13,
    })) # The sound played whenever the platform starts moving.
    stopsnd = _KeyValue(SymbolicValue(0, {
        'No Sound': 0,
        'big elev stop1': 1,
        'big elev stop2': 2,
        'freight elev stop': 3,
        'heavy elev stop': 4,
        'rack stop': 5,
        'rail stop': 6,
        'squeek stop': 7,
        'quick stop': 8,
    })) # The sound played when the platform stops moving.
    volume = _KeyValue('0.85') # Sound Volume 0.0 - 1.0


class _Trackchange(_Targetname, _Parentname, _RenderFields, _Global, _PlatSounds):
    '''
    Base class for train track changing entity.
    '''
    height = _KeyValue(0) # The vertical height above the track that the
        # train moves. Negative values moves the train below.
    spawnflags = _KeyValue(Flags({
        'Auto Activate train': 1,
        'Relink track': 2,
        'Start at Bottom': 8,
        'Rotate Only': 16,
        'X Axis': 64,
        'Y Axis': 128,
    }))
    rotation = _KeyValue(0) # The amount this platform should rotate as it moves, in degrees.
    train = _KeyValue('') # Train to Switch
    toptrack = _KeyValue('') # Top Track
    bottomtrack = _KeyValue('') # Bottom Track
    speed = _KeyValue(0) # Move/Rotate Speed


class _BaseTrain(_Targetname, _Parentname, _RenderFields, _Global, _Shadow):
    '''
    Base class for trains.
    '''
    spawnflags = _KeyValue(Flags({
        'No Pitch (X-rot)': 1,
        'No User Control': 2,
        'Passable': 8,
        'Fixed Orientation': 16,
        'HL1 Train': 128,
        'Use max speed for pitch shifting move sound': 256,
        'Is unblockable by player': 512,
    }))
    target = _KeyValue('') # The name of the first path_track in the train's path. The train 
    startspeed = _KeyValue(100) # The maximum speed that this train can move. 
    speed = _KeyValue(0.0) # The speed that the train
        # will move at after it spawns, 0 = stopped.
    velocitytype = _KeyValue(SymbolicValue(0, {
        'Instantaneously': 0,
        'Linear blend': 1,
        'Ease in/ease out': 2,
    })) # The method through which this train changes its velocity
        # as it moves along the path.
    orientationtype = _KeyValue(SymbolicValue(1, {
        'Never (fixed orientation)': 0,
        'Near path_tracks': 1,
        'Linear blend': 2,
        'Ease in/ease out': 3,
    })) # The method through which this train changes its orientation
        # as it moves along the path.
    wheels = _KeyValue(50) # Used for turning and stopping.
    height = _KeyValue(4) # The height above the track that this train moves.
    bank = _KeyValue('0') # Bank Angle on Turns
    dmg = _KeyValue(0) # The amount of damage this train does to entities that block it.
    _minlight = _KeyValue('') # The minimum level of ambient light that hits this brush.
    MoveSound = _KeyValue('') # A sound that is played (and looped) while the train is moving.
    MovePingSound = _KeyValue('') # A sound that is played more frequently
        # as the train speeds up.
    StartSound = _KeyValue('') # A sound played when the train starts moving.
    StopSound = _KeyValue('') # A sound played when the train stops moving.
    volume = _KeyValue(10) # Volume (10 = loudest)
    MoveSoundMinPitch = _KeyValue(60) # The sound pitch value that the train will
        # approach as it comes to a stop.
    MoveSoundMaxPitch = _KeyValue(200) # The sound pitch value that the train will
        # approach as it approaches its 
    MoveSoundMinTime = _KeyValue(0.0) # Minimum interval at which to play the move ping sound.
    MoveSoundMaxTime = _KeyValue(0.0) # Maximum interval at which to play the move ping sound.
    
    def SetSpeed(self, speed_scale, userid=0):
        '''
        Set the speed of the train, as a ratio of max speed [0, 1]
        '''
        speed = RangedValue(speed_scale, 0.0, 1.0)
        self.fire('SetSpeed', speed, userid=userid)
    
    def SetSpeedDir(self, speed_sacle, userid=0):
        '''
        Set the speed of the train, as a ratio of max speed.
        Negative values reverse the direction [-1, 1]
        '''
        speed = RangedValue(speed_scale, -1.0, 1.0)
        self.fire('SetSpeedDir', speed, userid=userid)
    
    def SetSpeedReal(self, speed, userid=0):
        '''
        Set the speed of the train. Must be a positive value from 0 to max speed.
        '''
        self.fire('SetSpeedReal', speed, userid=userid)
    
    def Stop(self, userid=0):
        '''
        Stop the train.
        '''
        self.fire('Stop', userid=userid)
    
    def StartForward(self, userid=0):
        '''
        Start the train moving forward.
        '''
        self.fire('StartForward', userid=userid)
    
    def StartBackward(self, userid=0):
        '''
        Start the train moving backward.
        '''
        self.fire('StartBackward', userid=userid)
    
    def Resume(self, userid=0):
        '''
        Resume the train moving in the current direction after it
        was stopped via the 'Stop' or 'Toggle' input.
        '''
        self.fire('Resume', userid=userid)
    
    def Reverse(self, userid=0):
        '''
        Reverse the direction of the train.
        '''
        self.fire('Reverse', userid=userid)
    
    def Toggle(self, userid=0):
        '''
        Toggle the train between start and stop.
        '''
        self.fire('Toggle', userid=userid)


class ai_speechfilter(_Targetname, _ResponseContext, _EnableDisable):
    '''
    An entity that can be used to control the idle speech patterns of a set of
    NPCs.
    '''
    subject = _KeyValue('') # This is the NPC(s) whose speech we're filtering.
        # May be a targetname or a classname.
    IdleModifier = _KeyValue(1.0) # Multiplier to the percentage chance that
        # our NPC(s) will idle speak. Set to 0 to prevent all idle speech.
    NeverSayHello = _KeyValue(SymbolicValue(0, {
        'Yes': 0,
        'No': 1,
    })) # If set to Yes, our NPC(s) won't greet the player when they first meet him.
    
    def SetIdleModifier(self, modifier, userid=0):
        '''
        Change the idle modifier at runtime
        '''
        self.fire('SetIdleModifier', modifier, userid=userid)


class ambient_generic(_Targetname):
    '''
    Universal ambient sound. Use it to play and control a single sound.
    '''
    preset_values = {
        'None': 0,
        'Huge Machine': 1,
        'Big Machine': 2,
        'Machine': 3,
        'Slow Fade in': 4,
        'Fade in': 5,
        'Quick Fade in': 6,
        'Slow Pulse': 7,
        'Pulse': 8,
        'Quick Pulse': 9,
        'Slow Oscillator': 10,
        'Oscillator': 11,
        'Quick Oscillator': 12,
        'Grunge pitch': 13,
        'Very low pitch': 14,
        'Low pitch': 15,
        'High pitch': 16,
        'Very high pitch': 17,
        'Screaming pitch': 18,
        'Oscillate spinup/down': 19,
        'Pulse spinup/down': 20,
        'Random pitch': 21,
        'Random pitch fast': 22,
        'Incremental Spinup': 23,
        'Alien': 24,
        'Bizarre': 25,
        'Planet X': 26,
        'Haunted': 27,
    }
    
    message = _KeyValue('')
    health = _KeyValue(RangedValue(10, 0, 10))
    preset = _KeyValue(SymbolicValue('None', preset_values))
    volstart = _KeyValue(0)
    fadeinsecs = _KeyValue(RangedValue(0, 0, 100))
    fadeoutsecs = _KeyValue(RangedValue(0, 0, 100))
    pitch = _KeyValue(RangedValue(100, 1, 255))
    pitchstart = _KeyValue(100)
    spinup = _KeyValue(RangedValue(0, 0, 100))
    spindown = _KeyValue(RangedValue(0, 0, 100))
    lfotype = _KeyValue(SymbolicValue(0, {
        'off': 0,
        'sqr': 1,
        'tri': 2,
        'rnd': 3,
    }))
    lforate = _KeyValue(RangedValue(0, 0, 1000))
    lfomodpitch = _KeyValue(RangedValue(0, 0, 100))
    lfomodvol = _KeyValue(RangedValue(0, 0, 100))
    cspinup = _KeyValue(0)
    radius = _KeyValue(1250) # why string?
    spawnflags = _KeyValue(Flags({
        'Play everywhere': 1,
        'Start Silent': 16,
        'Is NOT Looped': 32,
    }, 'Start Silent', 'Is NOT Looped'))
    SourceEntityName = _KeyValue('')
    
    def Pitch(self, pitch, userid=0):
        '''Set the sound pitch.'''
        self.fire('Pitch', pitch, userid=userid)
    
    def PlaySound(self, userid=0):
        '''Start the sound.'''
        self.fire('PlaySound', userid=userid)
    
    def StopSound(self, userid=0):
        '''Stop the sound if it is playing.'''
        self.fire('StopSound', userid=userid)
    
    def ToggleSound(self, userid=0):
        '''Toggle the sound between playing and stopping.'''
        self.fire('ToggleSound', userid=userid)
        
    def Volume(self, volume, userid=0):
        '''Set the sound volume.'''
        self.fire('Volume', volume, userid=userid)
    
    def FadeIn(self, seconds, userid=0):
        '''Fade the sound up to full volume over specified time.'''
        self.fire('FadeIn', seconds, userid=userid)
    
    def FadeOut(self, seconds, userid=0):
        '''Fade the sound to silence over specified time.'''
        self.fire('FadeOut', seconds, userid=userid)


class color_correction(_Targetname, _EnableDisable):
    '''
    An entity to control the color correction in the map.
    '''
    minfalloff = _KeyValue(0.0) # This is the distance to the start of the falloff
        # region (-1 = everywhere)
    maxfalloff = _KeyValue(200.0) # This is the distance to the end of the falloff
        # region (-1 = everywhere)
    maxweight = _KeyValue(1.0) # This is the maximum weight for this lookup
    filename = _KeyValue('') # This is the lookup table filename
    fadeInDuration = _KeyValue(0.0) # Duration of fade in on when enabled.
    fadeOutDuration = _KeyValue(0.0) # Dration of fade out on when disabled.
    
    def SetFadeInDuration(self, duration, userid=0):
        '''
        Set the 'fadeInDuration' variable,
        used to fade cc lookup usage when entity is enabled.
        '''
        self.fire('SetFadeInDuration', duration, userid=userid)
    
    def SetFadeOutDuration(self, duration, userid=0):
        '''
        Set the 'fadeOutDuration' variable,
        used to fade cc lookup usage when entity is disabled.
        '''
        self.fire('SetFadeOutDuration', duration, userid=userid)


class color_correction_volume(_Targetname, _EnableDisable):
    '''
    An entity to control the color correction in the map.
    '''
    fadeDuration = _KeyValue(10.0) # This is the duration for the lookup to
        # fade in/out on extry/exit
    maxweight = _KeyValue(1.0) # This is the maximum weight for this lookup
    filename = _KeyValue('') # This is the lookup table filename


class cycler(_Targetname, _Parentname, _RenderFields):
    '''
    An entity used to display a model for testing purposes.
    Shooting it will cycle through model's animations.
    '''
    spawnflags = _KeyValue(Flags({
        'Not Solid': 0,
    }))
    model = _KeyValue('')
    skin = _KeyValue(0)
    sequence = _KeyValue(0)
    
    def SetSequence(self, sequence, userid=0):
        '''Set the cycler's sequence.'''
        self.fire('SetSequence', sequence, userid=userid)


class env_beam(_Targetname, _Parentname, _RenderFields):
    '''
    Create a visible energy beam running between two entities, in either a line
    or ring shape. Similar to env_laser.

    For the list of attributes and methods, see the Valve Developer Community wiki
    http://developer.valvesoftware.com/wiki/Env_beam
    '''

    # These constants are defined in Source SDK
    MAX_BEAM_WIDTH = 25.5
    MAX_BEAM_NOISEAMPLITUDE = 64.0 # not 255?
    MAX_BEAM_SCROLLSPEED = 100.0

    TouchType_values = {
        'Not a tripwire': 0,
        'Player Only': 1,
        'NPC Only': 2,
        'Player or NPC': 3,
        'Player or NPC or Physprop': 4,
    }

    _flags = {
        'Start On': 1,
        'Toggle': 2,
        'Random Strike': 4,
        'Ring': 8,
        'StartSparks': 16,
        'EndSparks': 32,
        'Decal End': 64,
        'Shade Start': 128,
        'Shade End': 256,
        'Taper Out': 512,
    }

    spawnflags = _KeyValue(Flags(_flags))
    Radius = _KeyValue(256)
    life = _KeyValue(1.0) # 0.0 = infinite, not a string goddamnit
    BoltWidth = _KeyValue(RangedValue(2.0, 1.0, MAX_BEAM_WIDTH))
    NoiseAmplitude = _KeyValue(RangedValue(0.0, 0.0, MAX_BEAM_NOISEAMPLITUDE))
    texture = _KeyValue('sprites/laserbeam.vmt')
    TextureScroll = _KeyValue(RangedValue(35.0, 0.0, MAX_BEAM_SCROLLSPEED))
    framerate = _KeyValue(30)
    framestart = _KeyValue(0)
    StrikeTime = _KeyValue(1.0) # not a string
    damage = _KeyValue(0.0) # not a string
    LightningStart = _KeyValue('') # targetname
    LightningEnd = _KeyValue('') # targetname
    TouchType = _KeyValue(SymbolicValue('Not a tripwire', TouchType_values))
    filtername = _KeyValue('')
    decalname = _KeyValue('Bigshot')
    
    def TurnOn(self, userid=0):
        '''Turn the beam on.'''
        self.fire('TurnOn', userid=userid)

    def TurnOff(self, userid=0):
        '''Turn the beam off.'''
        self.fire('TurnOff', userid=userid)

    def Toggle(self, userid=0):
        '''Toggle the beam's state.'''
        self.fire('Toggle', userid=userid)

    def StrikeOnce(self, userid=0):
        '''
        Cause the beam to strike once. It will stay on for its set Life and
        then turn off (it will never turn off if Life is set to zero).
        '''
        self.fire('StrikeOnce', userid=userid)

    def Noise(self, amplitude, userid=0):
        '''Set the NoiseAmplitude.'''
        # NOTE: FGD specifies this as Amplitude, wiki says Noise
        self.fire('Noise', amplitude, userid=userid)

    def ScrollSpeed(self, speed, userid=0):
        '''Set the TextureScroll.'''
        self.fire('ScrollSpeed', speed, userid=userid)

    def Width(self, width, userid=0):
        '''Set the BoltWidth.'''
        self.fire('Width', width, userid=userid)


class env_beverage(_Targetname, _Parentname):
    '''HL1 Legacy: Beverage Dispenser.'''
    health = _KeyValue(10)
    beveragetype = _KeyValue(SymbolicValue(0, {
        'Coca-Cola': 0,
        'Sprite': 1,
        'Diet Coke': 2,
        'Orange': 3,
        'Surge': 4,
        'Moxie': 5,
        'Random': 6,
    }))
    
    def Activate(self, userid=0):
        '''Enable this dispenser.'''
        self.fire('Activate', userid=userid)


class env_blood(_Targetname, _Parentname):
    '''
    An entity used to spawn blood effects.
    '''
    spraydir = _KeyValue(Vector((0,0,0))) # Spray Direction (Pitch Yaw Roll). 
    color = _KeyValue(SymbolicValue(0, {
        'Red (Human)': 0,
        'Yellow (Alien)': 1,
        'Red': 0,
        'Yellow': 1,
        'Human': 0,
        'Alien': 1,
    })) # Blood Color. 
    amount = _KeyValue('100') # Amount of blood (damage to simulate). 
    spawnflags = _KeyValue(Flags({
        'Random Direction': 1,
        'Blood Stream': 2,
        'On Player': 4,
        'Spray decals': 8,
        'Cloud': 16, # possibly unsupported
        'Drops': 32, # possibly unsupported
        'Gore': 64, # possibly unsupported
    }))
     
    def EmitBlood(self, userid=0):
        '''
        Trigger the blood effect.
        '''
        self.fire('EmitBlood', userid=userid)


class env_bubbles(_Targetname, _Parentname):
    '''
    An entity used to create a volume in which to spawn bubbles.
    '''
    density = _KeyValue(2) # Bubble density. 
    frequency = _KeyValue(2) # Bubble frequency. 
    current = _KeyValue(0) # Speed of Current. 
    spawnflags = _KeyValue(Flags({
        'Start Off': 1,
    }))
     
    def Activate(self, userid=0):
        '''
        Activate the bubbles.
        '''
        self.fire('Activate', userid=userid)
     
    def Deactivate(self, userid=0):
        '''
        Deactivate the bubbles.
        '''
        self.fire('Deactivate', userid=userid)
     
    def Toggle(self, userid=0):
        '''
        Toggle the bubbles on and off.
        '''
        self.fire('Toggle', userid=userid)
     
    def SetDensity(self, density, userid=0):
        '''
        Set the bubble density.
        '''
        self.fire('SetDensity', density, userid=userid)
     
    def SetFrequency(self, frequency, userid=0):
        '''
        Set bubble emission rate in bubbles per second.
        '''
        self.fire('SetFrequency', frequency, userid=userid)
     
    def SetCurrent(self, current, userid=0):
        '''
        Set current speed in inches per second.
        '''
        self.fire('SetCurrent', current, userid=userid)


class env_credits(_Targetname):
    '''
    An entity to control the rolling credits.
    '''
    def RollCredits(self, userid=0):
        '''
        Start the intro credits rolling.
        '''
        self.fire('RollCredits', userid=userid)
    
    def RollOutroCredits(self, userid=0):
        '''
        Start the outro credits rolling.
        '''
        self.fire('RollOutroCredits', userid=userid)
    
    def ShowLogo(self, userid=0):
        '''
        Show the HL2 logo.
        '''
        self.fire('ShowLogo', userid=userid)


class env_cubemap(_BaseEntity):
    '''
    An entity that creates a sample point for the Cubic Environment Map.
    '''
    cubemapsize = _KeyValue(SymbolicValue(0, {
        'Default': 0,
        '1x1': 1,
        '2x2': 2,
        '4x4': 3,
        '8x8': 4,
        '16x16': 5,
        '32x32': 6,
        '64x64': 7,
        '128x128': 8,
        '256x256': 9,
    })) # Cubemap Size
    sides = _KeyValue('') # (Optional) Brushes faces to directly attach to the env_cubemap.
        # Hammer help: Press Pick then click on faces in the 3D View to select them. 
        # Use CTRL while clicking to add or remove from the selection.


class env_dustpuff(_Targetname, _Parentname):
    '''
    An entity that can emit dust puffs.
    '''
    scale = _KeyValue(8.0) # Size of the dust puff.
    speed = _KeyValue(16.0) # Speed at which the dust particles should move.
    color = _KeyValue(Color(128,128,128)) # Dust color
    
    def SpawnDust(self, userid=0):
        '''
        Spawn a dust puff.
        '''
        self.fire('SpawnDust', userid=userid)


class env_effectscript(_Targetname, _Parentname):
    '''
    An entity that allows you to script special visual effects via a script file.
    // NEEDHELP :D
    '''
    model = _KeyValue('models/Effects/teleporttrail.mdl') # Model to use for animation sequences.
    scriptfile = _KeyValue('scripts/effects/testeffect.txt') # Name of the script
        # to use for this model.
    
    def SetSequence(self, param, userid=0):
        '''
        Set the script model's sequence.
        '''
        self.fire('SetSequence', param, userid=userid)


class env_embers(_Targetname, _Parentname):
    '''
    An entity used to create a volume in which to spawn fire embers.
    '''
    particletype = _KeyValue(SymbolicValue(0, {
        'Normal': 0,
        'Smooth Fade': 1,
        'Pulled': 2,
    })) # Ember type. 
    density = _KeyValue(50) # Density (particles per second). 
    lifetime = _KeyValue(4) # Particle Lifetime (seconds). 
    speed = _KeyValue(32) # Particle Speed (units per second). 
    rendercolor = _KeyValue(Color(255,255,255)) # Ember Color (R G B). 
    spawnflags = _KeyValue(Flags({
        'Start On': 1,
        'Toggle': 2,
    }))


class env_entity_igniter(_Targetname):
    '''
    An entity that catches a target entity on fire. If the entity is an animating model, it will have sprite flames attached to its skeleton. Otherwise the entity will emit particle flame puffs.
    '''
    target = _KeyValue('') # Entity to ignite. 
    lifetime = _KeyValue(10.0) # Lifetime in seconds. 
     
    def Ignite(self, userid=0):
        '''
        Ignite the target entity.
        '''
        self.fire('Ignite', userid=userid)


class env_entity_maker(_Targetname, _Parentname):
    '''
    Spawn the specified entity template at its origin.
    If set to auto-spawn, it will spawn the template whenever there's room
    and the player is looking elsewhere.
    '''
    spawnflags = _KeyValue(Flags({
        "Enable AutoSpawn": 1,
        'AutoSpawn: Wait for entity destruction': 2,
        'AutoSpawn: Even if the player is looking': 4,
        "ForceSpawn: Only if there's room": 8,
        "ForceSpawn: Only if the player isn't looking": 16,
    }))
    EntityTemplate = _KeyValue('') # Name of the point_template to spawn here.
    PostSpawnSpeed = _KeyValue(0.0) # If specified, all the entities created in
        # the template will move this fast in the specified PostSpawn Movement Direction.
    PostSpawnDirection = _KeyValue(Vector((0,0,0))), # If a PostSpawn Movement Speed
        # is specified all the entities created in the template will move in this direction.
    PostSpawnDirectionVariance = _KeyValue(0.15) # This variance is applied to the
        # PostSpawn Movement Direction for each spawned entity in the template.
        # Use it to apply some randomness to the directions.
    PostSpawnInheritAngles = _KeyValue(SymbolicValue(0, {
        'No': 0,
        'Yes': 1,
    })) # If in hierarchy, is spawn direction in world space,
        # or object local space of parent
    
    def ForceSpawn(self, userid=0):
        '''
        Spawn an instance of the template at this origin and angle.
        '''
        self.fire('ForceSpawn', userid=userid)
    
    def ForceSpawnAtEntityOrigin(self, targetname, userid=0):
        '''
        Spawn an instance of the template that the same origin and angle as
        the specified entity (specify by targetname in parameters)
        '''
        self.fire('ForceSpawnAtEntityOrigin', targetname, userid=userid)


class env_explosion(_Targetname, _Parentname):
    '''
    An entity that creates an explosion at its origin.
    '''
    iMagnitude = _KeyValue(100) # Magnitude. 
    iRadiusOverride = _KeyValue(0) # Radius Override. 
    fireballsprite = _KeyValue('sprites/zerogxplode.spr') # Fireball Sprite. 
    rendermode = _KeyValue(SymbolicValue(5, {
        'Normal': 0,
        'Solid': 4,
        'Additive': 5,
    })) # Render Mode. 
    spawnflags = _KeyValue(Flags({
        'No Damage': 1,
        'Repeatable': 2,
        'No Fireball': 4,
        'No Smoke': 8,
        'No Decal': 16,
        'No Sparks': 32,
        'No Sound': 64,
        'Random Orientation': 128,
        'No Fireball Smoke': 256,
        'No particles': 512,
        'No DLights': 1024,
        "Don't clamp Min": 2048,
        "Don't clamp Max": 4096,
        'Damage above surface only': 8192,
        'Generic damage': 16384,
    }))
    ignoredEntity = _KeyValue('') # Ignored Entity. 
     
    def Explode(self, userid=0):
        '''
        Trigger the explosion.
        '''
        self.fire('Explode', userid=userid, fatal=True)


class env_fade(_Targetname):
    '''
    An entity that controls screen fades.
    '''
    spawnflags = _KeyValue(Flags({
        'Fade From': 1,
        'Modulate': 2,
        'Stay Out': 8,
    }))
    duration = _KeyValue(2.0) # why string :/
    holdtime = _KeyValue(0.0) # why string :\
    renderamt = _KeyValue(RangedValue(255, 0, 255))
    rendercolor = _KeyValue(Color(0,0,0))
    
    def Fade(self, userid=0):
        '''Start the screen fade.'''
        self.fire('Fade', userid=userid)


class env_fire(_Targetname, _Parentname, _EnableDisable):
    '''
    An entity that handles a single flame at its origin.
    The flame causes heat 'damage' to other env_fire entities around it,
    and will eventually ignite non-flaming env_fire entities nearby,
    causing the fire to spread.
    '''
    health = _KeyValue(30) # Duration. 
    firesize = _KeyValue(64) # Size. 
    fireattack = _KeyValue(4) # Attack. 
    firetype = _KeyValue(SymbolicValue(0, {
        'Natural': 0,
        'Plasma': 1,
    })) # Type. 
    spawnflags = _KeyValue(Flags({
        'Infinite Duration': 1,
        'Smokeless': 2,
        'Start On': 4,
        'Start Full': 8,
        "Don't drop": 16,
        'No glow': 32,
        'Delete when out': 128,
        'Visible from above': 256,
    }))
    ignitionpoint = _KeyValue(32.0) # Ignition Point. 
    damagescale = _KeyValue(1.0) # Damage Scale. 
     
    def StartFire(self, userid=0):
        '''
        Start the fire.
        '''
        self.fire('StartFire', userid=userid)
     
    def Extinguish(self, seconds, userid=0):
        '''
        Put out the fire permanently in the number of seconds specified.
        '''
        self.fire('Extinguish', seconds, userid=userid)
     
    def ExtinguishTemporary(self, seconds, userid=0):
        '''
        Put out the fire temporarily in the number of seconds specified.
        '''
        self.fire('ExtinguishTemporary', seconds, userid=userid)


class env_firesensor(_Targetname, _Parentname):
    '''
    An entity that detects changes in heat nearby.
    '''
    spawnflags = _KeyValue(Flags({
        'Start On': 1,
    }, 'Start On'))
    fireradius = _KeyValue(128.0) # Radius. 
    heatlevel = _KeyValue(32.0) # Heat level. 
    heattime = _KeyValue(0.0) # Time at level. 
     
    def Enable(self, userid=0):
        '''
        Enable fire sensor.
        '''
        self.fire('Enable', userid=userid)
     
    def Disable(self, userid=0):
        '''
        Disable fire sensor.
        '''
        self.fire('Disable', userid=userid)


class env_firesource(_Targetname, _Parentname):
    '''
    An entity that provides heat to all nearby env_fire entities.
    Cannot be extinguished.
    '''
    spawnflags = _KeyValue(Flags({
        'Start On': 1,
    }))
    fireradius = _KeyValue(128.0) # Radius. 
    firedamage = _KeyValue(10.0) # Intensity / Damage. 
     
    def Enable(self, userid=0):
        '''
        Enable fire source.
        '''
        self.fire('Enable', userid=userid)
     
    def Disable(self, userid=0):
        '''
        Disable fire source.
        '''
        self.fire('Disable', userid=userid)


class env_fog_controller(_Targetname, _DXLevelChoice):
    '''
    An entity that controls the fog and view distance in the map.
    '''
    fogenable = _KeyValue(SymbolicValue(0, {
        'No': 0,
        'Yes': 1,
    })) # Fog Enable. 
    fogblend = _KeyValue(SymbolicValue(0, {
        'No': 0,
        'Yes': 1,
    })) # Fog Blend. 
    use_angles = _KeyValue(SymbolicValue(0, {
        'No': 0,
        'Yes': 1,
    })) # Use Angles for Fog Dir. 
    fogcolor = _KeyValue(Color(255,255,255)) # Primary Fog Color. 
    fogcolor2 = _KeyValue(Color(255,255,255)) # Secondary Fog Color. 
    fogdir = _KeyValue(Vector((1,0,0))) # Primary Fog Direction. 
    fogstart = _KeyValue(500.0) # Fog Start. 
    fogend = _KeyValue(2000.0) # Fog End. 
    fogmaxdensity = _KeyValue(RangedValue(1.0, 0.0, 1.0)) # Fog Max Density [0..1]. 
    foglerptime = _KeyValue(0.0) # Interpolate time. 
    farz = _KeyValue(-1) # Far Z Clip Plane. 
    spawnflags = _KeyValue(Flags({
        'Master': 1,
    }))
    
    def SetStartDist(self, distance, userid=0):
        '''
        Set the fog start distance.
        '''
        self.fire('SetStartDist', distance, userid=userid)
    
    def SetEndDist(self, distance, userid=0):
        '''
        Set the fog end distance.
        '''
        self.fire('SetEndDist', distance, userid=userid)
    
    def TurnOn(self, userid=0):
        '''
        Turn the fog on.
        '''
        self.fire('TurnOn', userid=userid)
    
    def TurnOff(self, userid=0):
        '''
        Turn the fog off.
        '''
        self.fire('TurnOff', userid=userid)
    
    def SetColor(self, color, userid=0):
        '''
        Set the primary fog color.
        '''
        self.fire('SetColor', Color(color).for_input(), userid=userid)
    
    def SetColorSecondary(self, color, userid=0):
        '''
        Set the secondary fog color.
        '''
        self.fire('SetColorSecondary', Color(color).for_input(), userid=userid)
    
    def SetFarZ(self, farz, userid=0):
        '''
        Set the far clip plane distance.
        '''
        self.fire('SetFarZ', farz, userid=userid)
    
    def SetAngles(self, angles, userid=0):
        '''
        Set the angles to use for the secondary fog direction.
        '''
        self.fire('SetAngles', Vector(angles).for_input(), userid=userid)
    
    def SetColorLerpTo(self, color, userid=0):
        '''
        Set the primary fog color.
        '''
        self.fire('SetColorLerpTo', Color(color).for_input(), userid=userid)
    
    def SetColorSecondaryLerpTo(self, color, userid=0):
        '''
        Set the secondary fog color.
        '''
        self.fire('SetColorSecondaryLerpTo', Color(color).for_input(), userid=userid)
    
    def SetStartDistLerpTo(self, distance, userid=0):
        '''
        Set the fog start distance.
        '''
        self.fire('SetStartDistLerpTo', distance, userid=userid)
    
    def SetEndDistLerpTo(self, distance, userid=0):
        '''
        Set the fog end distance.
        '''
        self.fire('SetEndDistLerpTo', distance, userid=userid)
    
    def StartFogTransition(self, userid=0):
        '''
        Start fog transition.
        '''
        self.fire('StartFogTransition', userid=userid)


class env_funnel(_Targetname, _Parentname):
    '''
    HL1 Legacy: Large Portal Funnel
    '''
    pass


class env_hudhint(_Targetname):
    '''
    An entity to control the display of HUD hints.
    HUD hints are used to show the player what key is bound to a particular command.
    '''
    message = _KeyValue('') # Hint Text (localized).
        # This should be set to match the desired
        # HUD hint entry in the hl2\resource\valve_english.txt.
    
    def ShowHudHint(self, userid=0):
        '''
        Show the hint message.
        '''
        self.fire('ShowHudHint', userid=userid)
    
    def HideHudHint(self, userid=0):
        '''
        Hide the hint message.
        '''
        self.fire('HideHudHint', userid=userid)


class env_laser(_Targetname, _Parentname, _RenderFxChoices):
    '''
    An entity that creates a laser beam between itself and a given target.
    '''
    LaserTarget = _KeyValue('') # Target of Laser. 
    renderamt = _KeyValue(RangedValue(100, 1, 255)) # Brightness (1 - 255). 
    rendercolor = _KeyValue(Color(255,255,255)) # Beam Color (R G B). 
    width = _KeyValue(2.0) # Width of Beam. 
    NoiseAmplitude = _KeyValue(0) # Amount of noise (0-255). 
    texture = _KeyValue('sprites/laserbeam.spr') # Sprite Name. 
    EndSprite = _KeyValue('') # End Sprite. 
    TextureScroll = _KeyValue(35) # Texture Scroll Rate (0-100). 
    framestart = _KeyValue(0) # Starting Frame.
    damage = _KeyValue(100.0) # Damage / second.
    dissolvetype = _KeyValue(SymbolicValue('None', {
        'None': -1,
        'Energy': 0,
        'Heavy electrical': 1,
        'Light electrical': 2,
    })) # Dissolve Type. 
    spawnflags = _KeyValue(Flags({
        'Start On': 1,
        'StartSparks': 16,
        'EndSparks': 32,
        'Decal End': 64,
    }))
    
    def TurnOn(self, userid=0):
        '''
        Turn the laser on.
        '''
        self.fire('TurnOn', userid=userid)
    
    def TurnOff(self, userid=0):
        '''
        Turn the laser off.
        '''
        self.fire('TurnOff', userid=userid)
    
    def Toggle(self, userid=0):
        '''
        Toggle the laser between on and off.
        '''
        self.fire('Toggle', userid=userid)


class env_lightglow(_Parentname, _Targetname):
    '''
    An entity that puts an additive glow in the world, mostly used over light
    sources.
    '''
    rendercolor = _KeyValue(Color(255,255,255))
    VerticalGlowSize = _KeyValue(30)
    HorizontalGlowSize = _KeyValue(30)
    MinDist = _KeyValue(500)
    MaxDist = _KeyValue(2000)
    OuterMaxDist = _KeyValue(0)
    GlowProxySize = _KeyValue(2.0)
    HDRColorScale = _KeyValue(1.0)
    spawnflags = _KeyValue(Flags({
        'Visible only from front': 1,
    }))
    
    def Color(self, color, userid=0):
        '''Change the render color of the glow.'''
        self.fire('Color', Color(color).for_input(), userid=userid)


class env_message(_Targetname):
    '''
    An entity that draws a text message on player's HUDs.
    '''
    message = _KeyValue('') # Message Text. 
    spawnflags = _KeyValue(Flags({
        'Play Once': 1,
        'All Clients': 2,
    }))
    messagesound = _KeyValue('') # Sound Effect. 
    messagevolume = _KeyValue(RangedValue(10, 0, 10)) # Volume 0-10. 
    messageattenuation = _KeyValue(SymbolicValue(0, {
        'Small Radius': 0,
        'Medium Radius': 1,
        'Large  Radius': 2,
        'Play Everywhere': 3,
    })) # Sound Radius. 
    
    def ShowMessage(self, userid=0):
        '''
        Show the message and plays the sound.
        '''
        self.fire('ShowMessage', userid=userid)


class env_microphone(_Targetname, _Parentname, _EnableDisable):
    '''
    An entity that acts as a microphone. It works in one of two modes.
    If it has a 'Speaker' set, it picks up all sounds within the specified
    sound range, and rebroadcasts them through the Speaker entity.
    In this Speaker mode, it ignores the Hears X spawnflags and does not fire
    the SoundLevel output. If it has no Speaker set, it measures the sound
    level at a point, and outputs the sound level as a value between 0 and 1.
    In Measuring mode, it only hears sounds that match the Hear X spawnflags.
    '''
    target = _KeyValue(''), # If the speaker is in Measuring mode
        # this is the name of the entity where the sound level is to be measured.
    SpeakerName = _KeyValue('') # The name of an info_target entity through which
        # to play any sounds heard by this microphone.
        # If specified, the microphone will consider itself in Speaker mode.
    ListenFilter = _KeyValue('') # The name of an filter entity which specifies
        # the only entities the microphone can hear. Sounds emitted
        # by other entities will not be heard.
    speaker_dsp_preset = _KeyValue(SymbolicValue(0, {
        'Use Default': 0,
        '1 NO EFFECT': 50,
        '2 (DUPLICATE OF 1)': 51,
        '3 (DUPLICATE OF 1)': 52,
        '4 (DUPLICATE OF 1)': 53,
        '5 (DUPLICATE OF 1)': 54,
        '6 SPEAKER, LOUDER': 55,
        '7 SPEAKER VERY SMALL': 56,
        '8 LOUDSPEAKER, ECHO': 57,
        '9 SPEAKER SMALL': 58,
        '10 SPEAKER TINY': 59,
    })) # Only useful in Speaker mode. If specified, when the microphone is enabled,
        # it'll set the global dsp_speaker preset to this value.
        # Sounds played back through speakers will then be affected by the selected DSP.
    spawnflags = _KeyValue(Flags({
        'Hears combat sounds': 1,
        'Hears world sounds': 2,
        'Hears player sounds': 4,
        'Hears bullet impacts': 8,
        'Swallows sounds routed through speakers': 16,
        'Hears explosions': 32,
        'Ignores non-attenuated sounds': 64,
    }, 'Hears combat sounds', 'Hears world sounds', 'Hears player sounds',
        'Hears bullet impacts'))
    Sensitivity = _KeyValue(RangedValue(1.0, 0.0, 10.0)), # Microphone sensitivity
        # 0=deaf, 1=default, 10=extremely sensitive). Only applicable in Measuring mode.
    SmoothFactor = _KeyValue(RangedValue(0.0, 0.0, 1.0)), # Smoothing factor
        # 0=no smoothing, 1=maximum smoothing). Only applicable in Measuring mode.
    MaxRange = _KeyValue(240.0) # Sounds beyond this range won't be heard,
        # irrelevant of attenuation. 
    
    def SetSpeakerName(self, targetname, userid=0):
        '''
        Set the microphone to output through a different speaker entity.
        '''
        self.fire('SetSpeakerName', targetname, userid=userid)


class env_particle_performance_monitor(_Targetname):
    '''
    An entity for turning on and off measuring and display of particle throughput.
    '''
    def TurnOnDisplay(self, userid=0):
        '''
        Turn on display of particle simulation benchmark
        '''
        self.fire('TurnOnDisplay', userid=userid)
    
    def TurnOffDisplay(self, userid=0):
        '''
        Turn off display of particle simulation benchmark
        '''
        self.fire('TurnOffDisplay', userid=userid)
    
    def StartMeasuring(self, userid=0):
        '''
        Start measuring particle simulation speed
        '''
        self.fire('StartMeasuring', userid=userid)
    
    def StopMeasuring(self, userid=0):
        '''
        Stop measuring particle simulation speed
        '''
        self.fire('StopMeasuring', userid=userid)


class env_particlelight(_Parentname):
    '''
    An entity that can be used to light the smoke particles emitted by
    env_smokestack entities. Does not light any other particle types.
    '''
    Color = _KeyValue(Color(255,0,0))
    Intensity = _KeyValue(5000)
    directional = _KeyValue(SymbolicBoolean(0))
    PSName = _KeyValue('')


class env_particlescript(_Targetname, _Parentname):
    '''
    An entity that has special animation events that can be fired by a model
    with an animation inside its .qc designed for use by this entity.
    '''
    model = _KeyValue('models/Ambient_citadel_paths.mdl') # Model to use for animation sequences.
    
    def SetSequence(self, sequence, userid=0):
        '''
        Set the script model's sequence.
        '''
        self.fire('SetSequence', sequence, userid=userid)


class env_physexplosion(_Targetname, _Parentname):
    '''
    An entity that creates an explosion at its origin.
    If the no-damage spawnflag is set, the explosion won't be visible,
    but will apply force to any physics objects within its radius.
    '''
    magnitude = _KeyValue('100') # Magnitude. 
    radius = _KeyValue('0') # Clamp Radius (0 = auto). 
    targetentityname = _KeyValue('') # Limit to Entity. 
    spawnflags = _KeyValue(Flags({
        'No Damage - Only Force': 1,
        'Push players': 2,
        'Push radially - not as a sphere': 4,
        'Test LOS before pushing': 8,
        'Disorient player if pushed': 16,
    }, 'No Damage - Only Force'))
    inner_radius = _KeyValue(0.0) # Inner radius. 
     
    def Explode(self, userid=0):
        '''
        Trigger the explosion.
        '''
        self.fire('Explode', userid=userid)


class env_physimpact(_Targetname, _Parentname):
    '''
    An entity that will cause a physics impact on another entity.
    '''
    magnitude = _KeyValue(100) # Magnitude. 
    distance = _KeyValue(0) # Distance. 
    directionentityname = _KeyValue('') # Point to Entity. 
    spawnflags = _KeyValue(Flags({
        'No fall-off': 1,
        'Infinite Length': 2,
        'Ignore Mass': 4,
        'Ignore Surface Normal When Applying Force': 8,
    }, 'Ignore Surface Normal When Applying Force'))
     
    def Impact(self, userid=0):
        '''
        Trigger the impact
        '''
        self.fire('Impact', userid=userid)


class env_player_surface_trigger(_Targetname):
    '''
    An entity that monitors the material of the surface the player is standing
    on, and fires outputs whenever it changes to/from a specific material.
    '''
    gamematerial_values = {
        "None": 0,
        "Concrete": 67,
        "Dirt": 68,
        "Flesh": 70,
        "Grate": 71,
        "Clip": 73,
        "Metal": 77,
        "Sand": 78,
        "Foliage": 79,
        "Computer": 80,
        "Slosh": 83,
        "Tile": 84,
        "Vent": 86,
        "Wood": 87,
        "Glass": 89,
    }

    gamematerial = _KeyValue(SymbolicValue("None", gamematerial_values))

    def Enable(self, userid=0):
        '''
        Start watching the player's surface.
        '''
        self.fire('Enable', userid=userid)

    def Disable(self, userid=0):
        '''
        Stop watching the player's surface.
        '''
        self.fire('Disable', userid=userid)


class env_projectedtexture(_Targetname, _Parentname):
    '''
    Projected texture entity.
    '''
    spawnflags = _KeyValue(Flags({
        'Enabled': 1,
    }, 'Enabled'))
    target = _KeyValue('') # target
    lightfov = _KeyValue(90.0) # FOV
    nearz = _KeyValue(4.0) # Near Z for projected texture
    farz = _KeyValue(750.0) # Far Z for projected texture
    enableshadows = _KeyValue(SymbolicValue(0, {
        'No': 0,
        'Yes': 1,
    })) # Enables/disables shadows from this projected texture.
    shadowquality = _KeyValue(SymbolicValue(1, {
        'Low': 0,
        'High': 1,
    })) # Quality of shadows.
    lightonlytarget = _KeyValue(SymbolicValue(0, {
        'No': 0,
        'Yes': 1,
    })) # Limit flashlight effect to only effect target entity.
    lightworld = _KeyValue(SymbolicValue(1, {
        'No': 0,
        'Yes': 1,
    })) # Control whether flashlight effects static world geometry.
    lightcolor = _KeyValue(Color(255,255,255,200)) # Light Color RGB-Intensity
    cameraspace = _KeyValue(0) # Angles are interpreted as being relative to camera.
    
    def TurnOn(self, userid=0):
        '''
        Turn on the texture
        '''
        self.fire('TurnOn', userid=userid)
    
    def TurnOff(self, userid=0):
        '''
        Turn off the texture
        '''
        self.fire('TurnOff', userid=userid)
    
    def SetFOV(self, fov, userid=0):
        '''
        Set FOV
        '''
        self.fire('SetFOV', fov, userid=userid)


class env_rotorshooter(_gibshooterbase, _RenderFields):
    '''
    An entity that creates gibs when it's within the influence of a helicopter's rotor wash.
    '''
    shootmodel = _KeyValue('') # Model. Thing to shoot out.  Can be a .mdl or a .vmt.
    shootsounds = _KeyValue(SymbolicValue(-1, {
        'None': -1,
        'Glass': 0,
        'Wood': 1,
        'Metal': 2,
        'Flesh': 3,
        'Concrete': 4,
    })) # Material Sound. 
    simulation = _KeyValue(SymbolicValue(0, {
        'Point': 0,
        'Physics': 1,
        'Ragdoll': 2,
    })) # Simulate. 
    skin = _KeyValue(0), # Gib Skin. Some models have multiple versions of their textures
               # called skins. Set this to a number other than 0 to use that skin
               # on all gibs produced by this shooter.
    spawnflags = _KeyValue(Flags({
        'On fire': 2,
    }))
    rotortime = _KeyValue(1.0) # Time Under Rotor.
        # The average time it has to be under the rotor before it shoots a gib.
    rotortimevariance = _KeyValue(0.3) # Time variance.
        # The random amount to vary the time it has to be under
        # the rotor before it shoots a gib.


class env_rotorwash_emitter(_Targetname, _Parentname):
    '''
    Create rotorwash.
    '''
    altitude = _KeyValue(1024.0) # Altitude the rotorwash will show up.


class env_screeneffect(_Targetname):
    '''
    Allow screenspace effects to be played onthe player's view.
    '''
    type = _KeyValue(SymbolicValue(0, {
        'Advisor Stun': 0,
        'Intro Blur': 1,
        'Groggy Vision': 2,
    }))
    
    def StartEffect(self, seconds, userid=0):
        '''Start the effect for specific time.'''
        self.fire('StartEffect', seconds, userid=userid)
    
    def StopEffect(self, seconds=None, userid=0):
        '''Stop the effect.'''
        # FIXME: what's up with parameter here?!?
        if seconds is None:
            self.fire('StopEffect', userid=userid)
        else:
            self.fire('StopEffect', seconds, userid=userid)


class env_screenoverlay(_Targetname):
    '''
    An entity that can display and control a set of screen overlays,
    to be displayed over the player's view. Useful for view effects like
    drunkenness, or teleporter afterimages, etc.
    '''
    OverlayName1 = _KeyValue('')
    OverlayTime1 = _KeyValue(1.0)
    OverlayName2 = _KeyValue('')
    OverlayTime2 = _KeyValue(1.0)
    OverlayName3 = _KeyValue('')
    OverlayTime3 = _KeyValue(1.0)
    OverlayName4 = _KeyValue('')
    OverlayTime4 = _KeyValue(1.0)
    OverlayName5 = _KeyValue('')
    OverlayTime5 = _KeyValue(1.0)
    OverlayName6 = _KeyValue('')
    OverlayTime6 = _KeyValue(1.0)
    OverlayName7 = _KeyValue('')
    OverlayTime7 = _KeyValue(1.0)
    OverlayName8 = _KeyValue('')
    OverlayTime8 = _KeyValue(1.0)
    OverlayName9 = _KeyValue('')
    OverlayTime9 = _KeyValue(1.0)
    OverlayName10 = _KeyValue('')
    OverlayTime10 = _KeyValue(1.0)
    
    def StartOverlays(self, userid=0):
        '''Start displaying the first overlay.'''
        self.fire('StartOverlays', userid=userid)
    
    def StopOverlays(self, userid=0):
        '''Stop displaying any overlays.'''
        self.fire('StopOverlays', userid=userid)
    
    def SwitchOverlay(self, overlay, userid=0):
        '''Switch to displaying a specific overlay.'''
        # why float and not integer??
        self.fire('SwitchOverlay', overlay, userid=userid)


class env_shake(_Targetname, _Parentname):
    '''
    An entity to control screen shake on players.
    '''
    spawnflags = _KeyValue(Flags({
        'GlobalShake': 1,
        'In Air': 4,
        'Physics': 8,
        'Ropes': 16,
        "DON'T shake view": 32,
        "DON'T Rumble Controller": 64,
    }))
    amplitude = _KeyValue(RangedValue(4.0, 0.0, 16.0)) # Amplitude (0-16). 
    radius = _KeyValue(500.0) # Effect Radius. 
    duration = _KeyValue(1.0) # Duration (seconds). 
    frequency = _KeyValue(RangedValue(2.5, 0.0, 255.0)) # Frequency. 
    
    def Amplitude(self, amplitude, userid=0):
        '''
        Set the amplitude (0-16)
        '''
        self.fire('Amplitude', amplitude, userid=userid)
    
    def Frequency(self, frequency, userid=0):
        '''
        Set the frequency.
        Should be a value between 0 and 255, where 0.1 = jerk, and 255.0 = rumble.
        '''
        self.fire('Frequency', frequency, userid=userid)
    
    def StartShake(self, userid=0):
        '''
        Start the shake.
        '''
        self.fire('StartShake', userid=userid)
    
    def StopShake(self, userid=0):
        '''
        Stop the shake.
        '''
        self.fire('StopShake', userid=userid)


class env_shooter(_gibshooterbase, _RenderFields):
    '''
    An entity that shoots models, or sprites, out of its origin.
    '''
    shootmodel = _KeyValue('') # Model. Thing to shoot out. Can be a .mdl or a .vmt.
    shootsounds = _KeyValue(SymbolicValue(-1, {
        'None': -1,
        'Glass': 0,
        'Wood': 1,
        'Metal': 2,
        'Flesh': 3,
        'Concrete': 4,
    })) # Material Sound. 
    simulation = _KeyValue(SymbolicValue(0, {
        'Point': 0,
        'Physics': 1,
        'Ragdoll': 2,
    })) # Simulate. 
    skin = _KeyValue(0), # Gib Skin. Some models have multiple versions of their textures
               # called skins. Set this to a number other than 0 to use that skin
               # on all gibs produced by this shooter.
    spawnflags = _KeyValue(Flags({
        'On fire': 2,
        'strict remove after lifetime': 4,
    }))
    nogibshadows = _KeyValue(SymbolicValue(0, {
        'No': 0,
        'Yes': 1,
    })) # Disable Shadows on Gibs. 
    gibgravityscale = _KeyValue(1.0) # Gib gravity scale. ONLY WORKS FOR POINT GIBS.
                            # This field allows you to scale gravity so that gibs
                            # fall faster, slower, or not at all.
    massoverride = _KeyValue(0.0) # Specify an arbitrary mass for the gibs emitted by me.


class env_smokestack(_Parentname):
    '''
    An entity that spits out a constant stream of smoke.
    You can place up to two env_particlelight entities near the
    smoke stack to add ambient light to its particles.
    '''
    targetname = _KeyValue('')
    InitialState = _KeyValue(SymbolicBoolean(0))
    BaseSpread = _KeyValue(20)
    SpreadSpeed = KeyvalueInput(15) # :)
    Speed = KeyvalueInput(30) # :|
    StartSize = _KeyValue(20)
    EndSize = _KeyValue(30)
    Rate = KeyvalueInput(20) # :(
    JetLength = KeyvalueInput(180) # >:(
    WindAngle = _KeyValue(0)
    WindSpeed = _KeyValue(0)
    SmokeMaterial = _KeyValue('particle/SmokeStack.vmt')
    twist = _KeyValue(0)
    roll = _KeyValue(0.0)
    rendercolor = _KeyValue(Color(255,255,255))
    renderamt = _KeyValue(RangedValue(255, 0, 255))
    
    def TurnOn(self, userid=0):
        '''Turn on the smokestack.'''
        self.fire('TurnOn', userid=userid)
    
    def TurnOff(self, userid=0):
        '''Turn off the smokestack.'''
        self.fire('TurnOff', userid=userid)
    
    def Toggle(self, userid=0):
        '''Toggle the smokestack between on and off state.'''
        self.fire('Toggle', userid=userid)


class env_smoketrail(_Targetname, _Parentname):
    '''
    An entity that creates a smoke trail.
    '''
    opacity = _KeyValue(0.75) # Sprite Opacity. 
    spawnrate = _KeyValue(20.0) # Spawn Rate. 
    lifetime = _KeyValue(5.0) # Particle Life Time. 
    startcolor = _KeyValue(Color(192,192,192)) # Start Color. 
    endcolor = _KeyValue(Color(160,160,160)) # End Color. 
    emittime = _KeyValue(0.0) # Emitter Life Time. 
    minspeed = _KeyValue(10.0) # Minimum Random Speed. 
    maxspeed = _KeyValue(20.0) # Maximum Random Speed. 
    mindirectedspeed = _KeyValue(0.0) # Minimum Directed Speed. 
    maxdirectedspeed = _KeyValue(0.0) # Maximum Directed Speed. 
    startsize = _KeyValue(15.0) # Starting particle size. 
    endsize = _KeyValue(50.0) # Ending particle size. 
    spawnradius = _KeyValue(15.0) # Spawn radius. 
    firesprite = _KeyValue('sprites/firetrail.spr') # Fire Sprite. 
    smokesprite = _KeyValue('sprites/whitepuff.spr') # Smoke Puff. 


class env_soundscape(_Targetname, _Parentname, _EnableDisable):
    '''
    An entity to control sound in an area.
    The active soundscape at any time is the last one that had line-of-sight
    to the player, and was within the radius.
    '''
    radius = _KeyValue(128) # Radius. If set to -1, then the player can hear the
        # soundscape as long as he can see it (regardless of distance to it).
    soundscape = _KeyValue('Nothing') # Soundscape. The name of the soundscape to use.
    #    'Automatic': 'Automatic',
    #    'Automatic (dialog)': 'Automatic_Dialog',
    #    'Indoor': 'GenericIndoor',
    #    'Outdoor': 'GenericOutdoor',
    #    'Nothing': 'Nothing',
    # Corresponds to an entry in the soundscapes*.txt file in the hl2\scripts directory.
    position0 = _KeyValue('') # Sound Position 0.
    position1 = _KeyValue('') # Sound Position 1.
    position2 = _KeyValue('') # Sound Position 2.
    position3 = _KeyValue('') # Sound Position 3.
    position4 = _KeyValue('') # Sound Position 4.
    position5 = _KeyValue('') # Sound Position 5.
    position6 = _KeyValue('') # Sound Position 6.
    position7 = _KeyValue('') # Sound Position 7.
    # A sound position that will be referenced inside the soundscape text file.
    # Usually used to position a set of sounds within the world.
    
    def Enable(self, userid=0):
        '''
        Enable the soundscape.
        '''
        self.fire('Enable', userid=userid)
    
    def Disabled(self, userid=0):
        '''
        Disable the soundscape.
        '''
        self.fire('Disabled', userid=userid)
    
    def ToggleEnabled(self, userid=0):
        '''
        Toggle the soundscape enabled state.
        '''
        self.fire('ToggleEnabled', userid=userid)


class env_soundscape_proxy(_Targetname, _Parentname):
    '''
    An entity that acts like a soundscape but gets all of its sound
    parameters from another env_soundscape entity.
    '''
    MainSoundscapeName = _KeyValue('') # Soundscape Entity.
                              # The soundscape to get all sound parameters from.
    radius = _KeyValue(128) # Radius. 


class env_soundscape_triggerable(env_soundscape):
    '''
    An entity that works like env_soundscape except that it works in
    conjunction with trigger_soundscape to determine when a player hears it.
    '''
    pass


class env_spark(_Targetname, _Parentname):
    '''
    An entity used to create sparks at its origin.
    '''
    MaxDelay = _KeyValue(0) # Max Delay. The longest delay between sparks (in seconds).
    Magnitude = _KeyValue(SymbolicValue(1, {
        'Small': 1,
        'Medium': 2,
        'Large': 5,
        'Huge': 8,
    })) # Magnitude. The size of the sparks.
    TrailLength = _KeyValue(SymbolicValue(1, {
        'Short': 1,
        'Medium': 2,
        'Long': 3,
    })) # Spark Trail Length. 
    spawnflags = _KeyValue(Flags({
        'Start ON': 64,
        'Glow': 128,
        'Silent': 256,
        'Directional': 512,
    }))
    
    def StartSpark(self, userid=0):
        '''
        Start the spark effect.
        '''
        self.fire('StartSpark', userid=userid)
    
    def StopSpark(self, userid=0):
        '''
        Stop the spark effect.
        '''
        self.fire('StopSpark', userid=userid)
    
    def ToggleSpark(self, userid=0):
        '''
        Toggle the on/off state of the spark effect.
        '''
        self.fire('ToggleSpark', userid=userid)
    
    def SparkOnce(self, userid=0):
        '''
        Spark once.
        '''
        self.fire('SparkOnce', userid=userid)


class env_splash(_Targetname):
    '''
    An entity that creates a splash effect at its origin.
    If the 'find water surface' spawnflag is set, it will instead trace
    down below itself to find the water surface on which to create splashes.
    '''
    scale = _KeyValue(8.0)
    spawnflags = _KeyValue(Flags({
        'Automatically find water surface': 1,
        'Diminish with depth': 2,
    }, 'Diminish with depth'))
    
    def Splash(self, userid=0):
        '''Create a splash effect.'''
        self.fire('Splash', userid=userid)


# NOTE: this entity type is unavailable until it can be implemented correctly
# says in the FGD. If it will be implemented, it will likely be another name
# than env_splash...
class DISABLED_env_splash(_Targetname, _Parentname):
    '''
    Can be used to create either a spash effect or a stream of falling or
    spurting liquid. Will create a splash decal on the collided surface the
    same color as the liquid.
    '''
    spawnrate = _KeyValue(10.0) # SpawnRate. 
    startcolor = _KeyValue(Color(100,100,100)) # StartColor. 
    endcolor = _KeyValue(Color(240,110,0)) # EndColor. 
    speed = _KeyValue(3.0) # Speed. 
    speedrange = _KeyValue(1.0) # SpeedRange. 
    widthmin = _KeyValue(2.0) # WidthMin. 
    widthmax = _KeyValue(8.0) # WidthMax. 
    noise = _KeyValue(0.10000000000000001) # Noise. 
    lifetime = _KeyValue(5.0) # Lifetime. 
    numdecals = _KeyValue(1) # Num Decals. 
    startactive = _KeyValue(SymbolicValue(1, {
        'No': 0,
        'Yes': 1,
    })) # Start On. 
     
    def SetSpawnRate(self, param, userid=0):
        '''
        Set how many particles come out
        '''
        self.fire('SetSpawnRate', param, userid=userid)
     
    def SetSpeed(self, param, userid=0):
        '''
        Set speed of emitted particle
        '''
        self.fire('SetSpeed', param, userid=userid)
     
    def SetNoise(self, param, userid=0):
        '''
        Set noise of emitted particle (0-1)
        '''
        self.fire('SetNoise', param, userid=userid)
     
    def SetLifetime(self, param, userid=0):
        '''
        Set lifetime of emitted particles (in seconds)
        '''
        self.fire('SetLifetime', param, userid=userid)
     
    def TurnOn(self, userid=0):
        '''
        Turn particles on
        '''
        self.fire('TurnOn', userid=userid)
     
    def TurnOff(self, userid=0):
        '''
        Turn particles off
        '''
        self.fire('TurnOff', userid=userid)


class env_sprite(_Targetname, _Parentname, _RenderFields, _DXLevelChoice):
    '''
    An entity that controls the drawing of a sprite in the world.
    '''
    framerate = _KeyValue(10.0) # Rate at which the sprite should animate, if at all.
    model = _KeyValue('sprites/glow01.spr') # Material of the sprite to be drawn.
    scale = _KeyValue(RangedValue(1.0, 0.0, 8.0)) # Scale multiplier of the sprite.
    spawnflags = _KeyValue(Flags({
        'Start on': 1,
        'Play Once': 2,
    }))
    GlowProxySize = _KeyValue(2.0) # Size of the glow to be rendered for visibility testing.
        # Must be larger than the distance from the sprite center to empty space.
        # So if this glow is inside geometry (like a light bulb),
        # set this value to be bigger than the bulb's radius.
        # Any time a sphere of this radius would be visible
        # (poking through any nearby geometry), the glow will be rendered.
   
    def SetScale(self, scale, userid=0):
        '''
        Set the sprite's scale (0 - 8.0).
        '''
        self.fire('SetScale', scale, userid=userid)
    
    def HideSprite(self, userid=0):
        '''
        Hide the sprite. Won't be drawn until the 'ShowSprite' input is received.
        '''
        self.fire('HideSprite', userid=userid)
    
    def ShowSprite(self, userid=0):
        '''
        Show the sprite.
        '''
        self.fire('ShowSprite', userid=userid)
    
    def ToggleSprite(self, userid=0):
        '''
        Toggle the sprite between hidden and shown.
        '''
        self.fire('ToggleSprite', userid=userid)


class env_sprite_oriented(env_sprite):
    '''
    A env_sprite that allows orientation.
    
    Hint: set angles for the orientation.
    '''
    pass


class env_spritetrail(_Parentname, _Targetname):
    '''
    A magical trail you can parent to anything your heart desires.
    '''
    lifetime = _KeyValue(0.5) # Lifetime
    startwidth = _KeyValue(8.0) # Start Width
    endwidth = _KeyValue(1.0) # End Width
    spritename = _KeyValue('sprites/bluelaser1.vmt') # Sprite Name
    renderamt = _KeyValue(255) # The FX amount is used by the selected Render Mode.
    rendercolor = _KeyValue(Color(255,255,255)) # The FX color is used by the
        # selected Render Mode.
    rendermode = _KeyValue(SymbolicValue(5, {
        'Normal': 0,
        'Solid': 4,
        'Additive': 5,
    })) # Render Mode


class env_steam(_Targetname, _Parentname):
    '''
    An entity used to create a jet of steam.
    '''
    spawnflags = _KeyValue(Flags({
        'Emissive': 1,
    }))
    InitialState = _KeyValue(SymbolicValue(0, {
        'Off': 0,
        'On': 1,
    })) # Initial State. 
    type = _KeyValue(SymbolicValue(0, {
        'Normal': 0,
        'Heat Wave': 1,
    })) # Particle Type. 
    SpreadSpeed = KeyvalueInput(15) # Spread Speed. 
    Speed = KeyvalueInput(120) # Speed. 
    StartSize = _KeyValue(10) # Particle start size. 
    EndSize = _KeyValue(25) # Particle end size. 
    Rate = KeyvalueInput(26) # Emission rate. 
    rendercolor = _KeyValue(Color(255,255,255)) # Color (R G B). 
    JetLength = KeyvalueInput(80) # Length of steam jet. 
    renderamt = _KeyValue(RangedValue(255, 0, 255)) # Translucency. 
    rollspeed = _KeyValue(8.0) # How fast do the particles spin. 
    
    def TurnOn(self, userid=0):
        '''
        Turn the steam jet on.
        '''
        self.fire('TurnOn', userid=userid)
    
    def TurnOff(self, userid=0):
        '''
        Turn the steam jet off.
        '''
        self.fire('TurnOff', userid=userid)
    
    def Toggle(self, userid=0):
        '''
        Toggle the steam jet between on and off.
        '''
        self.fire('Toggle', userid=userid)


class env_sun(_Targetname):
    '''
    An entity to control & draw a sun effect in the sky.
    '''
    target = _KeyValue('')
    use_angles = _KeyValue(SymbolicBoolean(0))
    pitch = _KeyValue(0)
    rendercolor = _KeyValue(Color(100,80,80))
    overlaycolor = _KeyValue(Color(0,0,0))
    size = _KeyValue(16)
    overlaysize = _KeyValue(-1)
    material = _KeyValue('sprites/light_glow02_add_noz')
    overlaymaterial = _KeyValue('sprites/light_glow02_add_boz')
    HDRColorScale = _KeyValue(1.0)
    
    def TurnOn(self, userid=0):
        '''Enable sun rendering.'''
        self.fire('TurnOn', userid=userid)
    
    def TurnOff(self, userid=0):
        '''Disable sun rendering.'''
        self.fire('TurnOff', userid=userid)
    
    def SetColor(self, color, userid=0):
        '''Change the sun's color.'''
        self.fire('SetColor', Color(color).for_input(), userid=userid)


class env_texturetoggle(_Targetname):
    '''
    An entity that allows you to change the textures on other brush-built
    entities.
    '''
    target = _KeyValue('')
    
    def IncrementTextureIndex(self, userid=0):
        '''Increment target brush's current texture frame by one.'''
        self.fire('IncrementTextureIndex', userid=userid)
    
    def SetTextureIndex(self, index, userid=0):
        '''Set target brush's texture frame to the specified index.'''
        self.fire('SetTextureIndex', index, userid=userid)


class env_tonemap_controller(_Targetname):
    '''
    An entity that controls the HDR tonemapping for the player.
    Think of it as a method of controlling the exposure of the player's eyes.
    '''
    def SetTonemapScale(self, scale, userid=0):
        '''
        Set the player's tonemap scale. It should be a value between 0 and 2,
        where 0 is the eyes fully closed, 1 is use the unchanged autoexposure
        (default), and 2 is the eye fully wide open.
        '''
        self.fire('SetTonemapScale', scale, userid=userid)
 
    def BlendTonemapScale(self, scale, duration, userid=0):
        '''
        Blend from the player's current tonemap scale to a new one.
        '''
        self.fire('BlendTonemapScale', "%s %s"%(scale, duration), userid=userid)
 
    def UseDefaultAutoExposure(self, userid=0):
        '''
        Revert to using the default tonemap auto exposure.
        '''
        self.fire('UseDefaultAutoExposure', userid=userid)
 
    def SetAutoExposureMin(self, value, userid=0):
        '''
        Set a custom tonemap auto exposure minimum.
        '''
        self.fire('SetAutoExposureMin', value, userid=userid)
 
    def SetAutoExposureMax(self, value, userid=0):
        '''
        Set a custom tonemap auto exposure maximum.
        '''
        self.fire('SetAutoExposureMax', value, userid=userid)
 
    def SetBloomScale(self, value, userid=0):
        '''
        Set a custom bloom scale.
        '''
        self.fire('SetBloomScale', value, userid=userid)
 
    def SetTonemapRate(self, value, userid=0):
        '''
        Set the rate for autoexposure adjustment.
        '''
        self.fire('SetTonemapRate', value, userid=userid)


class env_viewpunch(_Targetname, _Parentname):
    '''
    Cause a view punch on players.
    '''
    spawnflags = _KeyValue(Flags({
        'Punch all players (ignore radius)': 1,
        'Punch players in the air': 2,
    }))
    punchangle = _KeyValue(Vector((0,0,90))) # Punch angles. The punch angles to apply.
    radius = _KeyValue(500.0) # Effect Radius.
         #The radius around this entity in which to affect players.
    
    def ViewPunch(self, userid=0):
        '''
        Perform the view punch.
        '''
        self.fire('ViewPunch', userid=userid)


class env_wind(_Targetname):
    '''
    An entity to control wind in the map. Partially functional.
    '''
    minwind = _KeyValue(20) # Minimum speed of the wind while idling.
    maxwind = _KeyValue(50) # Maximum speed of the wind while idling.
    mingust = _KeyValue(100) # Minimum speed of wind gusts.
    maxgust = _KeyValue(250) # Maximum speed of wind gusts.
    mingustdelay = _KeyValue(10) # Minimum time delay between random gusts.
    maxgustdelay = _KeyValue(20) # Maximum time delay between random gusts.
    gustduration = _KeyValue(5) # How long will the wind gust for.
    gustdirchange = _KeyValue(20) # Maximum amount that the wind's
        # direction changes due to a gust.


class env_zoom(_Targetname):
    '''
    An entity that can be used to contorl the player's FOV.
    Useful for scenes where the player's view is being controlled,
    or player-usable binoculars/telescopes, etc.
    '''
    Rate = _KeyValue(1.0)
    FOV = _KeyValue(75)
    spawnflags = _KeyValue(Flags({
        'Allow Suit Zoom': 1,
    }))
    
    def Zoom(self, userid=0):
        '''Start controlling the player's FOV.'''
        self.fire('Zoom', userid=userid)
    
    def UnZoom(self, userid=0):
        '''Stop controlling the player's FOV.'''
        self.fire('UnZoom', userid=userid)


class filter_activator_class(_BaseFilter):
    '''
    A filter that filters by the class name of the activator.
    '''
    filterclass = _KeyValue(''), # The class name to filter by. If the filter mode is Allow
        # only entities whose class name matches the given string will pass the filter.
        # If the filter mode is Disallow, all entities EXCEPT those whose class
        # name matches the given string will pass the filter.


class filter_activator_mass_greater(_BaseFilter):
    '''
    A filter that filters by the mass of the activator.
    '''
    filtermass = _KeyValue(0.0) # The mass to filter by.
        # If the filter mode is Allow, only entities whose mass is greater
        # than the give float will pass the filter. If the filter mode is
        # Disallow, all entities EXCEPT those whose mass is greater than
        # the given float will pass the filter.


class filter_activator_name(_BaseFilter):
    '''
    A filter that filters by the name of the activator.
    '''
    filtername = _KeyValue(''), # The name to filter by. If the filter mode is Allow
        # only entities whose name matches the given string will pass the filter.
        # If the filter mode is Disallow, all entities EXCEPT those whose
        # name matches the string will pass the filter.


class filter_damage_type(_BaseFilter):
    '''
    A damage filter that filters by the type of damage inflicted.
    This can only be used as a damage filter, not as an activator filter.
    '''
    damagetype = _KeyValue(SymbolicValue(64, {
        'GENERIC': 0,
        'CRUSH': 1,
        'BULLET': 2,
        'SLASH': 4,
        'BURN': 8,
        'FREEZE': 16,
        'FALL': 32,
        'BLAST': 64,
        'CLUB': 128,
        'SHOCK': 256,
        'SONIC': 512,
        'ENERGYBEAM': 1024,
        'DROWN': 16384,
        'PARALYSE': 32768,
        'NERVEGAS': 65536,
        'POISON': 131072,
        'RADIATION': 262144,
        'DROWNRECOVER': 524288,
        'CHEMICAL': 1048576,
        'SLOWBURN': 2097152,
        'SLOWFREEZE': 4194304,
    })) # The damage type to filter by. If the filter mode is Allow,
        # only damage types that match will pass the filter. If the filter
        # mode is Disallow, all damage types EXCEPT those who match will
        # pass the filter.


class filter_enemy(_BaseFilter):
    '''
    A filter that filters a potential enemy entity by a set of criteria.
    '''
    filtername = _KeyValue('') # The classname or entity name to filter by.
        # If the filter mode is Allow, only entities whose class name
        # matches the given string will pass the filter. If the filter
        # mode is Disallow, all entities EXCEPT those whose class name
        # matches the given string will pass the filter.
    filter_radius = _KeyValue(0.0) # Radius by which to test the proximity of the
        # enemy. If the filter mode is Allow, only entities whose distance
        # is equal to or closer than the radius will pass the filter.
        # If the filter mode is Disallow, all entities outside the radius
        # will pass the filter.
    filter_outer_radius = _KeyValue(0.0) # Enemies outside this radius are
        # considered invalid if Allow is set and valid if Disallow is set.
    filter_max_per_enemy = _KeyValue(0) # Maximum number of squadmates allowed to
        # target any given entity.
    spawnflags = _KeyValue(Flags({
        'Do not lose target if already aquired but filter failed.': 1,
    }))


class filter_multi(_BaseFilter):
    '''
    A filter that tests the activator against multiple filters.
    This allows you to build more complex filters, such as
    'Allow anyone on Team 1 who is also class engineer', or
    'Allow everyone except classes npc_zombie and npc_headcrab'.
    '''
    filtertype = _KeyValue(SymbolicValue('AND', {
        'AND': 0,
        'OR': 1,
    })) # Logic Type
    Negated = _KeyValue(SymbolicValue(0, {
        'No': 0,
        'Yes': 1,
    })) # Whether to negate the result of the subfilters,
        # after combining them using the Logic Type chosen.
        # Negating the outcome using the AND logic type means that any
        # subfilter must fail for this filter to pass.
        # Negating the outcome using the OR logic type means that all
        # subfilters must fail for this filter to pass.
    Filter01 = _KeyValue('') # Activator filter to test.
    Filter02 = _KeyValue('') # Activator filter to test.
    Filter03 = _KeyValue('') # Activator filter to test.
    Filter04 = _KeyValue('') # Activator filter to test.
    Filter05 = _KeyValue('') # Activator filter to test.


class func_areaportal(_Targetname):
    '''
    A portal brush used to manage visibility in maps.
    Portals define areas, which are spaces that are connected in the map.
    Both sides of a portal cannot touch the same area, for example, a doughnut
    shaped map would require at least two portals to divide the map into two
    areas. A linear map could be divided into two areas with a single area portal.
    '''
    target = _KeyValue('') # (Optional) The name of a door whose open/closed
        # state controls the on/off state of this area portal.
    StartOpen = _KeyValue(SymbolicValue(1, {
        'Closed': 0,
        'Open': 1,
    })) # Initial State
    PortalVersion = _KeyValue(1, readonly=True)
        # Differentiates between shipping HL2 maps and maps using new engine features.
    
    def Open(self, userid=0):
        '''
        Open the portal. When the portal is open is can be seen through.
        '''
        self.fire('Open', userid=userid)
    
    def Close(self, userid=0):
        '''
        Close the portal. When the portal is closed it cannot be seen through.
        '''
        self.fire('Close', userid=userid)
    
    def Toggle(self, userid=0):
        '''
        Toggle the open/closed state of the portal.
        '''
        self.fire('Toggle', userid=userid)


class func_areaportalwindow(_Targetname):
    '''
    An entity that can be used to optimize the visibility in a map.
    If you seal off an area with them, when the viewer moves the specified
    distance away from them, they will go opaque and the parts inside the area
    will not be drawn. The 'target' brush model should enclose the
    func_areaportal window so no parts of it are culled by the window.
    If you use the optional foreground brush model, then it should enclose
    the 'target' brush model.
    '''
    PortalVersion = _KeyValue(1)
    target = _KeyValue("")
    FadeStartDist = _KeyValue(128)
    BackgroundBModel = _KeyValue("")
    FadeDist = _KeyValue(512)
    TranslucencyLimit = _KeyValue(0.2) # string?
    
    def SetFadeStartDistance(self, distance, userid=0):
        '''
        Set fade start distance.
        '''
        self.fire('SetFadeStartDistance', distance, userid=userid)
    
    def SetFadeEndDistance(self, distance, userid=0):
        '''
        Set fade end distance.
        '''
        self.fire('SetFadeEndDistance', distance, userid=userid)


class func_breakable(_BreakableBrush, _RenderFields, _Shadow):
    '''
    A brush entity that can be broken from damage, or an input.
    '''
    minhealthdmg = _KeyValue(0) # The prop will ignore any damage events if the
        # damage is less than this amount.
    _minlight = _KeyValue('') # The minimum level of ambient light that hits this brush.
    physdamagescale = _KeyValue(1.0) # Scales damage energy when this object is hit
        # by a physics object. Note: 0 means this feature is disabled for
        # backwards compatibility.
        # Set to 1.0 for materials as strong as flesh, smaller numbers
        # indicate stronger materials.


class func_breakable_surf(_BreakableBrush, _RenderFields, _Shadow):
    '''
    A breakable surface, for partially breakable glass / tile / etc.
    All faces but the desired visible one must be marked as NODRAW and
    thatface must be 4 sided.
    The material applied to the visible face must be set up to be breakable.
    '''
    spawnflags = _KeyValue(Flags({
        'Physics damage decals': 1,
        'Take damage from held objects': 2,
    }))
    health = _KeyValue(5) # The amount of damage the surface takes before breaking.
    fragility = _KeyValue(100) # If the 'Surface Type' is set to Glass, this value
        # sets how fragile the glass pieces are after the surface has been
        # broken.
    surfacetype = _KeyValue(SymbolicValue(0, {
        'Glass': 0,
        'Tile': 1,
    })) # Surface Type
    
    def Shatter(self, x_center, y_center, radius, userid=0):
        '''
        Shatter the window. Input a vector.
        First two coordinates are the X,Y center of the shattering
        (as values from from 0-1).
        The third coordinate is the radius of the shatter, in inches.
        '''
        x_cent = RangedValue(x_center, 0.0, 1.0)
        y_cent = RangedValue(y_center, 0.0, 1.0)
        self.fire('Shatter', "%s %s %s"%(x_cent, y_cent, radius), userid=userid)


class func_brush(_Targetname, _Parentname, _RenderFields, _Global,
    _Inputfilter, _EnableDisable, _Shadow):
    '''
    An brush built entity with various features.
    '''
    Solidity_values = {
        "Toggle": 0,
        "Never Solid": 1,
        "Always Solid": 2,
    }
    
    excludednpc = _KeyValue("")
    _minlight = _KeyValue("")
    Solidity = _KeyValue(SymbolicValue("Toggle", Solidity_values))
    solidbsp = _KeyValue(SymbolicBoolean(0))
    vrad_brush_cast_shadows = _KeyValue(SymbolicBoolean(0))
    spawnflags = _KeyValue(Flags({
        "Ignore player +USE": 2,
    }, "Ignore player +USE"))
    invert_exclusion = _KeyValue(SymbolicBoolean(0))
    
    def SetExcluded(self, excludednpc, userid=0):
        '''
        Change the NPC class excluded from collisions.
        '''
        self.fire('SetExcluded', excludednpc, userid=userid)
    
    def SetInvert(self, value, userid=0):
        '''
        Set the state of inversion for NPC class exclusion (0 or 1).
        '''
        self.fire('SetInvert', value, userid=userid)


class func_button(_Targetname, _Parentname, _RenderFields, _DamageFilter, _Button):
    '''
    A brush entity that's designed to be used for a player-useable button.
    When used by the player, it moves to a pressed position.
    '''
    movedir = _KeyValue(Vector((0,0,0))) # Specifies the direction of motion to move
        # when the button is used.
    speed = _KeyValue(5) # The speed that the button moves, in inches per second.
    health = _KeyValue(0) # Legacy method of specifying whether or not the button
        # can be shot to activate it. Use the 'Damage Activates' spawnflag instead.
    lip = _KeyValue(0) # The amount, in inches, of the button to leave sticking
        # out of the wall it recedes into when pressed.
        # Negative values make the button recede even further into the wall.
    master = _KeyValue('') # Legacy support
    sounds = _KeyValue(SymbolicValue(0, {
        'None (Silent)': 0,
        'Big zap & Warmup': 1,
        'Access Denied': 2,
        'Access Granted': 3,
        'Quick Combolock': 4,
        'Power Deadbolt 1': 5,
        'Power Deadbolt 2': 6,
        'Plunger': 7,
        'Small zap': 8,
        'Keycard Sound': 9,
        'Buzz': 10,
        'Buzz Off': 11,
        'latch locked': 12,
        'Latch Unlocked': 13,
        'Lightswitch': 14,
        'small bleek': 15,
        'small deny': 16,
        'small doop': 17,
        'small tech deny': 18,
        'click and combine screen fuzz': 19,
        'roomy beep': 20,
        'lever or wheel: turn + move sqeek': 21,
        'lever or wheel: latch + release gas': 22,
        'lever or wheel: ratchet + sqeek': 23,
        'lever or wheel: large ratchet': 24,
        'lever or wheel: clanky + gas release': 25,
        'lever or wheel: latch + large metal thud': 26,
        'lever or wheel: smaller ratchet': 27,
        'lever or wheel: smaller lever move': 28,
        'shock buzz': 31,
        'clickbeep': 32,
        'tech blip': 33,
        'clickbeepbeep open': 34,
        'small high blip': 35,
        'small tech fuzz blip': 36,
        'small click bleep (change to lightswitch)': 37,
        'combine door lock - locked': 40,
        'combine blip growl': 41,
        'combine squick growl': 42,
        'combine whine purr': 43,
        'combine click talk': 44,
        'combine click growl fizz': 45,
        'combine click fizz (deny)': 46,
        'combine click talker': 47,
    })) # Sounds
    wait = _KeyValue(3), # Amount of time, in seconds
        # after the button has been pressed before it returns to the starting position.
        # Once it has returned, it can be used again.
        # If the value is set to -1, the button never returns.
    spawnflags = _KeyValue(Flags({
        "Don't move": 1,
        'Toggle': 32,
        'Touch Activates': 256,
        'Damage Activates': 512,
        'Use Activates': 1024,
        'Starts locked': 2048,
        'Sparks': 4096,
    }, 'Use Activates'))
    locked_sound = _KeyValue(SymbolicValue(0, {
        'None': 0,
        'Access Denied': 2,
        'Small zap': 8,
        'Buzz': 10,
        'Buzz Off': 11,
        'Latch Locked': 12,
    })) # Sound played when the player tries to use the button,
        # and fails because it's locked.
    unlocked_sound = _KeyValue(SymbolicValue(0, {
        'None': 0,
        'Big zap & Warmup': 1,
        'Access Granted': 3,
        'Quick Combolock': 4,
        'Power Deadbolt 1': 5,
        'Power Deadbolt 2': 6,
        'Plunger': 7,
        'Small zap': 8,
        'Keycard Sound': 9,
        'Buzz': 10,
        'Latch Unlocked': 13,
        'Lightswitch': 14,
    })) # Sound played when the button is unlocked.
    locked_sentence = _KeyValue(SymbolicValue(0, {
        'None': 0,
        'Gen. Access Denied': 1,
        'Security Lockout': 2,
        'Blast Door': 3,
        'Fire Door': 4,
        'Chemical Door': 5,
        'Radiation Door': 6,
        'Gen. Containment': 7,
        'Maintenance Door': 8,
        'Broken Shut Door': 9,
    })) # A sentence played when the player tries to use the button,
        # and fails because it's locked.
    unlocked_sentence = _KeyValue(SymbolicValue(0, {
        'None': 0,
        'Gen. Access Granted': 1,
        'Security Disengaged': 2,
        'Blast Door': 3,
        'Fire Door': 4,
        'Chemical Door': 5,
        'Radiation Door': 6,
        'Gen. Containment': 7,
        'Maintenance area': 8,
    })) # A sentence played when the button is unlocked.
    _minlight = _KeyValue('') # The minimum level of ambient light that hits this brush.


class func_clip_vphysics(_Targetname, _EnableDisable):
    '''
    A brush entity that's considered solid to vphysics.
    '''
    filtername = _KeyValue('')


class func_conveyor(_Targetname, _Parentname, _RenderFields, _Shadow):
    '''
    Conveyor Belt
    NEEDHELP?
    '''
    movedir = _KeyValue(Vector((0,0,0))) # The direction conveyor moves.
    spawnflags = _KeyValue(Flags({
        'No Push': 1,
        'Not Solid': 2,
    }))
    speed = _KeyValue('100') # Conveyor Speed
    _minlight = _KeyValue('') # The minimum level of ambient light that hits this brush.
    
    def ToggleDirection(self, userid=0):
        '''
        ToggleDirection
        '''
        self.fire('ToggleDirection', userid=userid)
    
    def SetSpeed(self, param, userid=0):
        '''
        SetSpeed
        '''
        self.fire('SetSpeed', param, userid=userid)


class func_detail(_DXLevelChoice):
    '''
    An entity that turns its brushes into detail brushes.
    Detail brushes do NOT contribute to visibility in the PVS.
    World geometry is not clipped to detail brushes, so if you have a small
    detail clump attached to a wall, the wall won't be cut up by the detail
    brush. func_detail is great for high-frequency brush geometry that's
    visual detail only. It is also ideal for reducing map VIS time.
    '''
    pass


class func_door(_Door):
    '''
    A brush entity for use as a player-useable door.
    '''
    movedir = _KeyValue(Vector((0,0,0))) # The direction the door will move, when it opens.
    filtername = _KeyValue('') # Filter to use to determine entities that block the door.
        # ( Half-Life: Source port only )


class func_door_rotating(_Door):
    '''
    A brush entity for use as a rotating player-useable door.
    '''
    spawnflags = _KeyValue(Flags({
        'Reverse Dir': 2,
        'One-way': 16,
        'X Axis': 64,
        'Y Axis': 128,
        'New func_door +USE rules': 65536,
    }))
    distance = _KeyValue(90) # The amount, in degrees, that the button should
        # rotate when it's pressed.
    solidbsp = _KeyValue(SymbolicValue(0, {
        'VPhysics': 0,
        'BSP': 1,
    })) # Solid Type


class func_dustcloud(_Targetname, _BModelParticleSpawner):
    '''
    A brush entity that spawns a translucent dust cloud within its volume.
    '''
    Alpha = _KeyValue(30) # Alpha
    SizeMin = _KeyValue(100) # Minimum Particle Size STRING?
    SizeMax = _KeyValue(200) # Maximum Particle Size STRING?


class func_dustmotes(_Targetname, _BModelParticleSpawner):
    '''
    A brush entity that spawns sparkling dust motes within its volume.
    '''
    SizeMin = _KeyValue(10) # Minimum Particle Size
    SizeMax = _KeyValue(20) # Maximum Particle Size
    Alpha = _KeyValue(255) # Alpha


class func_fish_pool(_BaseEntity):
    '''
    Create a school of interactive fish that swim near this entity.
    '''
    model = _KeyValue('models/Junkola.mdl') # World model
    fish_count = _KeyValue(10) # Number of Fish in this Pool
    max_range = _KeyValue(RangedValue(150.0, 0.0, 255.0)) # How far away
        # a Fish can wander (max 255)


class func_guntarget(_Targetname, _Parentname, _RenderFields, _Global):
    '''
    This is a moving target that moves along a path of path_tracks.
    It can be shot and killed.
    '''
    speed = _KeyValue(100) # The speed at which the target moves along its path.
    target = _KeyValue('') # The name of the first path_track entity in the path
        # that this target should follow.
    health = _KeyValue(0) # The amount of damage taken before this target is killed.
    _minlight = _KeyValue('') # The minimum level of ambient light that hits this brush.
    
    def Start(self, userid=0):
        '''
        Start the target moving.
        '''
        self.fire('Start', userid=userid)
    
    def Stop(self, userid=0):
        '''
        Stop the target from moving.
        '''
        self.fire('Stop', userid=userid)
    
    def Toggle(self, userid=0):
        '''
        Toggle the target between moving and stopped.
        '''
        self.fire('Toggle', userid=userid)


class func_illusionary(_Targetname, _Parentname, _RenderFields, _Shadow):
    '''
    Legacy support. Use func_brush instead.
    '''
    _minlight = _KeyValue('') # The minimum level of ambient light that hits this brush.


class func_instance(_BaseEntity):
    '''
    An entity for placing an instance of a map file.
    You may translate and rotate this entity. You can use the replace
    keys to do parameter changes on the instance contents.
    In a $ at the beginning of a variable name.
    Then just use the $variable name inside of the instance contents on
    any value portion of a key/value pair.
    '''
    targetname = _KeyValue('') # The name that all entities will be fixed up
        # with based upon the fix up style.
    file = _KeyValue('') # This indicates a map file relative to the map's file name
    fixup_style = _KeyValue(SymbolicValue(0, {
        'Prefix': 0,
        'Postfix': 1,
        'None': 2,
    })) # Fixup style for instanced entity names.
        # Uses the 'Fix Up Name' field.
    replace01 = _KeyValue('') # This is a replacement parameter.
    replace02 = _KeyValue('') # This is a replacement parameter.
    replace03 = _KeyValue('') # This is a replacement parameter.
    replace04 = _KeyValue('') # This is a replacement parameter.
    replace05 = _KeyValue('') # This is a replacement parameter.
    replace06 = _KeyValue('') # This is a replacement parameter.
    replace07 = _KeyValue('') # This is a replacement parameter.
    replace08 = _KeyValue('') # This is a replacement parameter.
    replace09 = _KeyValue('') # This is a replacement parameter.
    replace10 = _KeyValue('') # This is a replacement parameter.
        # It goes in the form of $variable value.
        # All entities inside of that instance that have $variable
        # somewhere will be replaced with the value contents.
        # Example: $color 255 0 0


class func_instance_parms(_BaseEntity):
    '''
    Place one copy of this entity inside of an instance.
    Whenever you add a $parameter for the instance, get the properties
    of this entity. It will auto-populate it with the variables
    and allow you to indicate the variable type.
    '''
    parm1 = _KeyValue('') # This is a parameter.
    parm2 = _KeyValue('') # This is a parameter.
    parm3 = _KeyValue('') # This is a parameter.
    parm4 = _KeyValue('') # This is a parameter.
    parm5 = _KeyValue('') # This is a parameter.
    parm6 = _KeyValue('') # This is a parameter.
    parm7 = _KeyValue('') # This is a parameter.
    parm8 = _KeyValue('') # This is a parameter.
    parm9 = _KeyValue('') # This is a parameter.
    parm10 = _KeyValue('') # This is a parameter.
        # It goes in the form of $variable type.


class func_ladderendpoint(_Targetname, _Parentname):
    '''
    An entity used to specify the endpoints of a ladder.
    This entity is functional but has been replaced by the easier-to-use
    func_useableladder entity. Left in only for backwards-compatibility!
    
    To be valid, a full sized player hull traced between the start and end
    points must not be obstructed at level activation time. The angle
    determines in which direction the player leaves the ladder if the player
    presses the +jump button.
    
    Note: This entity is non-functional in Counter-Strike: Source. Use
    func_ladder instead.
    '''
    target = _KeyValue('')


class func_lod(_Targetname):
    '''
    Brush-built model that fades out over a specified distance.
    '''
    DisappearDist = _KeyValue(2000)
    Solid = _KeyValue(SymbolicValue('Solid', {
        'Solid': 0,
        'Nonsolid': 1,
    }))


class func_movelinear(_Targetname, _Parentname, _RenderFields):
    '''
    A brush entity that moves linearly along a given distance, in a given direction.
    '''
    movedir = _KeyValue(Vector((0,0,0))) # The direction the brushes will move, when told to.
    spawnflags = _KeyValue(Flags({
        'Not Solid': 8,
    }))
    startposition = _KeyValue(RangedValue(0.0, 0.0, 1.0)) # Position of brush when spawned.
        # The range is a value between 0.0 and 1.0, where 0 is the starting position
        # and 1 is the starting position + (move direction * move distance).
    speed = _KeyValue(100) # The speed that the brush moves, in inches per second.
    movedistance = _KeyValue(100.0) # The distance from the starting point that
        # the brush should move, in inches.
    blockdamage = _KeyValue(0.0) # The amount of damage to do to any entity
        # that blocks the brushes, per frame.
    startsound = _KeyValue('') # Sound played when the brush starts moving.
    stopsound = _KeyValue('') # Sound played when the brush stops moving.
    
    def Open(self, userid=0):
        '''
        Move the brush to the end position
        (starting position + (move direction * move distance)).
        '''
        self.fire('Open', userid=userid)
    
    def Close(self, userid=0):
        '''
        Move the brush to the starting position.
        '''
        self.fire('Close', userid=userid)
    
    def SetPosition(self, distance, userid=0):
        '''
        Move the brush to a specific position between 0.0 and 1.0,
        where 0 is the starting position and 1 is the
        starting position + (move direction * move distance).
        '''
        self.fire('SetPosition', distance, userid=userid)
    
    def SetSpeed(self, speed, userid=0):
        '''
        Set the speed and update immediately.
        '''
        self.fire('SetSpeed', speed, userid=userid)


class func_occluder(_Targetname):
    '''
    A occluder brush used to manage dynamic visibility in maps.
    Occluders are used to dynamically determine what things are behind them,
    to prevent trying to draw them at all.
    '''
    StartActive = _KeyValue(SymbolicValue(1, {
        'Inactive': 0,
        'Active': 1,
    })) # Initial State
    
    def Deactivate(self, userid=0):
        '''
        Deactivate the occluder, When inactive, it can be seen through.
        '''
        self.fire('Deactivate', userid=userid)
    
    def Activate(self, userid=0):
        '''
        Activate the occluder. When active, it cannot be seen through.
        '''
        self.fire('Activate', userid=userid)
    
    def Toggle(self, userid=0):
        '''
        Toggle the active/inactive state of the occluder.
        '''
        self.fire('Toggle', userid=userid)


class func_physbox(_BreakableBrush, _Targetname, _RenderFields, _Shadow):
    '''
    A brush entity that's physically simulated.
    '''
    _minlight = _KeyValue('') # The minimum level of ambient light that hits this brush.
    spawnflags = _KeyValue(Flags({
        'Start Asleep': 4096,
        'Ignore +USE for Pickup': 8192,
        "Debris - Don't collide with the player or other debris": 16384,
        'Motion Disabled': 32768,
        'Use Preferred Carry Angles': 65536,
        'Enable motion on Physcannon grab': 131072,
        'Not affected by rotor wash': 262144,
        'Generate output on +USE ': 524288,
        'Physgun can ALWAYS pick up. No matter what.': 1048576,
        'Physgun is NOT allowed to pick this up.': 2097152,
        'Physgun is NOT allowed to punt this object.': 4194304,
        'Prevent motion enable on player bump': 8388608,
    }, 'Generate output on +USE '))
    Damagetype = _KeyValue(SymbolicValue(0, {
        'Blunt': 0,
        'Sharp': 1,
    })) # Impact Damage Type
    massScale = _KeyValue(0.0) # A scale multiplier for the object's mass.
    overridescript = _KeyValue('') # A list of physics key/value pairs that are
        # usually in a physics prop .qc file.
        # Format is 'key,value,key,value,etc'.
    damagetoenablemotion = _KeyValue(0) # If specified, this object will start
        # motion disabled. Once its health has dropped below this specified
        # amount, it will enable motion.
    forcetoenablemotion = _KeyValue(0.0) # If specified, this object will start
        # motion disabled. Any impact that imparts a force greater than
        # this value on the physbox will enable motion.
    health = _KeyValue(0) # Number of points of damage to take before breaking.
        # 0 means don't break.
    preferredcarryangles = _KeyValue(Vector((0,0,0))) # If the
        # 'Use Preferred Carry Angles' spawnflag is set, this angle is
        # the angle which the object should orient to when the player
        # picks it up, with the physgun or +USE.
    notsolid = _KeyValue(SymbolicValue(0, {
        'Solid to World': 0,
        'Passes through World': 1,
    })) # Not solid to world
    
    def Wake(self, userid=0):
        '''
        Wake up this physics object, if it is sleeping.
        '''
        self.fire('Wake', userid=userid)
    
    def Sleep(self, userid=0):
        '''
        Put this physics object to sleep. It will wake if given the Wake input,
        or if force is applied to it. Note that physics objects go to sleep
        automatically after coming to rest for a while, so you don't really
        need to use this.
        '''
        self.fire('Sleep', userid=userid)
    
    def EnableMotion(self, userid=0):
        '''
        Enable physics motion/collision response.
        '''
        self.fire('EnableMotion', userid=userid)
    
    def DisableMotion(self, userid=0):
        '''
        Disable physics motion/collision response.
        '''
        self.fire('DisableMotion', userid=userid)
    
    def ForceDrop(self, userid=0):
        '''
        If this object is being carried by a player, with the physgun or +USE,
        force it to be dropped.
        '''
        self.fire('ForceDrop', userid=userid)


class func_platrot(_Targetname, _Parentname, _RenderFields, _BasePlat, _Shadow):
    '''
    A brush entity that moves vertically, and can rotate while doing so.
    '''
    spawnflags = _KeyValue(Flags({
        'Toggle': 1,
        'X Axis': 64,
        'Y Axis': 128,
    }, 'Toggle'))
    noise1 = _KeyValue('') # The sound to play when the brush moves.
    noise2 = _KeyValue('') # The sound to play when the brush stops moving.
    speed = _KeyValue(50) # Speed at which the brush rotates, in degrees per second.
    height = _KeyValue(0) # The vertical distance from the starting position that
        # this platform moves. If negative, the platform will lower.
    rotation = _KeyValue(0) # The amount this platform should rotate as it moves, in degrees.
    _minlight = _KeyValue('') # The minimum level of ambient light that hits this brush.


class func_precipitation(_Targetname, _Parentname):
    '''
    A brush entity that creates rain and snow inside its volume.
    '''
    renderamt = _KeyValue(5) # Density (0-100%)
    rendercolor = _KeyValue(Color(100,100,100)) # Color (R G B)
    preciptype = _KeyValue(SymbolicValue(0, {
        'Rain': 0,
        'Snow': 1,
        'Ash': 2,
        'Snowfall': 3,
    })) # Precipitation Type


class func_reflective_glass(func_brush):
    '''
    Used to produce perfectly reflective glass that renders world + entities.
    Typically, you want your glass brush to have nodraw on all non-reflective
    surfaces and you want to use a shader like lightmappedreflective in your
    material applied to the non-nodraw surfaces.
    See hl2/materials/glass/reflectiveglass001.vmt for an example.
    
    NOTE: currently, you cannot use reflective glass in scenes with water,
    and you can only have 1 reflective glass in your view frustum ( + pvs )
    at a time.
    '''
    pass


class func_rot_button(_Targetname, _Parentname, _Global, _Button, _EnableDisable):
    '''
    A brush entity that's designed to be used for a rotating player-useable button.
    When used by the player, it rotates to a pressed position.
    '''
    master = _KeyValue('') # Legacy support
    speed = _KeyValue(50) # The speed that the button rotates, in degrees per second.
    health = _KeyValue(0) # Legacy method of specifying whether or not the button
        # can be shot to activate it. Use the 'Damage Activates' spawnflag instead.
    sounds = _KeyValue(SymbolicValue(21, {
        'None (Silent)': 0,
        'Squeaky': 21,
        'Squeaky Pneumatic': 22,
        'Ratchet Groan': 23,
        'Clean Ratchet': 24,
        'Gas Clunk': 25,
    })) # Sounds
    wait = _KeyValue(3) # Amount of time, in seconds, after the button has been pressed before
        # it returns to the starting position.
        # Once it has returned, it can be used again.
        # If the value is set to -1, the button never returns.
    distance = _KeyValue(90) # The amount, in degrees, that the button
        # should rotate when it's pressed.
    spawnflags = _KeyValue(Flags({
        'Not solid': 1,
        'Reverse Dir': 2,
        'Toggle': 32,
        'X Axis': 64,
        'Y Axis': 128,
        'Touch Activates': 256,
        'Damage Activates': 512,
        'Use Activates': 1024,
        'Starts locked': 2048,
    }))
    _minlight = _KeyValue('') # The minimum level of ambient light that hits this brush.


class func_rotating(_Targetname, _Parentname, _RenderFields, _Shadow):
    '''
    A rotating brush entity.
    '''
    maxspeed = _KeyValue(100) # The maximum rotation speed of the brushes, in degrees per second.
    fanfriction = _KeyValue(RangedValue(20, 0, 100)) # The amount of rotational friction.
        # Value must be between 0 and 100 %.
    message = _KeyValue('') # Sound to play while rotating.
    volume = _KeyValue(10) # The volume of the rotation sound.
    spawnflags = _KeyValue(Flags({
        'Start ON': 1,
        'Reverse Direction': 2,
        'X Axis': 4,
        'Y Axis': 8,
        'Acc/Dcc': 16,
        'Fan Pain': 32,
        'Not Solid': 64,
        'Small Sound Radius': 128,
        'Medium Sound Radius': 256,
        'Large Sound Radius': 512,
    }, 'Large Sound Radius'))
    _minlight = _KeyValue('') # The minimum level of ambient light that hits this brush.
    dmg = _KeyValue(0) # Damage done to any entity that blocks the rotation, per frame.
    solidbsp = _KeyValue(SymbolicValue(0, {
        'VPhysics': 0,
        'BSP': 1,
    })) # Solid Type
    
    def SetSpeed(self, speed, userid=0):
        '''
        Set the speed as a ratio of the specified Max Rotation Speed,
        where 0 is stopped and 1 is the Max Rotation Speed..
        '''
        self.fire('SetSpeed', speed, userid=userid)
    
    def Start(self, userid=0):
        '''
        Start the rotator rotating.
        '''
        self.fire('Start', userid=userid)
    
    def Stop(self, userid=0):
        '''
        Stop the rotator from rotating.
        '''
        self.fire('Stop', userid=userid)
    
    def StopAtStartPos(self, userid=0):
        '''
        Stop the rotator from rotating when it gets around to the start position again
        (on its rotation axis).
        '''
        self.fire('StopAtStartPos', userid=userid)
    
    def StartForward(self, userid=0):
        '''
        Start the rotator rotating forward.
        '''
        self.fire('StartForward', userid=userid)
    
    def StartBackward(self, userid=0):
        '''
        Start the rotator rotating backward.
        '''
        self.fire('StartBackward', userid=userid)
    
    def Toggle(self, userid=0):
        '''
        Toggle the rotator between rotating and not rotating.
        '''
        self.fire('Toggle', userid=userid)
    
    def Reverse(self, userid=0):
        '''
        Reverse the direction of rotation of the rotator.
        '''
        self.fire('Reverse', userid=userid)


class func_smokevolume(_Targetname):
    '''
    A brush entity that spawns smoke particles within its volume.
    '''
    spawnflags = _KeyValue(Flags({
        'Emissive': 1,
    }))
    Color1 = _KeyValue(Color(255,255,255)) # Particle Color1 (R G B)
    Color2 = _KeyValue(Color(255,255,255)) # Particle Color2 (R G B)
    material = _KeyValue('particle/particle_smokegrenade') # The material
        # to use for the particles
    ParticleDrawWidth = _KeyValue(120.0) # The size of the particles, in units/inches.
    ParticleSpacingDistance = _KeyValue(80.0) # The distance between the particles inside
        # the volume. The lower the number, the denser the particles,
        # and the more overdraw there will be.
        # It is best to keep it as high as you can without it looking bad.
    DensityRampSpeed = _KeyValue(1.0) # Time to go from density 0 to density 1, in seconds.
    RotationSpeed = _KeyValue(10.0), # The speed that the particles should rotate
        # in degrees per second.
    MovementSpeed = _KeyValue(10.0), # The speed that the particles should move around
        # in units/inches per second.
    Density = _KeyValue(RangedValue(1.0, 0.0, 1.0)) # Density [0..1]
    
    def SetRotationSpeed(self, speed, userid=0):
        '''
        Set the particle rotation speed (in degrees per second).
        '''
        self.fire('SetRotationSpeed', speed, userid=userid)
    
    def SetMovementSpeed(self, speed, userid=0):
        '''
        Set the particle movement speed (in inches per second).
        '''
        self.fire('SetMovementSpeed', speed, userid=userid)
    
    def SetDensity(self, density, userid=0):
        '''
        Set the particle density. It should be a range from 0 to 1.
        '''
        self.fire('SetDensity', density, userid=userid)


class func_tanktrain(_BaseTrain):
    '''
    A moving train that follows a path of path_track entities,
    shoots at the player, and can be killed.
    
    NOTE: Build your train so that the front of the train is facing down
    the X axis. When it spawns it will automatically rotate to face
    the next path_track on the path.
    '''
    health = _KeyValue(100) # Health


class func_trackautochange(_Trackchange):
    '''
    An entity that works as a rotating/moving platform that will carry
    a train to a new track. It must be larger in X-Y planar area
    than the train, since it must contain the train within these dimensions
    in order to operate when the train is near it.
    '''
    _minlight = _KeyValue('') # The minimum level of ambient light that hits this brush.
    
    def Trigger(self, userid=0):
        '''
        Trigger the track change.
        '''
        self.fire('Trigger', userid=userid)


class func_trackchange(_Trackchange):
    '''
    An entity that works as a rotating/moving platform that will carry
    a train to a new track. It must be larger in X-Y planar area than
    the train, since it must contain the train within these dimensions
    in order to operate when the train is near it.
    '''
    _minlight = _KeyValue('') # The minimum level of ambient light that hits this brush.


class func_tracktrain(_BaseTrain):
    '''
    A moving platform that the player can ride. It follows a path of
    path_track entities.
    
    NOTE: Build your train so that the front of the train is facing down
    the X axis. When it spawns it will automatically rotate to face the
    next path_track on the path.
    '''
    ManualSpeedChanges = _KeyValue(SymbolicValue(0, {
        'Off': 0,
        'On': 1,
    })) # Train Speed is controlled through IO, handles accel, decel times.
    ManualAccelSpeed = _KeyValue(0.0) # Units per second to accelerate to target speed.
    ManualDecelSpeed = _KeyValue(0.0) # Units per second to decelerate to target speed.
    
    def SetSpeedDirAccel(self, accel_scale, userid=0):
        '''
        Accel/Decel to the specified speed, as a ratio of max speed.
        Negative values reverse the direction [-1, 1]
        '''
        accel = RangedValue(accel_scale, -1.0, 1.0)
        self.fire('SetSpeedDirAccel', accel, userid=userid)
    
    def TeleportToPathTrack(self, targetname, userid=0):
        '''
        Teleport train to the designated path track.
        '''
        self.fire('TeleportToPathTrack', targetname, userid=userid)
    
    def SetSpeedForwardModifier(self, modifier, userid=0):
        '''
        Apply the given modifier to all forward speeds. [0, 1]
        '''
        mod = RangedValue(modifier, 0.0, 1.0)
        self.fire('SetSpeedForwardModifier', mod, userid=userid)


class func_traincontrols(_Parentname, _Global):
    '''
    When used by the player, this entity overrides the player's
    controls to let them drive a train.
    '''
    target = _KeyValue('') # The target train to control when the player uses these controls.


class func_useableladder(_Targetname, _Parentname):
    '''
    Ladder that handles player auto mount/unmount as well as +use to get
    onto the ladder.
    
    See also info_ladder_dismount, used to specify auto-dismount points.
    
    Note: This entity is non-functional in Counter-Strike: Source.
    Use func_ladder instead.
    '''
    StartDisabled = _KeyValue(SymbolicBoolean(0))
    point0 = _KeyValue(Vector(0,0,0))
    point1 = _KeyValue(Vector(0,0,0))
    ladderSurfaceProperties = _KeyValue('')
    spawnflags = _KeyValue(Flags({
        "Fake Ladder": 1,
    }))
 
    def Enable(self, userid=0):
        '''
        Enable this ladder.
        '''
        self.fire('Enable', userid=userid)
 
    def Disable(self, userid=0):
        '''
        Disable this ladder.
        '''
        self.fire('Disable', userid=userid)


class func_viscluster(_BaseEntity):
    '''
    Any leaves touching this brush will have their vis merged together
    into a single cluster.
    Use multiple func_viscluster entities to reduce vis time
    '''
    pass


class func_wall(_Targetname, _RenderFields, _Global, _Shadow):
    '''
    Legacy support. Use func_brush instead.
    '''
    _minlight = _KeyValue('')


class func_wall_toggle(func_wall):
    '''
    A brush entity that can be toggled on/off.
    When off, the brush will be non-solid and invisible.
    Does not cast lightmap shadows.
    '''
    spawnflags = _KeyValue(Flags({
        'Starts Invisible': 1,
    }))
    
    def Toggle(self, userid=0):
        '''
        Toggle the brush on/off.
        When off, the brush will be non-solid and invisible.
        '''
        self.fire('Toggle', userid=userid)


# FIXME: Why Commented out from FGD?
class func_water(_Door):
    '''
    Liquid
    '''
    spawnflags = _KeyValue(Flags({
        'Starts Open': 1,
        'Use Only': 256,
    }))
    movedir = _KeyValue(Vector((0,0,0))) # The direction the water will move
        # when it is told to 'Open'.
    WaveHeight = _KeyValue('3.0') # Wave Height


class func_water_analog(_Targetname, _Parentname):
    '''
    A water brush entity that moves linearly along a given distance, in a given direction
    '''
    movedir = _KeyValue(Vector((0,0,0))) # The direction the water will move, when told to 'Open'.
    startposition = _KeyValue(RangedValue(0.0, 0.0, 1.0)) # Position of the
        # water brush when spawned.
        # The range is a value between 0.0 and 1.0, where 0 is the starting position
        # and 1 is the starting position + (move direction * move distance).
    speed = _KeyValue(100) # The speed that the water brush moves, in inches per second.
    movedistance = _KeyValue(100.0) # The distance from the starting point that
        # the water brush should move, in inches.
    startsound = _KeyValue('') # Sound played when the water brush starts moving.
    stopsound = _KeyValue('') # Sound played when the water brush stops moving.
    WaveHeight = _KeyValue(3.0) # Wave Height
    
    def Open(self, userid=0):
        '''
        Move the water brush to the end position
        (starting position + (move direction * move distance)).
        '''
        self.fire('Open', userid=userid)
    
    def Close(self, userid=0):
        '''
        Move the water brush to the starting position.
        '''
        self.fire('Close', userid=userid)
    
    def SetPosition(self, distance, userid=0):
        '''
        Move the water brush to a specific position between 0.0 and 1.0,
        where 0 is the starting position and 1 is the
        starting position + (move direction * move distance).
        '''
        self.fire('SetPosition', distance, userid=userid)


class game_end(_Targetname):
    '''
    An entity that ends a multiplayer game.
    '''
    master = _KeyValue('') # Legacy support
    
    def EndGame(self, userid=0):
        '''
        End the multiplayer game.
        '''
        self.fire('EndGame', userid=userid)


class game_gib_manager(_Targetname):
    '''
    An entity to control the number of gibs in the world, for
    performance reasons.
    '''
    maxpieces = _KeyValue(-1)
    maxpiecesdx8 = _KeyValue(-1)
    allownewgibs = _KeyValue(SymbolicBoolean(0))
    
    def SetMaxPieces(self, count, userid=0):
        '''Set the max gib Count.'''
        self.fire('SetMaxPieces', count, userid=userid)
    
    def SetMaxPiecesDX8(self, count, userid=0):
        '''Set the max gib count under DX8.'''
        self.fire('SetMaxPiecesDX8', count, userid=userid)


class game_player_equip(_Targetname, _UnknownKeys):
    '''
    An entity that gives equipment to the player who activates it.
    To use, add new keys to this entity, where each key is the classname of a
    weapon/item, and the corresponding value is the number of those
    weapons/items to give to the player who uses this entity.
    If the 'Use Only' spawnflag isn't set, then players can just
    touch this entity to get the equipment.
    '''
    spawnflags = _KeyValue(Flags({
        'Use Only': 1,
    }))
    master = _KeyValue('') # Legacy support


class game_player_team(_Targetname):
    '''
    An entity that changes the team of the player who activates it.
    '''
    spawnflags = _KeyValue(Flags({
        'Remove On fire': 1,
        'Kill Player': 2,
        'Gib Player': 4,
    }))
    target = _KeyValue('') # game_team_master to use
    master = _KeyValue('') # Legacy support


class game_ragdoll_manager(_Targetname):
    '''
    An entity to control the number of ragdolls in the world, for
    performance reasons.
    '''
    MaxRagdollCount = _KeyValue(-1)
    MaxRagdollCountDX8 = _KeyValue(-1)
    SaveImportant = _KeyValue(SymbolicBoolean(0))
    
    def SetMaxRagdollCount(self, count, userid=0):
        '''Set the Max Ragdoll Count.'''
        self.fire('SetMaxRagdollCount', count, userid=userid)
    
    def SetMaxRagdollCountDX8(self, count, userid=0):
        '''Set the Max Ragdoll Count on DX8 hardware.'''
        self.fire('SetMaxRagdollCountDX8', count, userid=userid)


class game_score(_Targetname):
    '''
    An entity that awards/deducts points from the player who activates it.
    '''
    spawnflags = _KeyValue(Flags({
        'Allow Negative': 1,
        'Team Points': 2,
    }))
    points = _KeyValue(1) # Points to add (+/-)
    master = _KeyValue('') # Legacy support
    
    def ApplyScore(self, userid=0):
        '''
        Add score to player.
        '''
        self.fire('ApplyScore', userid=userid)


class game_text(_Targetname):
    '''
    An entity that displays text on player's screens.
    '''
    spawnflags = _KeyValue(Flags({
        'All Players': 1,
    }))
    message = _KeyValue('') # Message to display onscreen.
    x = _KeyValue(RangedValue(-1.0, -1.0, 1.0))
        # Horizontal position on the player's screens to draw the text.
        # The value should be between 0 and 1, where 0 is the far left
        # of the screen and 1 is the far right. -1 centers the text.
    y = _KeyValue(RangedValue(-1.0, -1.0, 1.0))
        # Vertical position on the player's screens to draw the text.
        # The value should be between 0 and 1, where 0 is the top
        # of the screen and 1 is the bottom. -1 centers the text.
    effect = _KeyValue(SymbolicValue(0, {
        'Fade In/Out': 0,
        'Credits': 1,
        'Scan Out': 2,
    })) # Text Effect
    color = _KeyValue(Color(100,100,100)) # Color1
    color2 = _KeyValue(Color(240,110,0)) # Color2
    fadein = _KeyValue(1.5) # The time it should take for the text to fully fade in.
    fadeout = _KeyValue(0.5), # The time it should take for the text to fade out
        # after the hold time has expired.
    holdtime = _KeyValue(1.2), # The time the text should stay onscreen, after fading in
        # before it begins to fade out.
    fxtime = _KeyValue(0.25), # If the 'Text Effect' is set to Scan Out
        # this is the time it should take to scan out all the letters in the text.
    channel = _KeyValue(SymbolicValue(1, {
        'Channel 1': 1,
        'Channel 2': 2,
        'Channel 3': 3,
        'Channel 4': 4,
    })) # You can have up to four individual game_text messages onscreen at once,
        # stored in channels. Select which channel this text should be placed in,
        # which will overwrite any active message already in that channel.
    master = _KeyValue('') # Legacy support
    
    def Display(self, userid=0):
        '''
        Display the message text.
        '''
        self.fire('Display', userid=userid)


class game_ui(_Targetname):
    '''
    An entity used to override player input when the player is looking at it.
    '''
    spawnflags = _KeyValue(Flags({
        'Freeze Player': 32,
        'Hide Weapon': 64,
        '+Use Deactivates': 128,
        'Jump Deactivates': 256,
    }, 'Freeze Player', 'Hide Weapon', '+Use Deactivates', 'Jump Deactivates'))
    FieldOfView = _KeyValue(RangedValue(-1.0, -1.0, 1.0))
        # The amount of tolerance in the view checking
        # when determining whether the player's input is still under control.
        # 1.0 = straight ahead, 0.0 = +/- 90 degrees, -1.0 = all directions.
        # If the player isn't within the tolerance, the player regains control.
    
    def Deactivate(self, userid=0):
        '''
        Return Player Control.
        '''
        self.fire('Deactivate', userid=userid)
    
    # FGD says that Activate takes a string parameter, I believe not
    def Activate(self, userid=0):
        '''
        Take Player Control.
        '''
        self.fire('Activate', userid=userid)


class game_weapon_manager(_Targetname):
    '''
    An entity used to limit the number of a particular weapon type in the world.
    Useful in places where NPCs are spawning rapidly, dying, and dropping weapons.
    '''
    weaponname = _KeyValue('') # Classname of the weapon type to limit.
    maxpieces = _KeyValue(0) # The maximum amount of the specified weapon
        # type allowed in the world.
    ammomod = _KeyValue(1.0) # Modifier for ammount of ammo dropped by a weapon.
    
    def SetAmmoModifier(self, ammomod, userid=0):
        '''
        Adjust the ammo modifier.
        '''
        self.fire('SetAmmoModifier', ammomod, userid=userid)


class game_zone_player(_Targetname, _Parentname):
    '''
    An entity used to count the number of players within a zone.
    '''
    def CountPlayersInZone(self, userid=0):
        '''
        Count the number of players in the zone, and fire the corresponding outputs.
        '''
        self.fire('CountPlayersInZone', userid=userid)


class gibshooter(_gibshooterbase):
    '''
    An entity that shoots out gibs. Style of body part depends on language type.
    '''
    pass


class hammer_updateignorelist(_Targetname):
    '''
    Specifiy entities that are to be ignored by the
    hammer_update_safe_entities console command.
    Enter the targetnames of entities that you want to exclude into
    the list of fields here. Several of these may exist in a map.
    '''
    IgnoredName01 = _KeyValue('')
    IgnoredName02 = _KeyValue('')
    IgnoredName03 = _KeyValue('')
    IgnoredName04 = _KeyValue('')
    IgnoredName05 = _KeyValue('')
    IgnoredName06 = _KeyValue('')
    IgnoredName07 = _KeyValue('')
    IgnoredName08 = _KeyValue('')
    IgnoredName09 = _KeyValue('')
    IgnoredName10 = _KeyValue('')
    IgnoredName11 = _KeyValue('')
    IgnoredName12 = _KeyValue('')
    IgnoredName13 = _KeyValue('')
    IgnoredName14 = _KeyValue('')
    IgnoredName15 = _KeyValue('')
    IgnoredName16 = _KeyValue('')


class info_camera_link(_Targetname):
    '''
    An entity that can use point_cameras to render images for materials
    used by entities. To author the material, use the special identifier
    _rt_Camera for the $baseTexture (or whatever texture you want, like
    envmap, etc.) in the .vmt then connect the 'target' field to the entity
    which uses that material, and the 'PointCamera' field to the point_camera
    you want to have appear on that entity's material
    '''
    target = _KeyValue('') # Entity Whose Material Uses _rt_camera
    PointCamera = _KeyValue('') # The name of a point_camera entity in the map
        # that the material should be rendered from.
    
    def SetCamera(self, targetname, userid=0):
        '''
        Set the camera to use. The parameter should be the name of
        a point_camera entity in the map.
        '''
        self.fire('SetCamera', targetname, userid=userid)


class info_constraint_anchor(_Targetname, _Parentname):
    '''
    An entity used to attach constraints to a local position on an entity.
    Usually constraints will attach to the center of mass of an object.
    Attach the desired constraint to this entity, and then parent this
    entity to the entity you want the constraint to apply to.
    '''
    massScale = _KeyValue(1.0) # Amount to scale the mass of this body in the constraint solver


class info_hint(_Targetname, _HintNode):
    '''
    A hint that is not used for navigation.
    They don't go into the nodegraph, nor do they fall to the ground.
    Use these to provide some spatial context for NPCs, such as
    'look here if you can't find the player' or 'throw rocks at this spot'.
    '''
    pass


class info_intermission(_BaseEntity):
    '''
    An entity that defines an intermission spot where dead players will
    float until they respawn.
    '''
    target = _KeyValue('') # Name of entity that dead players will face while
        # in intermission at this spot.


class info_ladder_dismount(_Parentname):
    '''
    An entity to handle endpoints for multiple ladders that are too
    close to each other.
    '''
    target = _KeyValue('')


class info_landmark(_Targetname):
    '''
    An entity that acts as a landmark for transitions to another level.
    There should be a corresponding info_landmark entity in the next map.
    Entities will be transitioned to the next level relative
    to the info_landmark entities.
    '''
    pass


class info_lighting(_Targetname):
    '''
    An entity that can be used to change the lighting origin of a prop_static.
    Set the prop_static's Lighting Origin to point at this entity to make
    the prop_static light as if it was at the info_lighting's origin.
    Good for prop_static entities that are embedded in world geometry (like rocks/windows/etc).
    '''
    pass


class info_mass_center(_BaseEntity):
    '''
    An entity that overrides the mass center of the target physics prop,
    or func_physbox, by moving it to the info_mass_center's location.
    '''
    target = _KeyValue('') # The entity whose mass center will be overridden.


class info_no_dynamic_shadow(object):
    '''
    Use this entity to mark surfaces that shouldn't receive dynamic shadows.
    Useful to apply to walls and floors where shadows are drawn improperly,
    giving away the location of enemies.
    '''
    sides = _KeyValue('') # Brush faces


class info_node(_Node):
    '''
    A navigation node for ground moving NPCs.
    Navigation nodes are baked into the nodegraph so that NPCs can move to them.
    Ground nodes fall to the ground when they spawn.
    '''
    spawnflags = _KeyValue(Flags({
        'Force human permission': 1,
        'Force small_centered permission': 2,
        'Force wide_human permission': 4,
        'Force tiny permissiont': 8,
        'Force wide_short permission': 16,
        'Force medium permission': 32,
        'Force tiny_centered permission': 64,
        'Force large permission': 128,
        'Force large_centered permission': 256,
        'Keep editor position': 512,
    }))


class info_node_air(_Node):
    '''
    A navigation node for flying NPCs.
    Air navigation nodes are baked into the nodegraph so that NPCs can move
    to them. Air nodes do not fall to the ground when they spawn.
    '''
    nodeheight = _KeyValue(0) # NodeHeight


class info_node_air_hint(_Targetname, _HintNode):
    '''
    A navigation node for flying NPCs that includes some context information
    for NPCs that are interested in it. The hint might indicate a window that
    could be looked into, or an item of interest that could be commented on.
    Many hint nodes are NPC-specific, so it's helpful to use naming conventions
    like 'Crow: Fly to point' in the hint choices list.
    The angles of a hint node indicate what direction the NPC should
    face to perform the hint behavior.
    '''
    nodeheight = _KeyValue(0) # NodeHeight


class info_node_climb(_Targetname, _HintNode):
    '''
    A climb-node for AI navigation. Only usable by NPCs that can climb.
    '''
    pass


class info_node_hint(_Targetname, _HintNode):
    '''
    A navigation node for ground moving NPCs that includes some
    context information for NPCs that are interested in it.
    The hint might indicate a window that could be looked out of,
    or an item of interest that could be commented on.
    Many hint nodes are NPC-specific, so it's helpful to use naming conventions
    like 'Crow: Fly to point' in the hint choices list.
    The angles of a hint node indicate what direction the NPC should face
    to perform the hint behavior.
    
    It's important to understand the distinction between scripts,
    such as scripted_sequence and scripted_schedule, and info_hint entities.
    Scripts summon NPCs to specific cue points to play their parts,
    while hints provide context information to the AI that they use to
    perform their behaviors. Hints require code support in the NPC,
    while scripts are generic and may require only animations to play.
    Use a hint if the behavior is driven by the AI, use a script if
    the behavior is driven by the map.
    '''
    pass


class info_node_link(_Targetname):
    '''
    A dynamic connection between two navigation nodes.
    You specify the node IDs of the start and end nodes,
    and then you can use entity I/O to turn on and off the connection.
    This could be used to create or destroy a connection in the nodegraph
    because of some event in your map (a bridge being created/destroyed, etc).
    '''
    StartNode = _KeyValue('') # The node ID of one end of the node connection.
    EndNode = _KeyValue('') # The node ID of one end of the node connection.
    initialstate = _KeyValue(SymbolicValue(1, {
        'Off': 0,
        'On': 1,
    })) # Initial State
    linktype = _KeyValue(SymbolicValue(1, {
        'Ground': 1,
        'Jump': 2,
        'Fly': 4,
        'Climb': 8,
    })) # Type of Connection
    AllowUse = _KeyValue('') # Entity or class to allow passage even when node is off
    InvertAllow = _KeyValue(SymbolicValue(0, {
        'No': 0,
        'Yes': 1,
    })) # Allowed entity is the only entity NOT allowed when this is set to 'yes'
    spawnflags = _KeyValue(Flags({
        'Force human connect': 1,
        'Force small_centered connect': 2,
        'Force wide_human connect': 4,
        'Force tiny connect': 8,
        'Force wide_short connect': 16,
        'Force medium connect': 32,
        'Force tiny_centered connect': 64,
        'Force large connect': 128,
        'Force large_centered connect': 256,
        'Force medium_tall connect': 512,
    }))
    
    def TurnOn(self, userid=0):
        '''
        Turn the link on.
        '''
        self.fire('TurnOn', userid=userid)
    
    def TurnOff(self, userid=0):
        '''
        Turn the link off.
        '''
        self.fire('TurnOff', userid=userid)


class info_node_link_controller(_Targetname):
    '''
    An entity that controls all connections between nodes that intersect the
    controller's volume. This allows for mass enabling/disabling of all node
    connections through a volume.
    '''
    mins = _KeyValue(Vector((-8,-32,-36))) # Mins
    maxs = _KeyValue(Vector((8,32,36))) # Maxs
    initialstate = _KeyValue(SymbolicValue(1, {
        'Off': 0,
        'On': 1,
    })) # Initial State
    useairlinkradius = _KeyValue(SymbolicValue(0, {
        'No': 0,
        'Yes': 1,
    })) # Set this to 'Yes' if this controller is intended to control air links.
        # Air links connect using a larger search radius so leaving this at 'No'
        # might miss some air links.
    AllowUse = _KeyValue('') # Entity or class to allow passage even when node is off
    InvertAllow = _KeyValue(SymbolicValue(0, {
        'No': 0,
        'Yes': 1,
    })) # Allowed entity is the only entity NOT allowed when this is set to 'yes'
    
    def TurnOn(self, userid=0):
        '''
        Turn the link on.
        '''
        self.fire('TurnOn', userid=userid)
    
    def TurnOff(self, userid=0):
        '''
        Turn the link off.
        '''
        self.fire('TurnOff', userid=userid)
    
    def SetAllowed(self, allowed, userid=0):
        '''
        Change the allowed pass when off
        '''
        self.fire('SetAllowed', allowed, userid=userid)
    
    def SetInvert(self, invertallow, userid=0):
        '''
        Change the invert exclusion rule
        '''
        self.fire('SetInvert', invertallow, userid=userid)


class info_npc_spawn_destination(_Targetname, _Parentname):
    '''
    NPC Spawn Destination. See npc_template_maker.
    '''
    ReuseDelay = _KeyValue(1.0)
    RenameNPC = _KeyValue('')
 

class info_null(_Targetname):
    '''
    An entity that's immediately removed on spawning. 
    Useful as a spotlight target.
    '''
    pass


class info_overlay(_Targetname):
    '''
    An entity that places an overlay on the world.
    '''
    material = _KeyValue('') # Material
    sides = _KeyValue('') # Brush faces
    RenderOrder = _KeyValue(RangedValue(0, 0, 3))
        # Higher values render after lower values. This value can be 0-3.
    StartU = _KeyValue(0.0) # U Start
    EndU = _KeyValue(1.0) # U End
    StartV = _KeyValue(0.0) # V Start
    EndV = _KeyValue(1.0) # V End
    fademindist = _KeyValue(-1.0) # Distance at which the overlay starts to fade
        # (<0 = use fademaxdist).
    fademaxdist = _KeyValue(0.0) # Maximum distance at which the overlay is visible
        # (0 = don't fade out).

    BasisOrigin = _KeyValue(Vector((0,0,0)), readonly=True)
    BasisU = _KeyValue(Vector((0,0,0)), readonly=True)
    BasisV = _KeyValue(Vector((0,0,0)), readonly=True)
    BasisNormal = _KeyValue(Vector((0,0,0)), readonly=True)
    uv0 = _KeyValue(Vector((0,0,0)), readonly=True)
    uv1 = _KeyValue(Vector((0,0,0)), readonly=True)
    uv2 = _KeyValue(Vector((0,0,0)), readonly=True)
    uv3 = _KeyValue(Vector((0,0,0)), readonly=True)


class info_overlay_transition(_BaseEntity):
    '''
    Overlay Transition
    '''
    material = _KeyValue('') # Material
    sides = _KeyValue('') # Brush faces
    sides2 = _KeyValue('') # Water faces
    LengthTexcoordStart = _KeyValue(0.0) # Texcoord Length Start
    LengthTexcoordEnd = _KeyValue(1.0) # Texcoord Length End
    WidthTexcoordStart = _KeyValue(0.0) # Texcoord Width Start
    WidthTexcoordEnd = _KeyValue(1.0) # Texcoord Width End
    Width1 = _KeyValue(25.0) # Width Land
    Width2 = _KeyValue(25.0) # Width Water
    DebugDraw = _KeyValue(SymbolicBoolean(0)) # Boolean value (0 or 1).


class info_particle_system(_Targetname, _Parentname):
    '''
    An entity that spawns a particle system built using the particle editor.
    '''
    effect_name = _KeyValue('') # Particle System Name
    start_active = _KeyValue(SymbolicValue(0, {
        'No': 0,
        'Yes': 1,
    })) # Start Active?
    flag_as_weather = _KeyValue(SymbolicValue(0, {
        'No': 0,
        'Yes': 1,
    })) # Is this particle system going to be used as a weather effect?
    # Control points: if set, cpoint1(2,3,...) of the effect will be
    # at this entity's location.
    # If previous control point is not set, value is ignored
    cpoint1 = _KeyValue('') # targetname
    cpoint2 = _KeyValue('')
    cpoint3 = _KeyValue('')
    cpoint4 = _KeyValue('')
    cpoint5 = _KeyValue('')
    cpoint6 = _KeyValue('')
    cpoint7 = _KeyValue('')
    cpoint8 = _KeyValue('')
    cpoint9 = _KeyValue('')
    cpoint10 = _KeyValue('')
    cpoint11 = _KeyValue('')
    cpoint12 = _KeyValue('')
    cpoint13 = _KeyValue('')
    cpoint14 = _KeyValue('')
    cpoint15 = _KeyValue('')
    cpoint16 = _KeyValue('')
    cpoint17 = _KeyValue('')
    cpoint18 = _KeyValue('')
    cpoint19 = _KeyValue('')
    cpoint20 = _KeyValue('')
    cpoint21 = _KeyValue('')
    cpoint22 = _KeyValue('')
    cpoint23 = _KeyValue('')
    cpoint24 = _KeyValue('')
    cpoint25 = _KeyValue('')
    cpoint26 = _KeyValue('')
    cpoint27 = _KeyValue('')
    cpoint28 = _KeyValue('')
    cpoint29 = _KeyValue('')
    cpoint30 = _KeyValue('')
    cpoint31 = _KeyValue('')
    cpoint32 = _KeyValue('')
    cpoint33 = _KeyValue('')
    cpoint34 = _KeyValue('')
    cpoint35 = _KeyValue('')
    cpoint36 = _KeyValue('')
    cpoint37 = _KeyValue('')
    cpoint38 = _KeyValue('')
    cpoint39 = _KeyValue('')
    cpoint40 = _KeyValue('')
    cpoint41 = _KeyValue('')
    cpoint42 = _KeyValue('')
    cpoint43 = _KeyValue('')
    cpoint44 = _KeyValue('')
    cpoint45 = _KeyValue('')
    cpoint46 = _KeyValue('')
    cpoint47 = _KeyValue('')
    cpoint48 = _KeyValue('')
    cpoint49 = _KeyValue('')
    cpoint50 = _KeyValue('')
    cpoint51 = _KeyValue('')
    cpoint52 = _KeyValue('')
    cpoint53 = _KeyValue('')
    cpoint54 = _KeyValue('')
    cpoint55 = _KeyValue('')
    cpoint56 = _KeyValue('')
    cpoint57 = _KeyValue('')
    cpoint58 = _KeyValue('')
    cpoint59 = _KeyValue('')
    cpoint60 = _KeyValue('')
    cpoint61 = _KeyValue('')
    cpoint62 = _KeyValue('')
    cpoint63 = _KeyValue('')
    # Control point parents: If set and nonzero, control point # of the effect will use
    # this point for its parent
    cpoint1_parent = _KeyValue(0)
    cpoint2_parent = _KeyValue(0)
    cpoint3_parent = _KeyValue(0)
    cpoint4_parent = _KeyValue(0)
    cpoint5_parent = _KeyValue(0)
    cpoint6_parent = _KeyValue(0)
    cpoint7_parent = _KeyValue(0)
    
    def Start(self, userid=0):
        '''
        Tell the particle system to start emitting.
        '''
        self.fire('Start', userid=userid)
    
    def Stop(self, userid=0):
        '''
        Tell the particle system to stop emitting.
        '''
        self.fire('Stop', userid=userid)


class info_player_start(_PlayerClass):
    '''
    This entity indicates the position and facing direction at which the
    player will spawn. Any number of info_player_start entities may be
    placed in a map for when working in cordoned-off portions of the map.
    When multiple info_player_start entities are present in a map, set
    the 'Master' spawnflag on one of them to indicate which one should
    be used when running the entire map.
    '''
    spawnflags = _KeyValue(Flags({
        'Master': 1,
    }))


class info_projecteddecal(_Targetname):
    '''
    An entity that projects a decal onto the world (or props).
    If the decal has no target name, it will immediately apply itself when 
    the level is loaded. If it has a name specified, it won't apply until
    it receives the 'Activate' input.
    
    A decal to be applied to a prop or the world using specified origin
    and orientation and radius, uses an orientation angle and a radius to
    determine ray to cast for projection.
    '''
    texture = _KeyValue('')
    Distance = _KeyValue(64.0) # Distance from the origin to project the decal.
    
    def Activate(self, userid=0):
        '''
        Force the decal to apply itself to the world.
        '''
        self.fire('Activate', userid=userid)


class info_radial_link_controller(_Targetname, _Parentname):
    '''
    This entity automatically severs node connections that pass through its radius.
    If it moves, it will restore those connections.
    '''
    radius = _KeyValue(120.0) # Radius (Use Helper!)


class info_target(_Targetname, _Parentname):
    '''
    An entity that does nothing.
    Very useful as a positioning entity for other entities to refer to
    (i.e. the endpoint of an env_beam)
    '''
    spawnflags = _KeyValue(Flags({
        'Transmit to client': 1,
    }))


class info_teleport_destination(_Targetname, _Parentname, _PlayerClass):
    '''
    An entity that does nothing itself, but can be used to specify
    the destination for a trigger_teleport entity.
    OBSOLETE: An info_target can be used instead.
    '''
    pass


class infodecal(_Targetname):
    '''
    An entity that places a decal on the world.
    If the decal has no target name, it will immediately apply itself
    when the level is loaded. If it has a name specified, it won't apply
    until it receives the 'Activate' input. 
    '''
    texture = _KeyValue('')
    LowPriority = _KeyValue(SymbolicValue(0, {
        'No': 0,
        'Yes': 1,
    })) # Low Priority (can be replaced)
    
    def Activate(self, userid=0):
        '''
        Force the decal to apply itself to the world.
        '''
        self.fire('Activate', userid=userid)


class keyframe_rope(_Targetname, _Parentname, _KeyFrame, _RopeKeyFrame):
    '''
    A node entity that marks a point in a rope.
    The first node in the rope should be a move_rope, followed by 1 or more keyframe_ropes.
    '''
    pass


class keyframe_track(_Targetname, _Parentname, _KeyFrame):
    '''
    Animation KeyFrame
    '''
    pass


class light(_Targetname, _Light):
    '''
    An invisible omnidirectional lightsource.
    '''
    target = _KeyValue('') # The name of an entity in the map that the spotlight
        # will point at. This will override the spotlight's angles.
    spawnflags = _KeyValue(Flags({
        'Initially dark': 1,
    }))
    _distance = _KeyValue(0) # This is the distance that light is allowed to cast, in inches.


class light_dynamic(_Targetname, _Parentname):
    '''
    An invisible lightsource that changes in some way over time.
    '''
    target = _KeyValue('') # The name of an entity in the map that the dynamic
        # light will point at.
    _light = _KeyValue(Color(255,255,255,200)) # Light color
    brightness = KeyvalueInput(0) # Light brightness
    _inner_cone = KeyvalueInput(30) # Inner (bright) angle
    _cone = KeyvalueInput(45) # Outer (fading) angle
    pitch = _KeyValue(RangedValue(-90.0, -90.0, 90.0)) # Pitch
    distance = KeyvalueInput(120.0)
        # This is the distance that light is allowed to cast, in inches.
    spotlight_radius = KeyvalueInput(80.0)
        # This is the radius of the light, in inches,
        # at the object that it is hitting.
    style = KeyvalueInput(SymbolicValue(0, {
        'Normal': 0,
        'Flicker A': 1,
        'Slow, strong pulse': 2,
        'Candle A': 3,
        'Fast strobe': 4,
        'Gentle pulse': 5,
        'Flicker B': 6,
        'Candle B': 7,
        'Candle C': 8,
        'Slow strobe': 9,
        'Fluorescent flicker': 10,
        'Slow pulse, noblack': 11,
    })) # Appearance
    spawnflags = _KeyValue(Flags({
        'Do not light world': 1,
        'Do not light models': 2,
        'Add Displacement Alpha': 4,
        'Subtract Displacement Alpha': 8,
    }))
    
    def Color(self, color, userid=0):
        '''
        Set the light's render color (R G B).
        '''
        self.fire('Color', Color(color).for_input(), userid=userid)
    
    def TurnOn(self, userid=0):
        '''
        Turn the light off.
        '''
        self.fire('TurnOn', userid=userid)
    
    def TurnOff(self, userid=0):
        '''
        Turn the light on.
        '''
        self.fire('TurnOff', userid=userid)
    
    def Toggle(self, userid=0):
        '''
        Toggle the light on/off.
        '''
        self.fire('Toggle', userid=userid)


class light_environment(_BaseEntity):
    '''
    Set the color and angle of the light from the sun and sky.
    '''
    pitch = _KeyValue(RangedValue(0.0, -90.0, 90.0))
        # The downward pitch of the light from the sun.
        # 0 is horizontal, -90 is straight down.
    _light = _KeyValue(Color(255,255,255,200)) # Brightness
    _ambient = _KeyValue(Color(255,255,255,20)) # Ambient
    _lightHDR = _KeyValue(Color(-1,-1,-1,1)) # BrightnessHDR
    _lightscaleHDR = _KeyValue(1.0) # Amount to scale the light by when compiling for HDR.
    _ambientHDR = _KeyValue(Color(-1,-1,-1,1)) # AmbientHDR
    _AmbientScaleHDR = _KeyValue(1.0) # Amount to scale the ambient
        # light by when compiling for hdr.
    SunSpreadAngle = _KeyValue(0.0) # The angular extent of the sun for casting soft shadows.
        # Higher numbers are more diffuse. 5 is a good starting value.


class light_spot(_Targetname, _Light):
    '''
    An invisible and directional spotlight.
    '''
    target = _KeyValue('') # The name of an entity in the map that the spotlight
        # will point at. This will override the spotlight's angles.
    _inner_cone = _KeyValue(30) # Inner (bright) angle
    _cone = _KeyValue(45) # Outer (fading) angle
    _exponent = _KeyValue(1) # Focus
    _distance = _KeyValue(0) # This is the distance that light is allowed to cast, in inches.
    pitch = _KeyValue(-90) # Pitch
    spawnflags = _KeyValue(Flags({
        'Initially dark': 1,
    }))


class logic_active_autosave(_Targetname):
    '''
    An entity that is used to look for opportunities to autosave.
    '''
    MinimumHitPoints = _KeyValue(30) # Start looking for an opportunity to save
        # if player drops below this hitpoint level.
    TriggerHitPoints = _KeyValue(75), # If started looking for an opportunity
        # save when hitpoints reach this level.
    TimeToTrigget = _KeyValue(0.0) # If > 0, how long to try and get a save off before giving up
        # FIXME: should it be TimeToTrigger (Wiki agrees with typo!)?
    DangerousTime = _KeyValue(10.0) # If 0, just autosave.
        # Otherwise, do an autosavedangerous with this time threshold
    
    def Enable(self, userid=0):
        '''
        Enable the entity
        '''
        self.fire('Enable', userid=userid)
    
    def Disable(self, userid=0):
        '''
        Enable the entity
        '''
        self.fire('Disable', userid=userid)


class logic_auto(_BaseEntity):
    '''
    Fires outputs when a map spawns.
    If 'Remove on fire' flag is set the logic_auto is deleted after firing.
    It can be set to check a global state before firing.
    This allows you to only fire events based on what took place in a previous map.
    '''
    spawnflags = _KeyValue(Flags({
        'Remove on fire': 1,
    }, 'Remove on fire'))
    globalstate = _KeyValue(SymbolicValue('', {
        '--- None ---': '',
        'Antlions are player allies': 'antlion_allied',
        'Friendly encounter sequence (lower weapons, etc.)': 'friendly_encounter',
        'Gordon is invulnerable': 'gordon_invulnerable',
        'Gordon pre-criminal': 'gordon_precriminal',
        'Game is running on a console': 'is_console',
        'Game is running on a PC': 'is_pc',
        "Don't spawn seagulls on the jeep": 'no_seagulls_on_jeep',
        'Suit sprint function not yet enabled': 'suit_no_sprint',
        'Super phys gun is enabled': 'super_phys_gun',
    })) # If set, this specifies a global state to check before firing.
        # The OnMapSpawn output will only fire if the global state is set.


class logic_autosave(_Targetname):
    '''
    An entity that is used to force an autosave.
    '''
    NewLevelUnit = _KeyValue(SymbolicValue(0, {
        'No': 0,
        'Yes': 1,
    })) # If set, the save will discard any savedata from previous levels,
        # for the purpose of keeping savegame filesizes down.
        # Can only be safely used if there is no way for the player to return to
        # previous levels.
    MinimumHitPoints = _KeyValue(0) # Don't save dangerous when player has less than
        # this many hitpoints.
    MinHitPointsToCommit = _KeyValue(0) # Minimum hitpoints required to commit to save.
        # The save will be made if you have at least Minimum Hit Points,
        # but when the autosave timer expires, the autosave is only kept if
        # you have at least Min Hitpoints to Commit.
    
    def Save(self, userid=0):
        '''
        Force an autosave.
        '''
        self.fire('Save', userid=userid)
    
    def SaveDangerous(self, seconds, userid=0):
        '''
        Force an autosave as autosavedangerous.sav.
        If the player is alive after the passed number of seconds it replaces the
        standard auto save.
        '''
        self.fire('SaveDangerous', seconds, userid=userid)
    
    def SetMinHitpointsThreshold(self, hitpoints, userid=0):
        '''
        Set MinimumHitPoints to this.
        '''
        self.fire('SetMinHitpointsThreshold', hitpoints, userid=userid)


class logic_branch(_Targetname):
    '''
    Tests a boolean value and fires an output based on whether
    the value is true or false.
    Use this entity to branch between two potential sets of events.
    '''
    InitialValue = _KeyValue(SymbolicBoolean(0)) # Initial value for the boolean value (0 or 1).
    
    def SetValue(self, value, userid=0):
        '''
        Set the boolean value without performing the comparison.
        Use this to hold a value for a future test.
        '''
        self.fire('SetValue', value, userid=userid)
    
    def SetValueTest(self, value, userid=0):
        '''
        Set the boolean value and test it, firing OnTrue or OnFalse based on the new value.
        '''
        self.fire('SetValueTest', value, userid=userid)
    
    def Toggle(self, userid=0):
        '''
        Toggle the boolean value between true and false.
        '''
        self.fire('Toggle', userid=userid)
    
    def ToggleTest(self, userid=0):
        '''
        Toggle the boolean value and tests it, firing OnTrue or OnFalse based on the new value.
        '''
        self.fire('ToggleTest', userid=userid)
    
    def Test(self, userid=0):
        '''
        Test the input value and fire OnTrue or OnFalse based on the value.
        '''
        self.fire('Test', userid=userid)


class logic_branch_listener(_Targetname):
    '''
    Contains a list of logic_branch entities and fires outputs
    when the state of any of the logic_branches changes.
    
    This entity is used to fire an event when a set of conditions are all satisfied.
    '''
    Branch01 = _KeyValue('') # The name of one or more logic_branches (wildcards allowed).
    Branch02 = _KeyValue('') # The name of one or more logic_branches (wildcards allowed).
    Branch03 = _KeyValue('') # The name of one or more logic_branches (wildcards allowed).
    Branch04 = _KeyValue('') # The name of one or more logic_branches (wildcards allowed).
    Branch05 = _KeyValue('') # The name of one or more logic_branches (wildcards allowed).
    Branch06 = _KeyValue('') # The name of one or more logic_branches (wildcards allowed).
    Branch07 = _KeyValue('') # The name of one or more logic_branches (wildcards allowed).
    Branch08 = _KeyValue('') # The name of one or more logic_branches (wildcards allowed).
    Branch09 = _KeyValue('') # The name of one or more logic_branches (wildcards allowed).
    Branch10 = _KeyValue('') # The name of one or more logic_branches (wildcards allowed).
    Branch11 = _KeyValue('') # The name of one or more logic_branches (wildcards allowed).
    Branch12 = _KeyValue('') # The name of one or more logic_branches (wildcards allowed).
    Branch13 = _KeyValue('') # The name of one or more logic_branches (wildcards allowed).
    Branch14 = _KeyValue('') # The name of one or more logic_branches (wildcards allowed).
    Branch15 = _KeyValue('') # The name of one or more logic_branches (wildcards allowed).
    Branch16 = _KeyValue('') # The name of one or more logic_branches (wildcards allowed).
    
    def Test(self, userid=0):
        '''
        Tests the state of all the logic_branches in the list and fires the appropriate output.
        '''
        self.fire('Test', userid=userid)


class logic_case(_Targetname):
    '''
    Compares an input to up to 16 preset values.
    If the input value is the same as any of the preset values,
    an output corresponding to that value is fired.
    
    For example: if Case01 is set to 2 and Case02 is set to 5, and the input value is 5,
    the OnCase02 output will be fired.
    
    This entity can also be used to select from a number of random targets via the
    PickRandom input. One of the OnCase outputs that is connected to another
    entity will be picked at random and fired.
    '''
    Case01 = _KeyValue('') # Case 01
    Case02 = _KeyValue('') # Case 02
    Case03 = _KeyValue('') # Case 03
    Case04 = _KeyValue('') # Case 04
    Case05 = _KeyValue('') # Case 05
    Case06 = _KeyValue('') # Case 06
    Case07 = _KeyValue('') # Case 07
    Case08 = _KeyValue('') # Case 08
    Case09 = _KeyValue('') # Case 09
    Case10 = _KeyValue('') # Case 10
    Case11 = _KeyValue('') # Case 11
    Case12 = _KeyValue('') # Case 12
    Case13 = _KeyValue('') # Case 13
    Case14 = _KeyValue('') # Case 14
    Case15 = _KeyValue('') # Case 15
    Case16 = _KeyValue('') # Case 16
    
    def InValue(self, value, userid=0):
        '''
        Compare the Input value to the case values, and fire the appropriate output, if any.
        '''
        self.fire('InValue', value, userid=userid)
    
    def PickRandom(self, userid=0):
        '''
        Fires a random OnCase output with at least one connection.
        '''
        self.fire('PickRandom', userid=userid)
    
    def PickRandomShuffle(self, userid=0):
        '''
        Fires a random OnCase output with at least one connection,
        with no repeats until all cases have been picked,
        at which point the shuffle starts over.
        '''
        self.fire('PickRandomShuffle', userid=userid)


class logic_collision_pair(_Targetname):
    '''
    An entity that can be used to enables/disable vphysics collisions
    between two target entities.
    '''
    attach1 = _KeyValue('') # The first entity.
    attach2 = _KeyValue('') # The second entity.
    startdisabled = _KeyValue(SymbolicValue(1, {
        'No': 0,
        'Yes': 1,
    })) # Start with collisions disabled
    
    def EnableCollisions(self, userid=0):
        '''
        Enable collisions between the first and second entity.
        '''
        self.fire('EnableCollisions', userid=userid)
    
    def DisableCollisions(self, userid=0):
        '''
        Disable collisions between the first and second entity.
        '''
        self.fire('DisableCollisions', userid=userid)


class logic_compare(_Targetname):
    '''
    Compares an input value to another value.
    If the input value is less than the compare value,
    the OnLessThan output is fired with the input value.
    If the input value is equal to the compare value,
    the OnEqualTo output is fired with the input value.
    If the input value is greater than the compare value,
    the OnGreaterThan output is fired with the input value.
    '''
    InitialValue = _KeyValue(0) # Initial value for the input value.
    CompareValue = _KeyValue(0) # The value to compare against.
    
    def SetValue(self, value, userid=0):
        '''
        Set the value that will be compared against the compare value.
        '''
        self.fire('SetValue', value, userid=userid)
    
    def SetValueCompare(self, value, userid=0):
        '''
        Set the value that will be compared against the compare value
        and perform the comparison.
        '''
        self.fire('SetValueCompare', value, userid=userid)
    
    def SetCompareValue(self, value, userid=0):
        '''
        Set the compare value.
        '''
        self.fire('SetCompareValue', value, userid=userid)
    
    def Compare(self, userid=0):
        '''
        Force a compare of the input value with the compare value.
        '''
        self.fire('Compare', userid=userid)


class logic_lineto(_Targetname):
    '''
    An entity that calculates and outputs a vector from one entity to another.
    '''
    source = _KeyValue('') # Name of the entity the line should start from.
    target = _KeyValue('') # Name of the entity that line should end at.


class logic_measure_movement(_Targetname):
    '''
    An entity that can measure the movement of an entity relative
    to another entity and apply that movement to a third entity.
    '''
    MeasureTarget = _KeyValue('') # Entity whose movement you want to measure.
    MeasureReference = _KeyValue('') # The movement of Entity to Measure will be
        # measured relative to this entity.
    Target = KeyvalueInput('') # This entity will be moved to
        # mimic the motions of Entity to Measure.
    TargetReference = _KeyValue('') # The Entity to Move will move relative
        # to this entity.
    TargetScale = _KeyValue(1.0), # A scale to divide the measured movements by
        # before applying those movements to the Entity to Move.
        # 1 = target entity moves as much as the measured entity,
        # 2 = target entity moves half as far as the measured entity,
        # and 0.5 = target entity moves twice as far as the measured entity.
    MeasureType = _KeyValue(SymbolicValue(0, {
        'Position': 0,
        'Eye position': 1,
    })) # Measurement Type
    
    def SetMeasureTarget(self, targetname, userid=0):
        '''
        Set the Entity to Measure, whose movement should be measured.
        '''
        self.fire('SetMeasureTarget', targetname, userid=userid)
    
    def SetMeasureReference(self, targetname, userid=0):
        '''
        Set the Measure Reference entity.
        '''
        self.fire('SetMeasureReference', targetname, userid=userid)
    
    def SetTargetReference(self, targetname, userid=0):
        '''
        Set the Movement Reference entity.
        '''
        self.fire('SetTargetReference', targetname, userid=userid)
    
    def SetTargetScale(self, scale, userid=0):
        '''
        Set the scale to divide the measured movements by.
        '''
        self.fire('SetTargetScale', scale, userid=userid)
    
    def Enable(self, userid=0):
        '''
        Enable the logic_measure_movement.
        '''
        self.fire('Enable', userid=userid)
    
    def Disable(self, userid=0):
        '''
        Disable the logic_measure_movement.
        '''
        self.fire('Disable', userid=userid)


class logic_multicompare(_Targetname):
    '''
    Compares a set of inputs to each other.
    If they are all the same, fires an OnEqual output.
    If any are different, fires the OnNotEqual output.
    
    // NEEDHELP: Unused in HL2, not clear what it's useful for. (LOL)
    '''
    IntegerValue = _KeyValue(0) # Integer Value (optional) (eh)
    ShouldComparetoValue = _KeyValue(SymbolicValue(0, {
        'No': 0,
        'Yes': 1,
    })) # Should use Integer Value
    
    def InputValue(self, value, userid=0):
        '''
        Input value
        '''
        self.fire('InputValue', value, userid=userid)
    
    def CompareValues(self, userid=0):
        '''
        Compare the values and fires appropriate outputs
        '''
        self.fire('CompareValues', userid=userid)


class logic_navigation(_Targetname):
    '''
    An entity that is used to set navigation properties on other entities.
    Useful to make NPCs ignore physics props in their way that they can easily push.
    '''
    target = _KeyValue('') # Name of the entity to set navigation properties on.
    spawnflags = _KeyValue(Flags({
        'Start On': 1,
    }, 'Start On'))
    navprop = _KeyValue(SymbolicValue('Ignore', {
        "NPCs Ignore this when navigating": 'Ignore',
    })) # Nav Property
    
    def TurnOn(self, userid=0):
        '''
        Turn on. The Navigation Entity will have its navigation properties set.
        '''
        self.fire('TurnOn', userid=userid)
    
    def TurnOff(self, userid=0):
        '''
        Turn off. The Navigation Entity will have its navigation properties
        returned to the default settings.
        '''
        self.fire('TurnOff', userid=userid)
    
    def Toggle(self, userid=0):
        '''
        Toggle on/off.
        '''
        self.fire('Toggle', userid=userid)


class logic_playerproxy(_Targetname, _DamageFilter):
    '''
    An entity that is used to relay inputs/ouputs to the player and
    back to the world.
    '''
    def RequestPlayerHealth(self, userid=0):
        '''
        Requests the current player's health from the proxy.
        This will fire the PlayerHealth output with the value.
        '''
        self.fire('RequestPlayerHealth', userid=userid)
    
    def SetFlashlightSlowDrain(self, userid=0):
        '''
        Puts the player's flashlight in slow-power-drain mode
        (for Episodic darkness)
        '''
        self.fire('SetFlashlightSlowDrain', userid=userid)
    
    def SetFlashlightNormalDrain(self, userid=0):
        '''
        Puts the player's flashlight to default power drain
        '''
        self.fire('SetFlashlightNormalDrain', userid=userid)
    
    def SetPlayerHealth(self, health, userid=0):
        '''
        Sets the player's health to this value.
        '''
        self.fire('SetPlayerHealth', health, userid=userid)
    
    def RequestAmmoState(self, userid=0):
        '''
        Request the ammo state of the player.
        It will fire PlayerHasAmmo or PlayerHasNoAmmo outputs.
        '''
        self.fire('RequestAmmoState', userid=userid)
    
    def LowerWeapon(self, userid=0):
        '''
        Lowers the players weapon.
        '''
        self.fire('LowerWeapon', userid=userid)
    
    def EnableCappedPhysicsDamage(self, userid=0):
        '''
        Cause player to take less damage from physics objects,
        and never more than 30 points from any individual strike.
        '''
        self.fire('EnableCappedPhysicsDamage', userid=userid)
    
    def DisableCappedPhysicsDamage(self, userid=0):
        '''
        Undo effects of EnableCappedPhysicsDamage
        '''
        self.fire('DisableCappedPhysicsDamage', userid=userid)
    
    def SetLocatorTargetEntity(self, targetname, userid=0):
        '''
        Set the entity that the HUD locator should track. (Usually a vehicle)
        '''
        self.fire('SetLocatorTargetEntity', targetname, userid=userid)


class logic_relay(_Targetname, _EnableDisable):
    '''
    A message forwarder. Fires an OnTrigger output when triggered,
    and can be disabled to prevent forwarding outputs.
    
    Useful as an intermediary between one entity and another for turning
    on or off an I/O connection, or as a container for holding a set of
    outputs that can be triggered from multiple places.
    '''
    spawnflags = _KeyValue(Flags({
        'Only trigger once': 1,
        'Allow fast retrigger': 2,
    }))
    
    def Trigger(self, userid=0):
        '''
        Trigger the relay, causing its OnTrigger output to fire if it is enabled.
        '''
        self.fire('Trigger', userid=userid)
    
    def Toggle(self, userid=0):
        '''
        Toggle the relay between enabled and disabled.
        '''
        self.fire('Toggle', userid=userid)
    
    def CancelPending(self, userid=0):
        '''
        Cancel any events fired by this relay that are currently pending in the I/O event queue.
        '''
        self.fire('CancelPending', userid=userid)


class logic_timer(_Targetname, _EnableDisable):
    '''
    An entity that fires a timer event at regular, or random, intervals.
    It can also be set to oscillate betweena high and low end, in which case
    it will fire alternating high/low outputs each time it fires.
    '''
    spawnflags = _KeyValue(Flags({
        'Oscillator': 1,
    }))
    UseRandomTime = _KeyValue(SymbolicValue(0, {
        'No': 0,
        'Yes': 1,
    })) # Use Random Time
    LowerRandomBound = KeyvalueInput(0.0)
        # If 'Use Random Time' is set, this is the
        # minimum time between timer fires. The time will be a random
        # number between this and the 'UpperRandomBound'.
    UpperRandomBound = KeyvalueInput(0.0)
        # If 'Use Random Time' is set, this is the
        # maximum time between timer fires. The time will be a random
        # number between the 'LowerRandomBoundl' and this.
    RefireTime = _KeyValue(0.0) # If 'Use Random Time' isn't set, this is the time
        # between timer fires, in seconds.
    
    def RefireTime(self, seconds, userid=0):
        '''
        Set a new Refire Interval.
        '''
        self.fire('RefireTime', seconds, userid=userid)
    
    def ResetTimer(self, userid=0):
        '''
        Reset the timer. It will fire after the Refire Interval expires.
        '''
        self.fire('ResetTimer', userid=userid)
    
    def FireTimer(self, userid=0):
        '''
        Force the timer to fire immediately.
        '''
        self.fire('FireTimer', userid=userid)
    
    def Enable(self, userid=0):
        '''
        Enable the timer.
        '''
        self.fire('Enable', userid=userid)
    
    def Disable(self, userid=0):
        '''
        Disable the timer.
        '''
        self.fire('Disable', userid=userid)
    
    def Toggle(self, userid=0):
        '''
        Toggle the timer on/off.
        '''
        self.fire('Toggle', userid=userid)
    
    def AddToTimer(self, seconds, userid=0):
        '''
        Add time to the timer if it is currently enabled.
        Does not change the Refire Interval.
        '''
        self.fire('AddToTimer', seconds, userid=userid)
    
    def SubtractFromTimer(self, seconds, userid=0):
        '''
        Subtract time from the timer if it is currently enabled.
        Does not change the Refire Interval.
        '''
        self.fire('SubtractFromTimer', seconds, userid=userid)


class material_modify_control(_Parentname, _Targetname):
    '''
    An entity that can be used to directly control material vars. To use it,
    you need to read the .vmt of the material you intend to change. Parent
    this entity to a brush model entity who's material you want to control.
    '''
    materialName = _KeyValue('') # Material to modify.
    materialVar = _KeyValue('') # Material variable to modify.
    
    def SetMaterialVar(self, param, userid=0):
        '''
        Fire to modify a material variable.
        The argument is the value to set the variable to.
        '''
        self.fire('SetMaterialVar', param, userid=userid)
    
    def SetMaterialVarToCurrentTime(self, userid=0):
        '''
        This sets the material variable to the current time on the server.
        '''
        self.fire('SetMaterialVarToCurrentTime', userid=userid)
    
    def StartAnimSequence(self, frame_start, frame_end, frame_rate, loop, userid=0):
        '''
        Force an animated material with the MaterialModifyAnimated proxy
        to play a set of animation frames.
        
        Format is: <Frame Start> <Frame End> <Frame Rate> <Loop>
        
        Setting <Frame End> to -1 uses the last frame of the texture.
        <Loop> should be 1 or 0.
        '''
        self.fire('StartAnimSequence', "%s %s %s %s"%(
            frame_start, frame_end, frame_rate, loop,
        ), userid=userid)
    
    def StartFloatLerp(self, start_value, end_value, transition_time, loop, userid=0):
        '''
        Force a material with the MaterialModifyAnimated proxy to lerp a
        material var between two floating point values.
        
        Format is: <Start Value> <End Value> <Transition Time> <Loop>
        <Loop> should be 1 or 0.
        '''
        self.fire('StartFloatLerp', "%s %s %s %s"%(
            start_value, end_value, transition_time, loop,
        ), userid=userid)


class math_remap(_Targetname, _EnableDisable):
    '''
    An entity that remaps a range of input values to a given range of output values.
    '''
    spawnflags = _KeyValue(Flags({
        'Ignore out of range input values': 1,
        'Clamp output to output range': 2,
    }, 'Ignore out of range input values', 'Clamp output to output range'))
    in1 = _KeyValue(0) # Input values below this value will be ignored.
    in2 = _KeyValue(1) # Input values above this value will be ignored.
    out1 = _KeyValue(0), # When the input value is equal to 'Minimum Valid Input Value'
        # this is the output value.
    out2 = _KeyValue(1), # When the input value is equal to 'Maximum Valid Input Value'
        # this is the output value.
    
    def InValue(self, param, userid=0):
        '''
        Input value and fire the output with the remapped value.
        '''
        self.fire('InValue', param, userid=userid)


class math_colorblend(_Targetname):
    '''
    Used to create a blend between two colors for controlling the color of another entity.
    '''
    spawnflags = _KeyValue(Flags({
        'Ignore out of range input values': 1,
    }, 'Ignore out of range input values'))
    inmin = _KeyValue(0) # Input values below this value will be ignored.
    inmax = _KeyValue(1) # Input values above this value will be ignored.
    colormin = _KeyValue(Color(0,0,0)) # When the input value is equal to
        # 'Minimum Valid Input Value', this is the output RGB color.
    colormax = _KeyValue(Color(255,255,255)) # When the input value is equal to
        # 'Maximum Valid Input Value', this is the output RGB color.
    
    def InValue(self, param, userid=0):
        '''
        Input value and fire the output with the remapped value.
        '''
        self.fire('InValue', param, userid=userid)


class math_counter(_Targetname, _EnableDisable):
    '''
    Holds a numeric value and performs arithmetic operations upon it.
    If either the minimum or maximum legal value is nonzero, OutValue will be clamped
    to the legal range, and the OnHitMin/OnHitMax outputs will be fired at
    the appropriate times. If both min and max are set to zero, no clamping
    is performed and only the OutValue output will be fired.
    '''
    startvalue = _KeyValue(0) # Starting value for the counter.
    min = _KeyValue(0) # Minimum legal value for the counter.
        # If min=0 and max=0, no clamping is performed.
    max = _KeyValue(0) # Maximum legal value for the counter.
        # If min=0 and max=0, no clamping is performed.
    
    def Add(self, param, userid=0):
        '''
        Add an amount to the counter and fire the OutValue output with the result.
        '''
        self.fire('Add', param, userid=userid)
    
    def Divide(self, param, userid=0):
        '''
        Divide the counter by an amount and fire the OutValue output with the result.
        '''
        self.fire('Divide', param, userid=userid)
    
    def Multiply(self, param, userid=0):
        '''
        Multiply the counter by an amount and fire the OutValue output with the result.
        '''
        self.fire('Multiply', param, userid=userid)
    
    def SetValue(self, param, userid=0):
        '''
        Set the counter to a new value and fire the OutValue output with the result.
        '''
        self.fire('SetValue', param, userid=userid)
    
    def SetValueNoFire(self, param, userid=0):
        '''
        Set the counter to a new value without firing any outputs.
        '''
        self.fire('SetValueNoFire', param, userid=userid)
    
    def Subtract(self, param, userid=0):
        '''
        Subtract an amount from the counter and fire the OutValue output with the result.
        '''
        self.fire('Subtract', param, userid=userid)
    
    def SetHitMax(self, param, userid=0):
        '''
        Set the upper bound of the counter and fire the OutValue output with the current value.
        '''
        self.fire('SetHitMax', param, userid=userid)
    
    def SetHitMin(self, param, userid=0):
        '''
        Set the lower bound of the counter and fire the OutValue output with the current value.
        '''
        self.fire('SetHitMin', param, userid=userid)
    
    def GetValue(self, userid=0):
        '''
        Causes the counter fire its OnGetValue output with the current value of the counter.
        Used for polling the counter when you don't want constant updates from
        the OutValue output.
        '''
        self.fire('GetValue', userid=userid)


class momentary_rot_button(_Targetname, _Parentname, _RenderFields):
    '''
    A brush entity that's designed to be used for rotating wheels,
    where the player can rotate them to arbitrary positions before stopping.
    '''
    speed = _KeyValue(50) # The amount, in degrees, that the wheel turns per second.
    master = _KeyValue('') # Legacy support
    sounds = _KeyValue(SymbolicValue(0, {
        'None': 0,
        'Big zap & Warmup': 1,
        'Access Denied': 2,
        'Access Granted': 3,
        'Quick Combolock': 4,
        'Power Deadbolt 1': 5,
        'Power Deadbolt 2': 6,
        'Plunger': 7,
        'Small zap': 8,
        'Keycard Sound': 9,
        'Squeaky': 21,
        'Squeaky Pneumatic': 22,
        'Ratchet Groan': 23,
        'Clean Ratchet': 24,
        'Gas Clunk': 25,
    })) # Sounds
    distance = _KeyValue(90) # The maximum amount, in degrees
        # that the wheel is allowed to rotate.
    returnspeed = _KeyValue(0), # If the 'Toggle' spawnflag is not set
        # the speed at which the wheel auto-returns when left alone, in degrees per second.
    spawnflags = _KeyValue(Flags({
        'Not Solid': 1,
        'Toggle (Disable Auto Return)': 32,
        'X Axis': 64,
        'Y Axis': 128,
        'Use Activates': 1024,
        'Starts locked': 2048,
        'Jiggle when used while locked': 8192,
    }, 'Not Solid', 'Toggle (Disable Auto Return)', 'Use Activates'))
    _minlight = _KeyValue('') # The minimum level of ambient light that hits this brush.
    startposition = _KeyValue(RangedValue(0.0, 0.0, 1.0)) # Postion when spawned.
        # The value is a range between 0.0 and 1.0, where 0 is the unrotated position
        # and 1 is the rotated position + 'Distance'.
    startdirection = _KeyValue(SymbolicValue('Forward', {
        'Forward': -1,
        'Backward': 1,
    })) # Start Direction
    solidbsp = _KeyValue(SymbolicValue(0, {
        'No': 0,
        'Yes': 1,
    })) # Solid BSP
    
    def Lock(self, userid=0):
        '''
        Lock the button, preventing it from functioning.
        '''
        self.fire('Lock', userid=userid)
    
    def Unlock(self, userid=0):
        '''
        Unlock the button, allowing it to function.
        '''
        self.fire('Unlock', userid=userid)
    
    def SetPosition(self, position, userid=0):
        '''
        Move to a position.
        The parameter must be a value between 0 and 1, where 0 is the
        unrotated position and 1 is the rotated position + 'Distance'.
        '''
        self.fire('SetPosition', position, userid=userid)
    
    def SetPositionImmediately(self, position, userid=0):
        '''
        Immediately teleport to a position.
        The parameter must be a value between 0 and 1, where 0 is the unrotated
        position and 1 is the rotated position + 'Distance'.
        '''
        self.fire('SetPositionImmediately', position, userid=userid)


class move_keyframed(_Targetname, _Parentname, _KeyFrame, _Mover):
    '''
    Keyframed Move Behavior
    '''
    pass


class move_rope(_Targetname, _Parentname, _KeyFrame, _RopeKeyFrame):
    '''
    The first node in set of nodes that are used to place ropes in the world.
    It should connect to 1 or more keyframe_rope entities.
    '''
    PositionInterpolator = _KeyValue(SymbolicValue(2, {
        'Linear': 0,
        'Catmull-Rom Spline': 1,
        'Rope': 2,
    })) # Curve Type. Currently only type 2 (Rope) is fully supported.


class move_track(_Targetname, _Parentname, _Mover, _KeyFrame):
    '''
    Track Move Behavior
    '''
    WheelBaseLength = _KeyValue(50) # Distance between the wheels
    Damage = _KeyValue(0) # Damage done to blocking entities
    NoRotate = _KeyValue(SymbolicValue(0, {
        'Yes': 0,
        'No': 1,
    })) # Turn to face down path


class npc_furniture(_BaseNPC, _Parentname):
    '''
    An entity used for non-NPCs that need to synchronise their
    animation with an NPC in a scripted_sequence. Usually a piece of furniture
    or door that an NPC needs to manipulate within a scripted_sequence.
    '''
    model = _KeyValue('') # Model
    
    def DisablePlayerCollision(self, userid=0):
        '''
        Disable collisions against the player.
        '''
        self.fire('DisablePlayerCollision', userid=userid)
    
    def EnablePlayerCollision(self, userid=0):
        '''
        Enable collisions against the player.
        '''
        self.fire('EnablePlayerCollision', userid=userid)


class npc_puppet(_BaseNPC, _Parentname, _Studiomodel):
    '''
    Puppet entity that mimics the animations of a target NPC.
    '''
    animationtarget = _KeyValue('') # Animation target
    attachmentname = _KeyValue('') # Target attachment point name
    
    def SetAnimationTarget(self, param, userid=0):
        '''
        Set the animation target to use.
        An empty string will disable the puppet.
        '''
        self.fire('SetAnimationTarget', param, userid=userid)


class npc_template_maker(_BaseNPCMaker):
    '''
    An entity that creates NPCs.
    
    The NPCs it creates are clones of a template NPC. NPCs are spawned around
    this maker's origin, or at specified destination points.
    '''
    spawnflags = _KeyValue(Flags({
        'Always use radius spawn': 256,
        "Don't preload template models": 512,
    })) # rest of flags are copied from template
    TemplateName = _KeyValue('')
    Radius = _KeyValue(256.0)
    DestinationGroup = _KeyValue('')
    CriterionVisibility = _KeyValue(SymbolicValue("Don't Care", {
        "Yes": 0,
        "No": 1,
        "Don't Care": 2,
    }))
    CriterionDistance = _KeyValue(SymbolicValue("Don't Care", {
        "Nearest": 0,
        "Farthest": 1,
        "Don't Care": 2,
    }))
    MinSpawnDistance = _KeyValue(0)
    
    def SpawnNPCInRadius(self, userid=0):
        '''Spawn an NPC somewhere within the maker's radius.'''
        self.fire('SpawnNPCInRadius', userid=userid)
    
    def SpawnNPCInLine(self, userid=0):
        '''Spawn an NPC somewhere within a line behind the maker.'''
        self.fire('SpawnNPCInLine', userid=userid)
    
    def SpawnMultiple(self, amount, userid=0):
        '''Spawn multiple NPCs (uses destination group, else radisu).'''
        self.fire('SpawnMultiple', amount, userid=userid)
    
    def ChangeDestinationGroup(self, target, userid=0):
        '''Switch to a different set of destination entities.'''
        self.fire('ChangeDestinationGroup', target, userid=userid)
    
    def SetMinimumSpawnDistance(self, distance, userid=0):
        '''Set the minimum spawn distance from player to destination node.'''
        self.fire('SetMinimumSpawnDistance', distance, userid=userid)


class path_track(_Targetname, _Parentname):
    '''
    An entity used to build paths for other entities to follow.
    Each path_track is a node on the path, each holding the name of the next
    path_track in the path.
    '''
    spawnflags = _KeyValue(Flags({
        'Disabled': 1,
        'Fire once': 2,
        'Branch Reverse': 4,
        'Disable train': 8,
        'Teleport to THIS path track': 16,
    }))
    target = _KeyValue('') # The next path_track in the path.
    altpath = _KeyValue('') # An alternative path_track to be the next node in
        # the path. Useful for making branching paths. Use the
        # ToggleAlternatePath / EnableAlternatePath inputs to make
        # the alternative path active.
    speed = _KeyValue(0.0), # When the train reaches this path_track
        # it will set its speed to this speed. 
    radius = _KeyValue(0.0) # Used by NPCs who follow track paths
        # (attack chopper/gunship). This tells them the maximum distance
        # they're allowed to be from the path at this node.
    orientationtype = _KeyValue(SymbolicValue(1, {
        'No change': 0,
        'Face direction of motion': 1,
        "Face this path_track's angles": 2,
    })) # The way that the path follower faces as it moves through this path track.
    
    def ToggleAlternatePath(self, userid=0):
        '''
        Cause the track to toggle to/from its alternate path.
        '''
        self.fire('ToggleAlternatePath', userid=userid)
    
    def EnableAlternatePath(self, userid=0):
        '''
        Enable the alternate path of the track.
        '''
        self.fire('EnableAlternatePath', userid=userid)
    
    def DisableAlternatePath(self, userid=0):
        '''
        Disable the alternate path of the track.
        '''
        self.fire('DisableAlternatePath', userid=userid)
    
    def TogglePath(self, userid=0):
        '''
        Cause the track to toggle on/off/
        '''
        self.fire('TogglePath', userid=userid)
    
    def EnablePath(self, userid=0):
        '''
        Enable the track.
        '''
        self.fire('EnablePath', userid=userid)
    
    def DisablePath(self, userid=0):
        '''
        Disable the track.
        '''
        self.fire('DisablePath', userid=userid)


class phys_ballsocket(_TwoObjectPhysics):
    '''
    A constraint that keeps the position of two objects fixed,
    relative to the constraint's origin. It does not affect rotation.
    '''
    pass


class phys_constraint(_TwoObjectPhysics):
    '''
    A constraint that keeps the relative position and orientation of two objects fixed.
    '''
    pass


class phys_constraintsystem(_Targetname):
    '''
    An entity used to manage a group of interacting constraints and keep
    them stable. All constraints on a set of entities should be placed in
    the same system, or they will fight each other during simulation.
    '''
    additionaliterations = _KeyValue(0) # Adding iterations makes the interactions
        # among constraints in a system tighter. It will not compensate
        # for errors due to collision, but will help in cases where objects
        # of disparate mass are constrained to each other.


class phys_convert(_Targetname):
    '''
    Turns an arbitrary entity into a physically simulated entity.
    i.e. brush entities will behave like func_physbox,
    model entities behave like prop_physics.
    '''
    spawnflags = _KeyValue(Flags({
        'Convert Asleep': 1,
        'Convert As Debris': 2,
    }))
    target = _KeyValue('') # Name of the entity that will be converted to a physics
        # object when the ConvertTarget input is fired.
    swapmodel = _KeyValue('') # Model Swap Entity
    massoverride = _KeyValue(0.0) # Sets the mass when the object(s) are converted
        # (0 means auto-calculate)
    
    def ConvertTarget(self, userid=0):
        '''
        Converts this entity's target to a physically simulated object.
        '''
        self.fire('ConvertTarget', userid=userid)


class phys_hinge(_TwoObjectPhysics):
    '''
    A physically simulated hinge. Use the helper to define the axis of rotation.
    '''
    hingefriction = _KeyValue(0.0) # Resistance/friction in the hinge
    hingeaxis = _KeyValue(Vector((0,0,0))) # Hinge Axis
    SystemLoadScale = _KeyValue(1.0) # Scale of the load connected to this hinge
        # (1=just the objects directly connected)
    minSoundThreshold = _KeyValue(6.0) # When travelling below this many units/sec
        # will not play any sound.
    maxSoundThreshold = _KeyValue(80.0), # When travelling at this speed or above
        # will play sound at full volume.
    slidesoundfwd = _KeyValue('') # Play this sound when travelling forward on helper axis
    slidesoundback = _KeyValue('') # Play this sound when travelling backward on helper axis
    reversalsoundthresholdSmall = _KeyValue(0.0) # When accelerating by more than
        # this many units/sec^2 opposite to direction of travel,
        # play the small reversal sound.
    reversalsoundthresholdMedium = _KeyValue(0.0) # When accelerating by more than
        # this many units/sec^2 opposite to direction of travel, play
        # the medium reversal sound.
    reversalsoundthresholdLarge = _KeyValue(0.0) # When accelerating by more than
        # this many units/sec^2 opposite to direction of travel, play
        # the large reversal sound.
    reversalsoundSmall = _KeyValue('') # Play this sound when making a hard reverse
        # over the small threshold but less than medium
    reversalsoundMedium = _KeyValue('') # Play this sound when making a hard reverse
        # over the medium threshold but less than large
    reversalsoundLarge = _KeyValue('') # Play this sound when making a hard reverse
        # over the large threshold
    
    def SetAngularVelocity(self, angular_velocity, userid=0):
        '''
        Set angular velocity around the hinge (motor) in deg/sec
        '''
        self.fire('SetAngularVelocity', angular_velocity, userid=userid)


class phys_keepupright(_Targetname):
    '''
    A controller that tries to keep an entity facing a particular direction.
    '''
    spawnflags = _KeyValue(Flags({
        'Start inactive': 1,
    }))
    attach1 = _KeyValue('') # The entity to align to the desired angles.
    angularlimit = _KeyValue(15.0) # The maximum angular velocity that this
        # controller can compensate for, in degrees per second.
    
    def TurnOn(self, userid=0):
        '''
        Enable the controller.
        '''
        self.fire('TurnOn', userid=userid)
    
    def TurnOff(self, userid=0):
        '''
        Disable the controller.
        '''
        self.fire('TurnOff', userid=userid)


class phys_lengthconstraint(_TwoObjectPhysics):
    '''
    A constraint that preserves the distance between two entities.
    If the 'Keep Rigid' flag is set, think of it as a rod. If not,
    think off it as a virtual rope.
    '''
    addlength = _KeyValue(0.0) # Add (or subtract) this amount to the rest
        # length of the rope.
    minlength = _KeyValue(0.0) # If the constraint is not rigid, this is the
        # minimum length it can be.
    attachpoint = _KeyValue(Vector((0,0,0))) # The position the rope attaches to object 2
    spawnflags = _KeyValue(Flags({
        'No Collision until break': 1,
        'Keep Rigid': 2,
    }, 'No Collision until break'))


class phys_magnet(_Parentname, _Studiomodel):
    '''
    An entity that acts like a magnet, attaching metallic physics objects
    to itself when they touch it.
    '''
    spawnflags = _KeyValue(Flags({
        'Start Asleep': 1,
        'Motion Disabled': 2,
        'Suck On Touch': 4,
        'Allow Attached Rotation': 8,
        'Coast jeep pickup hack': 16,
    }))
    forcelimit = _KeyValue(0.0) # The amount of force necessary to break a stuck
        # object off the magnet. A way of calculating this is to set it
        # to the mass of an object that would break this constraint if
        # it were resting on the magnet.
    torquelimit = _KeyValue(0.0) # The amount of torque necessary to break a stuck
        # object off the magnet. A way of calculating this is to multiply
        # any reference mass by the resting distance (from the center
        # of mass of the object) needed to break the constraint.
    massScale = _KeyValue(0.0) # A scale multiplier for the object's mass.
    overridescript = _KeyValue('') # A list of physics key/value pairs that are
        # usually in a physics prop .qc file.
        # Format is 'key,value,key,value,etc'.
    maxobjects = _KeyValue(0) # The maximum number of physics objects that can be
        # stuck to the magnet at once. 0 = no limit.
    
    def TurnOn(self, userid=0):
        '''
        Turn the magnet on.
        '''
        self.fire('TurnOn', userid=userid)
    
    def TurnOff(self, userid=0):
        '''
        The the magnet off. This will detach anything current stuck to the magnet.
        '''
        self.fire('TurnOff', userid=userid)


class phys_motor(_Targetname):
    '''
    An entity that tries to spin a target entity at a particular speed.
    '''
    speed = _KeyValue('0') # Angular speed (units are degress/second)
    spinup = _KeyValue('1') # spin up time in seconds
        # (also affects the rate at which speed changes happen)
    inertiafactor = _KeyValue(1.0) # Make this larger if the object being
        # driven is constrained to a set of heavier objects.
    axis = _KeyValue(Vector((0,0,0))) # Rotation Axis
    spawnflags = _KeyValue(Flags({
        'Start On': 1, # starts on by default
        'No world collision': 2, # Disable world collisions on hinges
        'Hinge Object': 4, # motor also acts as a hinge constraining
            # the object to this axis
        'Orient Locally': 8, # Maintain local relationship with the attached object
            # (NOT WORKING YET)
    }, 'Start On', 'Hinge Object'))
    attach1 = _KeyValue('') # Object to apply the force to
    
    def SetSpeed(self, speed, userid=0):
        '''
        Sets target speed
        '''
        self.fire('SetSpeed', speed, userid=userid)
    
    def TurnOn(self, userid=0):
        '''
        Turns motor on
        '''
        self.fire('TurnOn', userid=userid)
    
    def TurnOff(self, userid=0):
        '''
        Turns motor off
        '''
        self.fire('TurnOff', userid=userid)


class phys_pulleyconstraint(_TwoObjectPhysics):
    '''
    A constraint that is essentially two length constraints and two points.
    Imagine it as a virtual rope connected to two objects, each suspended
    from a pulley above them. The constraint keeps the sum of the distances
    between the pulley points and their suspended objects constant.
    '''
    addlength = _KeyValue(0.0) # Add (or subtract) this amount to the rest length
        # of the pulley rope.
    gearratio = _KeyValue(1.0) # Add (or subtract) this amount to the rest length
        # of the pulley rope.
    position2 = _KeyValue(Vector((0,0,0))) # The position of the pulley for Entity 2.
        # The pulley for Entity 1 is the origin of this constraint entity.
        # Entity 1 is always suspended from pulley point 1, and Entity 2
        # is always suspended from pulley point 2.
    spawnflags = _KeyValue(Flags({
        'No Collision until break': 1,
        'Keep Rigid': 2,
    }, 'No Collision until break'))


class phys_ragdollconstraint(_TwoObjectPhysics):
    '''
    A constraint that fixes the position of two entities,
    relative to this constraint's origin. Also allows for limits on the
    rotation around each axis, in the space of this constraint.
    '''
    spawnflags = _KeyValue(Flags({
        'No Collision until break': 1,
        'Only limit rotation (free movement)': 2,
    }, 'No Collision until break'))
    xmin = _KeyValue(-90.0) # -180 min and 180 max = no constraint on this axis.
    xmax = _KeyValue(90.0) # -180 min and 180 max = no constraint on this axis.
    ymin = _KeyValue(0.0) # -180 min and 180 max = no constraint on this axis.
    ymax = _KeyValue(0.0) # -180 min and 180 max = no constraint on this axis.
    zmin = _KeyValue(0.0) # -180 min and 180 max = no constraint on this axis.
    zmax = _KeyValue(0.0) # -180 min and 180 max = no constraint on this axis.
    xfriction = _KeyValue(0.0) # X axis friction
    yfriction = _KeyValue(0.0) # Y axis friction
    zfriction = _KeyValue(0.0) # Z axis friction


class phys_ragdollmagnet(_Targetname, _EnableDisable, _Parentname):
    '''
    An entity that acts like a magnet for ragdolls.
    Useful for crafting exaggerated ragdoll behavior
    (i.e. guys falling over rails on death).
    If the Bar Magnet spawnflag is set, the magnet works like it was a cylindrical
    magnet i.e. it attracts ragdolls to the nearest point on a line.
    '''
    axis = _KeyValue(Vector((0,0,0))) # Bar Magnet Axis
    radius = _KeyValue(512.0) # Radius in which ragdolls are affected
        # around this entity's origin.
    force = _KeyValue(5000.0) # Magnetic force to apply to ragdolls within the radius.
        # Expressed as kilograms per inch per second.
        # So a force of 1000 will add 10 inches/second to a 100kg man.
        # It will add 100 inches per second to a 10kg headcrab.
    target = _KeyValue('') # If specified, the phys_ragdollmagnet will only
        # affect the target entity.
    spawnflags = _KeyValue(Flags({
        'Bar Magnet': 2,
    }))


class phys_slideconstraint(_TwoObjectPhysics):
    '''
    A constraint that constrains an entity along a line segment.
    '''
    spawnflags = _KeyValue(Flags({
        'No Collision until break': 1,
        'Limit Endpoints': 2,
    }, 'No Collision until break'))
    slideaxis = _KeyValue(Vector((0,0,0))) # Sliding Axis
    slidefriction = _KeyValue(0.0) # Resistance/friction in the constraint
    SystemLoadScale = _KeyValue(1.0) # Scale of the mass load connected to this
        # constraint (1=just the objects directly connected)
    minSoundThreshold = _KeyValue(6.0), # When travelling below this many units/sec
        # will not play any sound.
    maxSoundThreshold = _KeyValue(80.0), # When travelling at this speed or above
        # will play sound at full volume.
    slidesoundfwd = _KeyValue('') # Play this sound when travelling forward on helper axis
    slidesoundback = _KeyValue('') # Play this sound when travelling backward on helper axis
    reversalsoundthresholdSmall = _KeyValue(0.0) # When accelerating by more than
        # this many units/sec^2 opposite to direction of travel, play
        # the small reversal sound.
    reversalsoundthresholdMedium = _KeyValue(0.0) # When accelerating by more than
        # this many units/sec^2 opposite to direction of travel, play
        # the medium reversal sound.
    reversalsoundthresholdLarge = _KeyValue(0.0) # When accelerating by more than
        # this many units/sec^2 opposite to direction of travel, play
        # the large reversal sound.
    reversalsoundSmall = _KeyValue('') # Play this sound when making a hard reverse
        # over the small threshold but less than medium
    reversalsoundMedium = _KeyValue('') # Play this sound when making a hard reverse
        # over the medium threshold but less than large
    reversalsoundLarge = _KeyValue('') # Play this sound when making a hard reverse
        # over the large threshold
    
    def SetVelocity(self, velocity, userid=0):
        '''
        Set linear velocity along the constraint
        '''
        self.fire('SetVelocity', velocity, userid=userid)


class phys_spring(_Targetname):
    '''
    A physically simulated spring. 'Length' is what's known as the
    'natural spring length'. This is how long the spring would be if it
    was at rest (nothing hanging on it or attached). When you attach
    something to the spring, it will stretch longer than its 'natural length'.
    The amount of stretch is determined by the 'Sprint Constant'.
    The larger the spring constant the less stretch the spring.
    '''
    spawnflags = _KeyValue(Flags({
        'Force only on stretch': 1,
    }))
    attach1 = _KeyValue('') # Entity 1
    attach2 = _KeyValue('') # Entity 2
    springaxis = _KeyValue(Vector((0,0,0))) # Use the helper.
        # Drag it out to match the virtual spring.
    length = _KeyValue('0') # How long the spring would be if it was at rest
        # (nothing hanging on it or attached).
        # 0 means the length of the brush.
    constant = _KeyValue('50') # Stiffness of the spring.
        # The larger the number the less the spring will stretch.
    damping = _KeyValue('2.0') # How much energy the spring loses.
        # The larger the number, the less bouncy the spring.
    relativedamping = _KeyValue('0.1') # The amount of energy the spring loses
        # proportional to the relative velocity of the two objects
        # the spring is attached to.
    breaklength = _KeyValue('0'), # If the spring's length ever exceeds this length
        # the spring breaks.
    
    def SetSpringConstant(self, constant, userid=0):
        '''
        Set the Spring Constant.
        '''
        self.fire('SetSpringConstant', constant, userid=userid)
    
    def SetSpringLength(self, length, userid=0):
        '''
        Set the Spring Length.
        '''
        self.fire('SetSpringLength', length, userid=userid)
    
    def SetSpringDamping(self, damping, userid=0):
        '''
        Set the Spring Damping.
        '''
        self.fire('SetSpringDamping', damping, userid=userid)


class phys_thruster(_ForceController):
    '''
    An entity used to apply constant acceleration to a physics object.
    The force and torque is calculated using the position and direction
    of the thruster as an impulse. So moving those off the object's center
    will cause torque as well. Torque can be removed by unchecking the
    'apply torque' flag. The position of the thruster can be forced to be at
    the object's center by checking to 'ignore pos' flag.
    '''
    spawnflags = _KeyValue(Flags({
        'Ignore Pos': 32,
    }))
    force = _KeyValue('0') # Force (will be integrated, units are force kg*in/s^2)


class phys_torque(_ForceController):
    '''
    An angular thruster. Use it to apply angular force to an entity.
    '''
    force = _KeyValue('0') # Angular Acceleration
    axis = _KeyValue(Vector((0,0,0))) # Rotation Axis


class physics_cannister(_Targetname):
    '''
    A physically simulated gas cannister that can have its cap shot off,
    at which point gas will start escaping and cause the cannister to fly
    around. If it takes enough damage, it will explode.
    '''
    model = _KeyValue('models/fire_equipment/w_weldtank.mdl') # World model
    spawnflags = _KeyValue(Flags({
        'Start Asleep': 1,
        'Explodes': 2,
    }, 'Explodes'))
    expdamage = _KeyValue('200.0') # The amount of damage done by the explosion
        # created when the cannister blows up.
    expradius = _KeyValue('250.0') # The radius of the explosion to create when
        # the cannister blows up.
    health = _KeyValue(25) # The amount of damage the cannister takes before exploding.
    thrust = _KeyValue('3000.0') # When the cap has been blown off, and the escaping
        # gas is thrusting the cannister about, this is the amount of thrust generated.
    fuel = _KeyValue('12.0') # The amount of time that gas leaks from
        # the cannister before being considered empty.
    rendercolor = _KeyValue(Color(255,255,255)) # Smoke Color (R G B)
    renderamt = _KeyValue(128) # Smoke Alpha (0 - 255)
    gassound = _KeyValue('ambient/objects/cannister_loop.wav') # The sound played
        # when the gas is escaping from the cannister.
    
    def Activate(self, param, userid=0):
        '''
        Start gas escaping from the cannister.
        '''
        self.fire('Activate', param, userid=userid)
    
    def Deactivate(self, param, userid=0):
        '''
        Stop gas escaping from the cannister.
        '''
        self.fire('Deactivate', param, userid=userid)
    
    def Explode(self, param, userid=0):
        '''
        Force the cannister to explode.
        '''
        self.fire('Explode', param, userid=userid)
    
    def Wake(self, userid=0):
        '''
        Wakes up the cannister, if it is sleeping.
        '''
        self.fire('Wake', userid=userid)


class player(_Targetname, _Parentname, _RenderFields):
    '''
    The player. This entity still exists even if the player is dead.

    For the list of attributes and methods, see the Valve Developer Community wiki page
    http://developer.valvesoftware.com/wiki/Player
    '''
    # FIXME
    friction = _KeyValue(1.0)
    health = _KeyValue(100)
    max_health = _KeyValue(100)
    physdamagescale = _KeyValue(1.0)
    
    def IgnoreFallDamage(self, userid=0):
        '''IgnoreFallDamage (undocumented)'''
        self.fire('IgnoreFallDamage', userid=userid)

    def IgnoreFallDamageWithoutReset(self, userid=0):
        '''IgnoreFallDamage (do not reset)'''
        self.fire('IgnoreFallDamageWithoutReset', userid=userid)
    
    def OnSquadMemberKilled(self, param=None, userid=0):
        '''OnSquadMemberKilled (undocumented, shouldn't this be output?)'''
        self.fire('OnSquadMemberKilled', param, userid=userid)
    
    def DisableFlashlight(self, userid=0):
        '''Disables the player's flashlight'''
        self.fire('DisableFlashlight', userid=userid)
    
    def EnableFlashlight(self, userid=0):
        '''Enables the player's flashlight'''
        self.fire('EnableFlashlight', userid=userid)
    
    def ForceDropPhysObjects(self, userid=0):
        '''Forces the player drop any grabbed physics objects (undocumented)'''
        self.fire('ForceDropPhysObjects', userid=userid)
    
    def SetHealth(self, health, userid=0):
        '''Set player health'''
        self.fire('Health', health, userid=userid)
    
    def SetHUDVisibility(self, value, userid=0):
        '''Set HUD visibility (0 = off, 1 = on)'''
        self.fire('SetHUDVisibility', value, userid=userid)
    
    def SetFogController(self, controller, userid=0):
        '''Set Fog Controller for the player'''
        self.fire('SetFogController', controller, userid=userid)
    
    def KilledNPC(self, param, userid=0):
        '''Killed a non-playing character (undocumented)'''
        self.fire('KilledNPC', param, userid=userid)

    def skin(self, skin_name, userid=0):
        '''Set the model's skin'''
        self.fire('skin', skin_name, userid=userid)


class point_anglesensor(_Targetname, _Parentname, _EnableDisable):
    '''
    An entity that detects if another entity points in a given direction for a
    period of time.
    '''
    target = _KeyValue('') # Name of the entity whose angles will be sensed.
    lookatname = _KeyValue('') # The entity we want to check to see if the
        # Target Entity is looking at.
    duration = _KeyValue(0.0) # The amount of time the Target Entity must look at
        # the 'Look at Entity' to trigger this entity, in seconds.
    tolerance = _KeyValue(0) # The tolerance, in degrees, in the checking to
        # determine when the Target Entity is looking at the Look At Entity.
    spawnflags = _KeyValue(Flags({
        "Use target entity's angles (NOT position)": 1,
    }))
    
    def Toggle(self, userid=0):
        '''
        Toggle the sensor between enabled and disabled.
        '''
        self.fire('Toggle', userid=userid)
    
    def Test(self, userid=0):
        '''
        Check to see if the Target Entity is facing the Look At Entity
        within the specified tolerance, firing either the OnFacingLookat or
        OnNotFacingLookat output based on the result.
        '''
        self.fire('Test', userid=userid)


class point_angularvelocitysensor(_Targetname):
    '''
    An entity that detects if another entity's angular velocity
    meets or exceeds a threshold value.
    '''
    target = _KeyValue('') # Name of the entity whose angular velocity will be sensed.
    threshold = _KeyValue(0.0), # The threshold angular velocity to compare against
        # in degrees per second.
    fireinterval = _KeyValue(0.2) # Angular velocity must cross the threshold for
        # at least this long to fire.
    axis = _KeyValue(Vector((0,0,0))) # Axis
    usehelper = _KeyValue(SymbolicValue(0, {
        'No': 0,
        'Yes': 1,
    })) # Use axis helper to determine rotation values (clockwise/counter-clockwise).
    
    def Test(self, userid=0):
        '''
        Check to see if the Target Entity's angular velocity meets or
        exceeds the Threshold Velocity, 
        '''
        self.fire('Test', userid=userid)
    
    def TestWithInterval(self, userid=0):
        '''
        Check to see if the Target Entity's angular velocity meets or
        exceeds the Threshold Velocity. Once the Fire Interval expires, 
        '''
        self.fire('TestWithInterval', userid=userid)


class point_bonusmaps_accessor(_Targetname):
    '''
    An entity that relays bonus maps changes.
    '''
    filename = _KeyValue('') # File Name
    mapname = _KeyValue('') # Map Name
    
    def Unlock(self, userid=0):
        '''
        Unlock the filename/map combination.
        '''
        self.fire('Unlock', userid=userid)
    
    def Complete(self, userid=0):
        '''
        Complete the filename/map combination.
        '''
        self.fire('Complete', userid=userid)
    
    def Save(self, userid=0):
        '''
        Save bonus map data.
        '''
        self.fire('Save', userid=userid)


class point_clientcommand(_Targetname):
    '''
    An entity that issues commands to the client console,
    as if it was typed in by the player
    (if activator is a player, or the local player in single player).
    '''
    def Command(self, command, userid=0):
        '''
        Command to execute.
        '''
        self.fire('Command', command, userid=userid)


class point_devshot_camera(_BaseEntity):
    '''
    An entity used by the -makedevshots system, which automatically takes
    screenshots at the position of every devshot camera in the level.
    '''
    cameraname = _KeyValue('') # Used as the name of the directory to store
        # screenshots from this camera. Must be unique within the level.
    FOV = _KeyValue(75) # FOV of this camera.


class point_enable_motion_fixup(_Parentname):
    '''
    An entity used to move a motion-disabled prop when it enables motion.
    Parent this entity to the prop, and when the prop has its motion enabled,
    it will immediately teleport to the origin of this entity.
    '''
    pass


class point_gamestats_counter(_Targetname, _EnableDisable):
    '''
    Generic game statistics counter.
    '''
    Name = _KeyValue('') # Name of statistic counter
    
    def SetName(self, name, userid=0):
        '''
        Change name of statistic counter
        '''
        self.fire('SetName', name, userid=userid)
    
    def Increment(self, value, userid=0):
        '''
        Increment named statistic counter by specified value
        '''
        self.fire('Increment', value, userid=userid)


class point_hurt(_Targetname):
    '''
    An entity that does damage to all entities in a radius around itself,
    with a specified delay. If 'Target Entity' is specified, the damage is
    only done to that entity.
    '''
    DamageTarget = _KeyValue('') # If specified, only this entity will take damage.
        # Otherwise, all entities within the Radius will take damage.
    DamageRadius = _KeyValue(256.0) # All entities within this radius of this
        # entity will take damage. If a 'Target Entity' is specified,
        # only that entity will take damage.
    Damage = _KeyValue(5) # Damage done to all affected entities each time this entity fires.
    DamageDelay = _KeyValue(1.0) # Delay between refires, in seconds.
    DamageType = _KeyValue(SymbolicValue(0, {
        'GENERIC': 0,
        'CRUSH': 1,
        'BULLET': 2,
        'SLASH': 4,
        'BURN': 8,
        'FREEZE': 16,
        'FALL': 32,
        'BLAST': 64,
        'CLUB': 128,
        'SHOCK': 256,
        'SONIC': 512,
        'ENERGYBEAM': 1024,
        'DROWN': 16384,
        'PARALYSE': 32768,
        'NERVEGAS': 65536,
        'POISON': 131072,
        'RADIATION': 262144,
        'DROWNRECOVER': 524288,
        'CHEMICAL': 1048576,
        'SLOWBURN': 2097152,
        'SLOWFREEZE': 4194304,
    })) # Type of damage to inflict on entities damaged.
    
    def Hurt(self, userid=0):
        '''
        Force a single fire, damaging either the Target Entity or all
        entities within the radius.
        '''
        self.fire('Hurt', userid=userid)
    
    def TurnOn(self, userid=0):
        '''
        Enable this entity. It will start damaging entities everytime it
        fires, and refire based upon the specified Delay.
        '''
        self.fire('TurnOn', userid=userid)
    
    def TurnOff(self, userid=0):
        '''
        Disable this entity. It will stop damaging entities.
        '''
        self.fire('TurnOff', userid=userid)
    
    def Toggle(self, userid=0):
        '''
        Toggle this entity between On/Off state.
        '''
        self.fire('Toggle', userid=userid)


class point_message(_Targetname, _Parentname):
    '''
    An entity that displays a text message in the world, at its origin.
    '''
    spawnflags = _KeyValue(Flags({
        'Start Disabled': 1,
    }))
    message = _KeyValue('') # Entity Message
    radius = _KeyValue(128) # Distance the player must be within to see this message.
    developeronly = _KeyValue(SymbolicValue(0, {
        'No': 0,
        'Yes': 1,
    })) # If set, this message will only be visible when developer mode is on.
    
    def Enable(self, userid=0):
        '''
        Start displaying the message text, if the player is within the message radius.
        '''
        self.fire('Enable', userid=userid)
    
    def Disable(self, userid=0):
        '''
        Stop displaying the message text.
        '''
        self.fire('Disable', userid=userid)


class point_playermoveconstraint(_Targetname):
    '''
    An entity that constrains players to a radius around itself,
    slowing them down the closer they get to the edge of the radius.
    '''
    radius = _KeyValue(256.0) # Radius to constrain players to.
    width = _KeyValue(75.0) # Width of the constraint edge.
        # This is the distance in which to start slowing players down
        # as they approach the edge of the radius.
    speedfactor = _KeyValue(0.15) # Factor applied to the player's max speed as
        # they approach the radius edge.
    
    def TurnOn(self, userid=0):
        '''
        Start constraining any players within the radius.
        '''
        self.fire('TurnOn', userid=userid)
    
    def TurnOff(self, userid=0):
        '''
        Stop constraining any players previously constrained.
        '''
        self.fire('TurnOff', userid=userid)


class point_posecontroller(_Targetname):
    '''
    An entity that controls a pose parameter of a prop and cycles the pose clientside.
    '''
    PropName = _KeyValue('') # Name of the prop to control.
    PoseParameterName = _KeyValue('') # Name of the pose parameter to control.
    PoseValue = _KeyValue(RangedValue(0.0, 0.0, 1.0)) # Normalized value
        # for the pose parameter from 0.0 and 1.0 (maps to min and max range).
    InterpolationTime = _KeyValue(RangedValue(0.0, 0.0, 10.0)) # Number of seconds
        # (0.0 to 10.0) for client to match absolue pose values.
    InterpolationWrap = _KeyValue(SymbolicValue(0, {
        'No': 0,
        'Yes': 1,
    })) # Should wrap from 0.0 to 1.0 when interpolating.
    CycleFrequency = _KeyValue(RangedValue(0.0, -10.0, 10.0)) # Base cycles
        # per second from -10.0 to 10.0.
    FModulationType = _KeyValue(SymbolicValue(0, {
        'None': 0,
        'Sine': 1,
        'Square': 2,
        'Triangle': 3,
        'Sawtooth': 4,
        'Noise': 5,
    })) # Frequency Modulation Type
    FModTimeOffset = _KeyValue(RangedValue(0.0, -1.0, 1.0)) # Modulation time
        # offset from -1.0f to 1.0.
    FModRate = _KeyValue(RangedValue(0.0, -10.0, 10.0)) # Modulation cycles
        # per second from -10.0f to 10.0.
    FModAmplitude = _KeyValue(RangedValue(0.0, 0.0, 10.0)) # Modulation extents
        # from 0.0f to 10.0.
    
    def SetPoseParameterName(self, name, userid=0):
        '''
        Set the pose parameter to control.
        '''
        self.fire('SetPoseParameterName', name, userid=userid)
    
    def SetPoseValue(self, value, userid=0):
        '''
        Set the pose parameter to a normalized value between 0.0 and 1.0
        (maps to min and max range).
        '''
        self.fire('SetPoseValue', value, userid=userid)
    
    def SetInterpolationTime(self, seconds, userid=0):
        '''
        Set the interpolation time to a number of seconds between 0.0 and 10.0.
        '''
        self.fire('SetInterpolationTime', seconds, userid=userid)
    
    def SetCycleFrequency(self, frequency, userid=0):
        '''
        Set the pose parameter's base cycles per second from -10.0f to 10.0.
        '''
        self.fire('SetCycleFrequency', frequency, userid=userid)
    
    def SetFModType(self, fmodtype, userid=0):
        '''
        Set the type of frequency modulation.
        '''
        self.fire('SetFModType', fmodtype, userid=userid)
    
    def SetFModTimeOffset(self, offset, userid=0):
        '''
        Set the modulation time offset from -1.0f to 1.0.
        '''
        self.fire('SetFModTimeOffset', offset, userid=userid)
    
    def SetFModRate(self, rate, userid=0):
        '''
        Set the modulation cycles per second from -10.0f to 10.0.
        '''
        self.fire('SetFModRate', rate, userid=userid)
    
    def SetFModAmplitude(self, amplitude, userid=0):
        '''
        Set the modulation extents from 0.0f to 10.0.
        '''
        self.fire('SetFModAmplitude', amplitude, userid=userid)
    
    def RandomizeFMod(self, extremeness, userid=0):
        '''
        Randomize the frequency modulation by an extremeness of 0.0 to 1.0.
        '''
        self.fire('RandomizeFMod', extremeness, userid=userid)


class point_proximity_sensor(_Targetname, _Parentname, _EnableDisable):
    '''
    An entity that detects another entity's proximity to a target position.
    '''
    target = _KeyValue('') # Name of the entity whose angles will be sensed.
    spawnflags = _KeyValue(Flags({
        'Test the distance as measured along the axis specified by our direction.': 1,
    }))
    
    def Toggle(self, userid=0):
        '''
        Toggle the sensor between enabled and disabled.
        '''
        self.fire('Toggle', userid=userid)


class point_servercommand(_Targetname):
    '''
    An entity that issues commands to the server console.
    '''
    def Command(self, command, userid=0):
        '''
        Command to execute.
        '''
        self.fire('Command', command, userid=userid)


class point_spotlight(_Targetname, _Parentname, _RenderFields, _DXLevelChoice):
    '''
    An entity to draw a spotlight.
    Will draw a beam when the player views it side on, and a halo when it's
    facing towards the player. 
    '''
    spawnflags = _KeyValue(Flags({
        'Start On': 1,
        'No Dynamic Light': 2,
    }, 'Start On', 'No Dynamic Light'))
    spotlightlength = _KeyValue(500) # Length of the spotlight beam.
    spotlightwidth = _KeyValue(50) # Width of the spotlight beam.
    
    def LightOn(self, userid=0):
        '''
        Turn the spotlight on.
        '''
        self.fire('LightOn', userid=userid)
    
    def LightOff(self, userid=0):
        '''
        Turn the spotlight off
        '''
        self.fire('LightOff', userid=userid)


class point_teleport(_Targetname):
    '''
    An entity that teleports a target entity to this position and angles.
    If 'Teleport Home' spawn flag is set, teleports the target entity to its
    spawn position instead. If object is physically simulated, simulation is
    turned off when teleported.
    '''
    target = _KeyValue('') # Name of the entity that will be teleported.
    spawnflags = _KeyValue(Flags({
        'Teleport Home': 1,
        'Into Duck (episodic)': 2,
    }))
    
    def Teleport(self, userid=0):
        '''
        Teleport the target entity.
        '''
        self.fire('Teleport', userid=userid)


class point_template(_Targetname):
    '''
    Turn an entity, or set of entities, into a single template that
    can be instanced anywhere, and multiple times. If there are interdependencies
    (entity I/O, hierarchy, or other name references) between the entities
    in the template, the entities in the template will have their names
    changed and the interdependencies will be reconnected to the changes names.
    The name change format is as follows: '<original name>&0000',
    where the 0000 will be replaced with the current global template instance,
    so wildcard searches for '<original name>*' will still find them.
    
    If you don't want the name fixup to happen, because you're only spawning
    the template once, or you want inputs to trigger all instances of the template,
    check the 'Preserve entity names' spawnflag.
    
    To spawn the template in other places, use an env_entity_maker.
    '''
    spawnflags = _KeyValue(Flags({
        "Don't remove template entities": 1,
        "Preserve entity names": 2,
    }, "Preserve entity names"))
    Template01 = _KeyValue('') # Template 1
    Template02 = _KeyValue('') # Template 2
    Template03 = _KeyValue('') # Template 3
    Template04 = _KeyValue('') # Template 4
    Template05 = _KeyValue('') # Template 5
    Template06 = _KeyValue('') # Template 6
    Template07 = _KeyValue('') # Template 7
    Template08 = _KeyValue('') # Template 8
    Template09 = _KeyValue('') # Template 9
    Template10 = _KeyValue('') # Template 10
    Template11 = _KeyValue('') # Template 11
    Template12 = _KeyValue('') # Template 12
    Template13 = _KeyValue('') # Template 13
    Template14 = _KeyValue('') # Template 14
    Template15 = _KeyValue('') # Template 15
    Template16 = _KeyValue('') # Template 16
    
    def ForceSpawn(self, userid=0):
        '''
        Spawn an instance of the template at the original position.
        '''
        self.fire('ForceSpawn', userid=userid)


class point_tesla(_Targetname, _Parentname):
    '''
    An entity that creates tesla lightning arcs around its origin.
    '''
    m_SourceEntityName = _KeyValue('') # If specified, tesla lightning will
        # originate from the specified entity. Otherwise, they originate from this entity.
    m_SoundName = _KeyValue('DoSpark') # Sound to be played whenever lightning is created.
    texture = _KeyValue('sprites/physbeam.vmt') # Material to use for the tesla lightning beams.
    m_Color = _KeyValue(Color(255,255,255)) # Color
    m_flRadius = _KeyValue(200) # Radius around the origin to find a point to strike with
        # a tesla lightning beam.
    beamcount_min = _KeyValue(6) # Minimum number of tesla lightning beams to
        # create when creating an arc.
    beamcount_max = _KeyValue(8) # Maximum number of tesla lightning beams to
        # create when creating an arc.
    thick_min = _KeyValue(4) # Minimum width of the tesla lightning beams.
    thick_max = _KeyValue(5) # Maximum width of the tesla lightning beams.
    lifetime_min = _KeyValue(0.3) # Minimum lifetime of the tesla lightning beams.
    lifetime_max = _KeyValue(0.3) # Maximum lifetime of the tesla lightning beams.
    interval_min = _KeyValue(0.5) # Minimum time delay between random arcing.
    interval_max = _KeyValue(2) # Maximum time delay between random arcing.
    
    def TurnOn(self, userid=0):
        '''
        Turn emitter on.
        '''
        self.fire('TurnOn', userid=userid)
    
    def TurnOff(self, userid=0):
        '''
        Turn emitter off.
        '''
        self.fire('TurnOff', userid=userid)
    
    def DoSpark(self, userid=0):
        '''
        Force a single arc.
        '''
        self.fire('DoSpark', userid=userid)


class point_velocitysensor(_Targetname):
    '''
    An entity that detects and outputs an entity's velocity.
    '''
    target = _KeyValue('') # Name of the entity whose velocity will be sensed.
    axis = _KeyValue(Vector((0,0,0))) # Measurement Axis
    enabled = _KeyValue(SymbolicValue(1, {
        'No': 0,
        'Yes': 1,
    })) # Whether or not to start enabled and active.
    
    def Enable(self, userid=0):
        '''
        Enable the sensor.
        '''
        self.fire('Enable', userid=userid)
    
    def Disable(self, userid=0):
        '''
        Disable the sensor.
        '''
        self.fire('Disable', userid=userid)


class point_viewcontrol(_Targetname, _Parentname):
    '''
    A camera entity that controls the player's view.
    While it's active, the player will see out of the camera.
    '''
    target = _KeyValue('') # Name of the entity that the camera should point at
        # and track while active.
    targetattachment = _KeyValue('') # If set, the camera will focus on the
        # specified attachment on the 'target'.
    wait = _KeyValue(10) # The amount of time the camera should control the
        # player's view for, after which it deactivates itself. If the
        # camera should stay active until told to deactive, set the
        # 'Infinite Hold Time' spawnflag.
    moveto = _KeyValue('') # The first path corner in a track that the camera
        # should move along once it's activated. If not specified,
        # the camera won't move.
    interpolatepositiontoplayer = _KeyValue(SymbolicValue(0, {
        'No': 0,
        'Yes': 1,
    })) # Gradually interpolate player's position to here on start. (Episodic only)
    spawnflags = _KeyValue(Flags({
        'Start At Player': 1,
        'Follow Player': 2,
        'Freeze Player': 4,
        'Infinite Hold Time': 8,
        'Snap to goal angles': 16,
        'Make Player non-solid': 32,
        'Interruptable by Player': 64,
    }, 'Start At Player', 'Follow Player'))
    speed = _KeyValue('0') # The starting speed that the camera moves at,
        # if it's on a path track.
    acceleration = _KeyValue('500') # The speed at which the camera accelerates to
        # path corner's desired speeds.
    deceleration = _KeyValue('500') # The speed at which the camera decelerates to
        # path corner's desired speeds.
    
    def Enable(self, userid=0):
        '''
        Enable the point_viewcontrol, and start controlling the player's view.
        '''
        self.fire('Enable', userid=userid)
    
    def Disable(self, userid=0):
        '''
        Disable the point_viewcontrol, and stop controlling the player's view.
        '''
        self.fire('Disable', userid=userid)


class prop_detail(_prop_detail_base):
    '''
    Detail Prop
    '''
    pass


class prop_door_rotating(_Parentname, _Global, _Studiomodel):
    '''
    An entity used to place a door in the world.
    '''
    slavename = _KeyValue('') # The name of any doors that should be slaved to this door
        # (i.e. should open when this one opens, and close when this one closes).
    hardware = _KeyValue(SymbolicValue(1, {
        '<None>': 0,
        'Lever': 1,
        'Push bar': 2,
        'Keypad': 3,
    })) # Hardware Type
    ajarangles = _KeyValue(Vector((0,0,0))), # If the door 'Spawn Position' is set to Ajar
        # these are the angles to spawn at, instead of being open or closed.
    spawnpos = _KeyValue(SymbolicValue(0, {
        'Closed': 0,
        'Open forward': 1,
        'Open back': 2,
        'Ajar': 3,
    })) # Spawn Position
    axis = _KeyValue('') # Hinge Axis
    distance = _KeyValue(90.0) # The amount, in degrees, that the door should rotate when opened.
    speed = _KeyValue(100) # The speed at which the door moves.
    soundopenoverride = _KeyValue('') # Sound played when the door has finished opening.
    soundcloseoverride = _KeyValue('') # Sound played when the door has finished closing.
    soundmoveoverride = _KeyValue('') # Sound played when the door starts to move.
    returndelay = _KeyValue(-1) # Amount of time, in seconds, after the door has opened
        # before it closes. If the value is set to -1, the door never closes itself.
    dmg = _KeyValue(0) # Amount of damage done to entities that block the movement
        # of this door, per frame.
    health = _KeyValue(0) # Health (0 = Unbreakable)
    soundlockedoverride = _KeyValue(''), # Sound played when the player tries to open the door
        # and fails because it's locked.
    soundunlockedoverride = _KeyValue('') # Sound played when the door is unlocked.
    forceclosed = _KeyValue(SymbolicValue(0, {
        'No': 0,
        'Yes': 1,
    })) # If set, this door will close no matter what.
        # Useful for doors that have to close even if the player tries to block
        # them with objects.
    spawnflags = _KeyValue(Flags({
        'Starts Open': 1,
        'Starts locked': 2048,
        'Door silent (No sound, and does not alert NPCs)': 4096,
        'Use closes': 8192,
        'Door silent to NPCS (Does not alert NPCs)': 16384,
        'Ignore player +USE': 32768,
    }, 'Use closes'))
    opendir = _KeyValue(SymbolicValue(0, {
        'Open Both Directions': 0,
        'Open Forward Only': 1,
        'Open Backward Only': 2,
    })) # Force the door to open only forwards or only backwards.
        # Both directions is the standard door behavior.
    
    def Open(self, userid=0):
        '''
        Open the door, if it is not fully open.
        '''
        self.fire('Open', userid=userid)
    
    def OpenAwayFrom(self, targetname, userid=0):
        '''
        Open the door away from the specified entity.
        '''
        self.fire('OpenAwayFrom', targetname, userid=userid)
    
    def Close(self, userid=0):
        '''
        Close the door, if it is not fully closed.
        '''
        self.fire('Close', userid=userid)
    
    def Toggle(self, userid=0):
        '''
        Toggle the door between open and closed.
        '''
        self.fire('Toggle', userid=userid)
    
    def Lock(self, userid=0):
        '''
        Lock the door.
        '''
        self.fire('Lock', userid=userid)
    
    def Unlock(self, userid=0):
        '''
        Unlock the door.
        '''
        self.fire('Unlock', userid=userid)
    
    def SetRotationDistance(self, distance, userid=0):
        '''
        Set the distance (in degrees) between Open and Closed.
        '''
        self.fire('SetRotationDistance', distance, userid=userid)
    
    def SetSpeed(self, speed, userid=0):
        '''
        Set the speed at which the door rotates. 100 is default.
        '''
        self.fire('SetSpeed', speed, userid=userid)


class prop_dynamic(_prop_dynamic_base, _EnableDisable):
    '''
    A prop that can be placed in hierarchy and can play animations.
    It can also be configured to break when it takes enough damage.
    Note that the health of the object will be overridden by the health
    inside the model, to ensure consistent health game-wide. If the model
    used by the prop is configured to be used as a prop_physics
    (i.e. it should be physically simulated) then it CANNOT be used as
    a prop_dynamic. Upon level load it will display a warning in
    the console and remove itself. Use a prop_physics instead.
    '''
    pass


class prop_dynamic_ornament(_prop_dynamic_base):
    '''
    A way to attach one studio model to another as an ornament.
    It will render in the way that player/NPC weapons render.
    '''
    solid = _KeyValue(SymbolicValue(0, {
        'Not Solid': 0,
    })) # Collisions
    InitialOwner = _KeyValue('') # Name of the entity that this ornament
        # should attach to, at startup.
    
    def SetAttached(self, targetname, userid=0):
        '''
        Attach the ornament to a different entity.
        Parameter should be the name of entity to attach to.
        '''
        self.fire('SetAttached', targetname, userid=userid)
    
    def Detach(self, targetname, userid=0):
        '''
        Detach from the Target Entity and become invisible.
        The ornament can be re-attached with the SetAttached input.
        '''
        # FIXME: Should this have targetname?
        self.fire('Detach', targetname, userid=userid)


class prop_dynamic_override(_prop_dynamic_base):
    '''
    A prop that can be placed in hierarchy and can play animations.
    It can also be configured to break when it takes enough damage.
    
    prop_dynamic_override is a prototyping entity only. It will allow
    the use of models designed to be used as prop_physics.
    '''
    health = _KeyValue(0) # Number of points of damage to take before breaking.
        # 0 means don't break.


class prop_physics(_BasePropPhysics, _RenderFields):
    '''
    A prop that physically simulates as a single rigid body.
    It can be constrained to other physics objects using hinges or other
    constraints. It can also be configured to break when it takes enough
    damage. Note that the health of the object will be overridden by the
    health inside the model, to ensure consistent health game-wide.
    If the model used by the prop is configured to be used as a prop_dynamic
    (i.e. it should not be physically simulated) then it CANNOT be used as
    a prop_physics. Upon level load it will display a warning in the
    console and remove itself. Use a prop_dynamic instead.
    '''
    def Ignite(self, userid=0):
        '''
        Ignite, burst into flames.
        '''
        self.fire('Ignite', userid=userid)
    
    def IgniteLifetime(self, lifetime, userid=0):
        '''
        Ignite, with a parameter lifetime.
        '''
        self.fire('IgniteLifetime', lifetime, userid=userid)
    
    def IgniteNumHitboxFires(self, param, userid=0):
        '''
        Ignite, with a parameternumber of hitbox fires.
        '''
        self.fire('IgniteNumHitboxFires', param, userid=userid)
    
    def IgniteHitboxFireScale(self, param, userid=0):
        '''
        Ignite, with a parameter hitbox fire scale.
        '''
        self.fire('IgniteHitboxFireScale', param, userid=userid)


class prop_physics_multiplayer(prop_physics):
    '''
    This class is the same as prop_physics, except the runtime collisions
    use a more bouncy method that avoids the prediction errors normal
    physics objects get.
    '''
    physicsmode = _KeyValue(SymbolicValue(0, {
        'Auto Detect': 0,
        'Solid, Server-side': 1,
        'Non-Solid, Server-side': 2,
        'Non-Solid, Client-side': 3,
    })) # Physics Mode


class prop_physics_override(_BasePropPhysics):
    '''
    A prop that physically simulates as a single rigid body.
    It can be constrained to other physics objects using hinges or other
    constraints. It can also be configured to break when it takes enough
    damage. Health can be overridden on this version.
    '''
    health = _KeyValue(0) # Number of points of damage to take before breaking.
        # 0 means don't break.
    
    def Ignite(self, userid=0):
        '''
        Ignite, burst into flames.
        '''
        self.fire('Ignite', userid=userid)
    
    def IgniteLifetime(self, seconds, userid=0):
        '''
        Ignite, with a parameter lifetime.
        '''
        self.fire('IgniteLifetime', seconds, userid=userid)
    
    def IgniteNumHitboxFires(self, param, userid=0):
        '''
        Ignite, with a parameternumber of hitbox fires.
        '''
        self.fire('IgniteNumHitboxFires', param, userid=userid)
    
    def IgniteHitboxFireScale(self, param, userid=0):
        '''
        Ignite, with a parameter hitbox fire scale.
        '''
        self.fire('IgniteHitboxFireScale', param, userid=userid)


class prop_ragdoll(_Studiomodel, _DXLevelChoice, _BaseFadeProp, _EnableDisable):
    '''
    A prop that physically simulates and can be articulated with internal
    joints. The joint constraints are part of the physics model.
    '''
    spawnflags = _KeyValue(Flags({
        "Debris": 4,
        'Allow Dissolve': 8192,
        'Motion Disabled': 16384,
        'Allow stretch': 32768,
        'Start asleep': 65536,
    }, "Debris"))
    angleOverride = _KeyValue(''), # Filled in by the engine via wc_update_entity
        # do not edit by hand except to clear.
    
    def StartRagdollBoogie(self, userid=0):
        '''
        Begin ragdoll boogie effect. Parameter override = number of seconds to boogie.
        '''
        self.fire('StartRagdollBoogie', userid=userid)
    
    def EnableMotion(self, userid=0):
        '''
        Enable physics motion/collision response.
        '''
        self.fire('EnableMotion', userid=userid)
    
    def DisableMotion(self, userid=0):
        '''
        Disable physics motion/collision response.
        '''
        self.fire('DisableMotion', userid=userid)
    
    def FadeAndRemove(self, duration, userid=0):
        '''
        Fade out then remove (kill) self. Parameter = duration of fade
        '''
        self.fire('FadeAndRemove', duration, userid=userid)


class prop_static(_prop_static_base):
    '''
    A prop that doesn't move and doesn't animate.
    '''
    pass


class shadow_control(_Targetname):
    '''
    An entity to control the shadows in the map.
    '''
    angles = _KeyValue(Vector((80,30,0))) # This is the shadow direction.
        # Pitch is rotation around the Y axis, yaw is the rotation around
        # the Z axis, and roll is the rotation around the X axis.
    color = KeyvalueInput(Color(128,128,128))
        # This is the color of the shadows.
    distance = _KeyValue(75.0) # This is the maximum distance the shadow is allowed
        # to cast, in inches.
    disableallshadows = _KeyValue(SymbolicValue(0, {
        'No': 0,
        'Yes': 1,
    })) # All Shadows Disabled
    
    def direction(self, angles, userid=0):
        '''
        Set the shadow direction.
        '''
        self.fire('direction', Vector(angles).for_input(), userid=userid)
    
    def SetDistance(self, distance, userid=0):
        '''
        Set the maximum shadow cast distance.
        '''
        self.fire('SetDistance', distance, userid=userid)
    
    def SetAngles(self, angles, userid=0):
        '''
        Set the shadow direction.
        '''
        self.fire('SetAngles', Vector(angles).for_input(), userid=userid)
    
    def SetShadowsDisabled(self, disabled, userid=0):
        '''
        Set shadows disabled state.
        '''
        self.fire('SetShadowsDisabled', disabled, userid=userid)


class sky_camera(_BaseEntity):
    '''
    An entity used to control the 3D Skybox.
    Its origin is used to determine the 3D Skybox's position relative to the map.
    Place this entity, in the 3D Skybox, at the point where the origin of the map should be.
    '''
    scale = _KeyValue(16) # Scale of the skybox.
    fogenable = _KeyValue(SymbolicValue(0, {
        'No': 0,
        'Yes': 1,
    })) # Fog Enable
    fogblend = _KeyValue(SymbolicValue(0, {
        'No': 0,
        'Yes': 1,
    })) # Fog Blend
    use_angles = _KeyValue(SymbolicValue(0, {
        'No': 0,
        'Yes': 1,
    })) # Use Angles for Fog Dir
    fogcolor = _KeyValue(Color(255,255,255)) # Primary Fog Color
    fogcolor2 = _KeyValue(Color(255,255,255)) # Secondary Fog Color
    fogdir = _KeyValue(Vector((1,0,0))) # Primary Fog Dir
    fogstart = _KeyValue(500.0) # Distance at which the skybox fog should start.
    fogend = _KeyValue(2000.0) # Distance at which the skybox fog should be fully opaque.


class tanktrain_ai(_Targetname):
    '''
    Train chase AI
    '''
    target = _KeyValue('') # Train Name
    startsound = _KeyValue('vehicles/diesel_start1.wav') # Start Moving Sound
    enginesound = _KeyValue('vehicles/diesel_turbo_loop1.wav') # Engine Loop Sound
    movementsound = _KeyValue('vehicles/tank_treads_loop1.wav') # Vehicle Movement Sound


class tanktrain_aitarget(_Targetname):
    '''
    An entity that changes the target of a tanktrain_ai entity.
    '''
    target = _KeyValue('') # The tanktrain_ai entity to change the target of.
    newtarget = _KeyValue('') # The entity to tell the tanktrain_ai to target.


class test_traceline(_BaseEntity):
    '''
    A debugging tool for testing tracelines.
    '''
    pass


class trigger_autosave(_Targetname):
    '''
    A trigger volume that autosaves when the player touches it.
    '''
    master = _KeyValue('') # Legacy support: The name of a master entity.
        # If the master hasn't been activated, this entity will not activate.
    NewLevelUnit = _KeyValue(SymbolicValue(0, {
        'No': 0,
        'Yes': 1,
    })) # If set, the save will discard any savedata from previous levels,
        # for the purpose of keeping savegame filesizes down. Can only
        # be safely used if there is no way for the player to return to
        # previous levels.
    DangerousTimer = _KeyValue(0.0) # The number of seconds the player must survive
        # before this autosave takes effect.
    MinimumHitPoints = _KeyValue(0) # Don't save dangerous when player has less
        # than this many hitpoints.


class trigger_changelevel(_EnableDisable):
    '''
    An entity that triggers a level change.
    
    Place an info_landmark in both maps that marks the 'same' location
    in each map.
    
    TIPS & TRICKS: To fire events in the next level, use the OnLevelChange
    output to turn on an env_global in the current level. Create a logic_auto
    in the next level that checks for the state set by the env_global.
    
    To control which entities go through the level transition, create one
    or more trigger_transitions and give them the same name as the landmark.
    Any entities within the trigger_transition(s) will go to the next map.
    '''
    targetname = _KeyValue('') # Name
    map = _KeyValue('') # New Map Name
    landmark = _KeyValue('') # Landmark Name
    spawnflags = _KeyValue(Flags({
        'Disable Touch': 2,
        'To Previous Chapter': 4,
    }))
    
    def ChangeLevel(self, userid=0):
        '''
        Cause the level change. Use this when triggering the level
        change with a button, etc.
        '''
        self.fire('ChangeLevel', userid=userid)


class trigger_gravity(_Trigger):
    '''
    A trigger volume that changes the gravity on any entity that touches it.
    '''
    gravity = _KeyValue(RangedValue(1.0, 0.0, 1.0)) # Gravity (0-1)


class trigger_hurt(_Trigger, _Targetname):
    '''
    A trigger volume that damages entities that touch it.
    '''
    master = _KeyValue('') # Legacy support: The name of a master entity.
        # If the master hasn't been activated, this entity will not activate.
    damage = _KeyValue(10) # The amount of damage done to entities that touch
        # this trigger. The damage is done every half-second. See also
        # 'Damage Model' for extra details on how damage can be dealt.
    damagecap = _KeyValue(20) # Maximum damage dealt per second.
        # This field is only used if you select the Doubling
        # w/Forgiveness damage model, via the spawnflag.
    damagetype = _KeyValue(SymbolicValue(0, {
        'GENERIC': 0,
        'CRUSH': 1,
        'BULLET': 2,
        'SLASH': 4,
        'BURN': 8,
        'FREEZE': 16,
        'FALL': 32,
        'BLAST': 64,
        'CLUB': 128,
        'SHOCK': 256,
        'SONIC': 512,
        'ENERGYBEAM': 1024,
        'DROWN': 16384,
        'PARALYSE': 32768,
        'NERVEGAS': 65536,
        'POISON': 131072,
        'RADIATION': 262144,
        'DROWNRECOVER': 524288,
        'CHEMICAL': 1048576,
        'SLOWBURN': 2097152,
        'SLOWFREEZE': 4194304,
    })) # Damage Type
    damagemodel = _KeyValue(SymbolicValue(0, {
        'Normal': 0,
        'Doubling w/forgiveness': 1,
    })) # How damage is dealt. Normal always does the specified
        # amount of damage each half second. Doubling starts with
        # the specified amount and doubles it each time it hurts the
        # toucher. Forgiveness means that if the toucher gets out of
        # the trigger the damage will reset to the specified value.
        # Good for making triggers that are deadly over time without
        # having to cause massive damage on each touch.
    nodmgforce = _KeyValue(SymbolicValue(0, {
        'No': 0,
        'Yes': 1,
    })) # Should the damaged entity receive no physics force from this trigger.
    
    def SetDamage(self, damage, userid=0):
        '''
        Set a new amount of damage for this trigger.
        '''
        self.fire('SetDamage', damage, userid=userid)


class trigger_impact(_Targetname):
    '''
    A trigger volume that can be told to push all physics objects that
    are inside of it in the direction specified by this trigger's angles.
    
    Also outputs the force at the time of impact for anyone else that
    wants to use it.
    '''
    Magnitude = _KeyValue(200.0) # The strength of the impact.
        # Negative values reverse the direction.
    noise = _KeyValue(RangedValue(0.1, 0.0, 1.0)) # The amount of directional noise.
        # 0 = no noise, 1 = random direction.
    viewkick = _KeyValue(RangedValue(0.05, 0.0, 1.0)) # The amount to kick player's
        # view if the player is in the trigger.
        # Proportional to magnitude (0-1).
    
    def Impact(self, param, userid=0):
        '''
        Fire the impact, pushing all entities within the volume.
        '''
        # FIXME: what param?
        self.fire('Impact', param, userid=userid)
    
    def SetMagnitude(self, magnitude, userid=0):
        '''
        Set the magnitude of the impact.
        '''
        self.fire('SetMagnitude', magnitude, userid=userid)


class trigger_look(_Trigger):
    '''
    An entity used to trigger something when the player looks at something.
    It fires 'OnTrigger' when the player looks at a target entity for the
    given amount of time, while within the trigger volume. If the player
    leaves the trigger or looks away from the target entity the clock resets.
    If the 'Use Velocity instead of facing' spawnflag is checked, the trigger
    uses the player's velocity instead of the player's view, so it determines
    whenever the player is moving toward the target entity. Useful for
    triggering when players are driving a vehicle at something.
    NOTE: Only designed for single-player game. 
    '''
    spawnflags = _KeyValue(Flags({
        'Fire Once': 128,
        'Use Velocity instead of facing': 256,
    }, 'Fire Once'))
    target = _KeyValue('') # The name of the entity to be looked at.
    LookTime = _KeyValue(0.5) # The time, in seconds, that the player must look
        # the target before firing the output. Resets if player leaves
        # trigger, or looks outside the Field of View threshold.
    FieldOfView = _KeyValue(RangedValue(0.9, -1.0, 1.0)) # How close the player
        # has to be looking at the target.
        # 1.0 = straight ahead
        # 0.0 = +/- 90 degrees
        # -1.0 = all directions
    Timeout = _KeyValue(0.0) # The time, in seconds, to wait after player
        # enters the trigger before firing the OnTimeout output, 0 = never.


class trigger_multiple(_Trigger):
    '''
    A trigger volume that can be triggered multiple times.
    '''
    wait = _KeyValue(1) # Amount of time, in seconds, after the trigger_multiple
        # has triggered before it can be triggered again. If set to -1,
        # it will never trigger again (in which case you should
        # just use a trigger_once).
    
    def TouchTest(self, userid=0):
        '''
        Test if the trigger is being touched and fires an output
        based on whether the value is true or false.
        '''
        self.fire('TouchTest', userid=userid)


class trigger_once(_TriggerOnce):
    '''
    A trigger volume that removes itself after it is triggered once.
    '''
    pass


class trigger_playermovement(_Trigger):
    '''
    An entity that can be used to disable player's automatic
    ducking/unducking when jumping.
    '''
    spawnflags = _KeyValue(Flags({
        '(OBSOLETE, Uncheck me)': 16,
        'Disable auto player movement': 128,
        'Auto-duck while in trigger': 2048,
    }, 'Disable auto player movement'))


class trigger_proximity(_Trigger):
    '''
    Measure the distance of the player within the trigger volume from a given
    point (and within a given radius). The NearestPlayerDistance output will
    be 0 when the player is at the center point, and 1 when the player
    is at the radius.
    '''
    measuretarget = _KeyValue('') # The name of a target entity who's origin is
        # the point to measure the player's distance from.
    radius = _KeyValue(256) # The radius to which the distance should be mapped.
        # If the player is outside the radius he will be ignored.


class trigger_push(_Trigger):
    '''
    A trigger volume that pushes entities that touch it.
    '''
    pushdir = _KeyValue(Vector((0,0,0))) # Angles indicating the direction to
        # push touched entities.
    spawnflags = _KeyValue(Flags({
        'Once Only': 128,
        'Affects Ladders (Half-Life 2)': 256,
    }))
    speed = _KeyValue(40), # The speed at which to push entities away
        # in inches / second.
    alternateticksfix = _KeyValue(0.0) # If nonzero, scale the force by this amount
        # when running with alternate ticks. This fixes problems with
        # an overly large force due to the longer frametime on when
        # running with sv_alternateticks 1.


class trigger_remove(_Trigger, _Targetname):
    '''
    A trigger volume that removes any entities that touch it.
    Be careful, removing some entities can cause instability.
    This is not the same as killing entities. i.e. NPCs removed in
    this manner will not fire their OnKilled outputs.
    '''
    pass


class trigger_serverragdoll(_Targetname):
    '''
    A volume that forces any NPC inside it to create a server side
    ragdoll instead of a client one.
    '''
    pass


class trigger_soundscape(_Trigger):
    '''
    Soundscape trigger.
    It is not necessary to create outputs for this trigger.
    It automatically will trigger the soundscape referred to by
    its 'Soundscape' property.
    NEEDHELP?
    '''
    soundscape = _KeyValue('') # Soundscape


class trigger_teleport(_Trigger):
    '''
    A trigger volume that teleports entities that touch it.
    Entities are teleported to the Remote Destination, and have their
    angles set to that of the Remote Destination's.
    If a Local Destination Landmark is specified, teleported entities are
    offset from the target by their initial offset from the landmark,
    and their angles are left alone.
    '''
    target = _KeyValue('') # The entity specifying the point to which
        # entities should be teleported.
    landmark = _KeyValue('') # If specified, then teleported entities are offset
        # from the target by their initial offset from the landmark.
    spawnflags = _KeyValue(Flags({
        'Preserve angles even when a local landmark is not specified': 32,
    }))


class trigger_transition(_Targetname):
    '''
    A volume that's used to control which entities go through the level
    transition. Create one or more trigger_transitions and give them the same
    name as the changelevel landmark. Any entities within the
    trigger_transition(s) will go to the next map.
    See trigger_changelevel for more info.
    '''
    pass


class trigger_wind(_Trigger):
    '''
    A trigger volume that pushes physics objects that touch it.
    '''
    Speed = _KeyValue(200) # The baseline for how hard the wind blows.
    SpeedNoise = _KeyValue(0) # Noise added to wind speed +/-
    DirectionNoise = _KeyValue(10) # Noise added to wind direction.
    HoldTime = _KeyValue(0) # Baseline for how long to wait before changing wind.
    HoldNoise = _KeyValue(0) # Noise added to how long to wait before changing wind.
    
    def SetSpeed(self, speed, userid=0):
        '''
        Set the baseline for how hard the wind blows.
        '''
        self.fire('SetSpeed', speed, userid=userid)


class vgui_screen(_vgui_screen_base):
    '''
    A VGUI screen. Useful for in-world monitors.
    '''
    pass


class vgui_slideshow_display(_Targetname, _Parentname):
    '''
    Slideshow Display.
    '''
    nolistrepeat_values = {
        "Allow List Repeats": 0,
        "No List Repeats": 1,
    }
 
    cycletype_values = {
        "Random": 0,
        "Forward": 1,
        "Backward": 2,
    }
 
    maxslidetime = _KeyValue(0.5)
    nolistrepeat = _KeyValue(SymbolicValue("Allow List Repeats", nolistrepeat_values))
    height = _KeyValue(128)
    width = _KeyValue(256)
    cycletype = _KeyValue(SymbolicValue("Random", cycletype_values))
    minslidetime = _KeyValue(0.5)
    directory = _KeyValue("slideshow")
    displaytext = _KeyValue("")
 
    def Enable(self, userid=0):
        '''
        Make slideshow visible.
        '''
        self.fire('Enable', userid=userid)
 
    def Disable(self, userid=0):
        '''
        Make slideshow invisible.
        '''
        self.fire('Disable', userid=userid)
 
    def SetDisplayText(self, text, userid=0):
        '''
        Set the display text.
        '''
        self.fire('SetDisplayText', text, userid=userid)
 
    def RemoveAllSlides(self, userid=0):
        '''
        Remove all slides from slideshow.
        '''
        self.fire('RemoveAllSlides', userid=userid)
 
    def AddSlides(self, keyword, userid=0):
        '''
        Add slides by keyword.
        '''
        self.fire('AddSlides', keyword, userid=userid)
 
    def SetMinSlideTime(self, seconds, userid=0):
        '''
        Set min random time between slides.
        '''
        self.fire('SetMinSlideTime', seconds, userid=userid)
 
    def SetMaxSlideTime(self, seconds, userid=0):
        '''
        Set max random time between slides.
        '''
        self.fire('SetMaxSlideTime', seconds, userid=userid)
 
    def SetCycleType(self, value, userid=0):
        '''
        0 - random, 1 - forward, 2 - backward
        '''
        self.fire('SetCycleType', value, userid=userid)
 
    def SetNoListRepeat(self, value, userid=0):
        '''
        Set if lists can be randomly picked twice in a row.
        '''
        self.fire('SetNoListRepeat', value, userid=userid)


class water_lod_control(_Targetname):
    '''
    An entity used to control the LOD behavior of any water in the map.
    If your map has water, this entity is required.
    '''
    cheapwaterstartdistance = _KeyValue(1000.0) # This is the distance from the
        # camera that water will start transitioning to cheap water,
        # in inches.
    cheapwaterenddistance = _KeyValue(2000.0) # This is the distance from the
        # camera that water will finish transitioning to cheap water,
        # in inches.
    
    def SetCheapWaterStartDistance(self, param, userid=0):
        '''
        Set the distance that water starts transitioning to cheap water.
        '''
        self.fire('SetCheapWaterStartDistance', param, userid=userid)
    
    def SetCheapWaterEndDistance(self, param, userid=0):
        '''
        Set the distance that water finishes transitioning to cheap water.
        '''
        self.fire('SetCheapWaterEndDistance', param, userid=userid)


class worldspawn(_Targetname, _worldbase, _ResponseContext):
    '''
    This is the world entity.
    '''
    pass

